package filter;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import utils.Activity;
import utils.Pattern;

public class ActivityPatternFinder extends Filter
{
	private static String newLine = System.getProperties().getProperty("line.separator");
	private static String separator = "------------";

	private List<String> activityDescriptions;
	private Activity[] slidingWindow;
	private List<Pattern> patterns;
	private int nextIndex;
	private int filter;

	private boolean includeHierarchyInfo = false;
	private boolean ignoreSameActivities = true;
	private boolean verbose = false;


	public ActivityPatternFinder(int windowSize, String[] activities, int filter)
	{
		this.activityDescriptions = Arrays.asList(activities);
		slidingWindow = new Activity[windowSize];
		patterns = new LinkedList<Pattern>();
		nextIndex = 0;

		this.filter = filter;
	}


	public int getWindowSize()
	{
		return slidingWindow.length;
	}


	@Override
	public void computeInfo(String content)
	{
		Activity currentActivity = new Activity(content);

		if (!activityDescriptions.contains(currentActivity.getDescription()))
		{
			return;
		}

		if (ignoreSameActivities && nextIndex > 0
				&& currentActivity.getDescription().equals(slidingWindow[nextIndex - 1].getDescription()))
		{
			return;
		}

		slidingWindow[nextIndex++] = currentActivity;

		if (nextIndex >= slidingWindow.length)
		{
			String fingerprintOfSlidingWindow = Pattern.computeFingerprint(slidingWindow, includeHierarchyInfo);
			boolean patternContained = false;

			for (Pattern pattern : patterns)
			{
				if (pattern.getFingerprint().equals(fingerprintOfSlidingWindow))
				{
					pattern.addActivitySequenz(Arrays.copyOf(slidingWindow,slidingWindow.length));
					patternContained = true;
					break;
				}
			}

			if (!patternContained)
			{
				Pattern pattern = new Pattern();
				pattern.setFingerprint(fingerprintOfSlidingWindow);
				pattern.addActivitySequenz(Arrays.copyOf(slidingWindow,slidingWindow.length));
				
				patterns.add(pattern);
			}

			// Slide
			for (int i = 0; i < slidingWindow.length; i++)
			{
				if (i + 1 < slidingWindow.length)
				{
					slidingWindow[i] = slidingWindow[i + 1];
				}
			}

			nextIndex = slidingWindow.length - 1;
		}
	}


	@Override
	public String getInfo()
	{
		info += "Analysis of activity patterns".toUpperCase() + newLine;

		info += "Sliding window size: " + slidingWindow.length + newLine;

		info += "Filter numbers lower than: " + filter + newLine;
		
		info += "Highest count: " + Pattern.highestCount + newLine;

		info += "--- Activity patterns ---" + newLine;

		for (Pattern pattern: patterns)
		{
			if (pattern.getNumberOfPatterns() >= filter)
			{
				info += pattern.getFingerprint() + ": " + pattern.getNumberOfPatterns() + newLine;
				
				if (verbose)
				{
					info += separator + newLine;
					
					for (Activity[] activitySequenz : pattern.getActivityPatterns())
					{
						for (Activity activity : activitySequenz)
						{
							info += activity.toString() + newLine;
						}
						
						info += separator + newLine;
					}
				}
			}
		}

		return info;
	}
}
