package plg.overlay;/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

import com.intellij.openapi.Disposable;
import com.intellij.openapi.actionSystem.DataKey;
import com.intellij.openapi.actionSystem.DataProvider;
import com.intellij.openapi.editor.*;
import com.intellij.openapi.editor.ex.FoldingListener;
import com.intellij.openapi.editor.impl.FoldingModelImpl;
import com.intellij.openapi.project.Project;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import java.awt.*;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragGestureRecognizer;
import java.awt.dnd.DragSource;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

/**
 * Created by Peter on 05.07.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public final class OverlayComponent extends JComponent implements Disposable, DataProvider {
    public static final DataKey<OverlayComponent> DATA_KEY = DataKey.create("sketchlink.overlayComponent");
    @NotNull
    private final OverlayElement myOverlayElement;
    @Nullable
    private final PropertyChangeListener propertyChangeListener;
    @NotNull
    private final DragSource dragSource;
    @NotNull
    private final OverlayContainer overlayContainer;
    @NotNull
    private final FoldingListener foldingListener;
    private final OverlayElementCustomDrawable overlayDrawing;
    private DragGestureRecognizer defaultDragGestureRecognizer;


    protected OverlayComponent(@NotNull OverlayElement element, @NotNull OverlayContainer container) {
        overlayContainer = container;
        myOverlayElement = element;
        if (element instanceof OverlayElementCustomDrawable) {
            overlayDrawing = (OverlayElementCustomDrawable) element;
        } else {
            overlayDrawing = new MyDefaultOverlayElementDrawable(101, 2);
        }
        propertyChangeListener = createPropertyChangeListener(this);
        myOverlayElement.addPropertyChangeListener(propertyChangeListener);

        @Nullable TransferHandler transferHandler = element.getTransferHandler();
        if (transferHandler != null) {
            setTransferHandler(transferHandler);
        }
        @Nullable DragGestureListener dragGestureListener = element.getDragGestureListener();
        dragSource = new DragSource();
        if (dragGestureListener != null) {
            defaultDragGestureRecognizer =
                    dragSource.createDefaultDragGestureRecognizer(this, DnDConstants.ACTION_LINK, dragGestureListener);
        }
        @Nullable MouseListener mouseListener = element.getMouseListener();
        if (mouseListener != null) {
            this.addMouseListener(mouseListener);
        }
        @Nullable MouseMotionListener mouseMotionListener = element.getMouseMotionListener();
        if (mouseMotionListener != null) {
            this.addMouseMotionListener(mouseMotionListener);
        }
        //set the size, because width=0,height=0 isn't drawn
        overlayDrawing.updateLocationAndSize(this);
        this.setCursor(element.getMouseCursor());
        //create and register Folding Listener
        foldingListener = new FoldingListener() {
            /**
             * Informs that <code>'collapsed'</code> state of given fold region is just changed.
             * <p/>
             * <b>Note:</b> listener should delay fold region state processing until {@link #onFoldProcessingEnd()} is called.
             * I.e. folding model may return inconsistent data between current moment and {@link #onFoldProcessingEnd()}.
             *
             * @param region fold region that is just collapsed or expanded
             */
            @Override
            public void onFoldRegionStateChange(@NotNull FoldRegion region) {
                @NotNull OverlayElement overlayElement = OverlayComponent.this.getOverlayElement();
                int offset = overlayElement.getOffset(region.getEditor());
                if (region.getStartOffset() < offset && offset < region.getEndOffset()) {
                    OverlayComponent.this.setVisible(region.isExpanded());
                }
            }

            /**
             * Informs that fold processing is done.
             */
            @Override
            public void onFoldProcessingEnd() {

            }
        };
        registerFoldingListener();
    }

    @NotNull
    public OverlayElement getOverlayElement() {
        return myOverlayElement;
    }


    private void registerFoldingListener() {
        @NotNull Editor editor = overlayContainer.getEditor();
        if (!this.isEnabled()) {
            return;
        }
        @NotNull FoldingModel foldingModel = editor.getFoldingModel();
        if (foldingModel instanceof FoldingModelImpl) {
            ((FoldingModelImpl) foldingModel).addListener(foldingListener, this);
        }
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        OverlayComponent that = (OverlayComponent) o;

        if (defaultDragGestureRecognizer != null ? !defaultDragGestureRecognizer.equals(that.defaultDragGestureRecognizer) : that.defaultDragGestureRecognizer != null)
            return false;
        if (dragSource != null ? !dragSource.equals(that.dragSource) : that.dragSource != null) return false;
        if (!foldingListener.equals(that.foldingListener)) return false;
        if (!myOverlayElement.equals(that.myOverlayElement)) return false;
        if (!overlayContainer.equals(that.overlayContainer)) return false;
        if (!overlayDrawing.equals(that.overlayDrawing)) return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = myOverlayElement.hashCode();
        result = 31 * result + (dragSource != null ? dragSource.hashCode() : 0);
        result = 31 * result + overlayContainer.hashCode();
        result = 31 * result + foldingListener.hashCode();
        result = 31 * result + (defaultDragGestureRecognizer != null ? defaultDragGestureRecognizer.hashCode() : 0);
        result = 31 * result + overlayDrawing.hashCode();
        return result;
    }

    @NotNull
    private PropertyChangeListener createPropertyChangeListener(@NotNull final OverlayComponent myself) {
        return new PropertyChangeListener() {


            @Override
            public void propertyChange(@NotNull PropertyChangeEvent evt) {
                if ("transferHandler".equals(evt.getPropertyName())) {
                    myself.setTransferHandler((TransferHandler) evt.getNewValue());
                }
                if ("mouseListener".equals(evt.getPropertyName())) {
                    @NotNull MouseListener oldValue = (MouseListener) evt.getOldValue();
                    @NotNull MouseListener newValue = (MouseListener) evt.getNewValue();
                    if (oldValue != null) {
                        myself.removeMouseListener(oldValue);
                    }
                    if (newValue != null) {
                        myself.addMouseListener(newValue);
                    }
                }
                if ("mouseMotionListener".equals(evt.getPropertyName())) {
                    @NotNull MouseMotionListener oldValue = (MouseMotionListener) evt.getOldValue();
                    @NotNull MouseMotionListener newValue = (MouseMotionListener) evt.getNewValue();
                    if (oldValue != null) {
                        myself.removeMouseMotionListener(oldValue);
                    }
                    if (newValue != null) {
                        myself.addMouseMotionListener(newValue);
                    }
                }
                if ("positionalElement".equals(evt.getPropertyName())) {
                    @NotNull PsiElement element = (PsiElement) evt.getNewValue();
                    if (element == null || !element.isValid()) {
                        myself.setVisible(false);
                        myself.setEnabled(false);
                    }
                }
                if ("valid".equals(evt.getPropertyName())) {
                    @NotNull Boolean state = (Boolean) evt.getNewValue();
                    if (!state) {
                        myself.setVisible(false);
                        myself.setEnabled(false);
                    }
                }


            }
        };
    }

    @Override
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);
        updateLocationAndSize();
        if (!getOverlayElement().isValid()) {
            unregisterOverlayElement();
            return;
        }
        overlayDrawing.paintComponent(this, g);

    }

    @Nullable
    public Object getData(@NonNls String dataId) {
        if (DATA_KEY.getName().equals(dataId)) {
            return this;
        }
        if (OverlayElement.OVERLAY_ELEMENT_DATA_KEY.getName().equals(dataId)) {
            return myOverlayElement;
        }
        return getOverlayElement().getData(dataId);
    }

    @Override
    public void dispose() {
        System.out.println("plg.overlay.OverlayComponent disposed");
        for (MouseListener mouseListener : this.getMouseListeners()) {
            this.removeMouseListener(mouseListener);
        }
        for (MouseMotionListener listener : this.getMouseMotionListeners()) {
            this.removeMouseMotionListener(listener);
        }
        if (propertyChangeListener != null)
            myOverlayElement.removePropertyChangeListener(propertyChangeListener);
        this.setEnabled(false);
        this.setVisible(false);
    }

    private void unregisterOverlayElement() {
        myOverlayElement.setValid(false);
        this.setVisible(false);
        this.setEnabled(false);
        @Nullable Project project = overlayContainer.getEditor().getProject();

        if (project == null) {
            overlayContainer.removeOverlayElement(getOverlayElement());
            return;
        }
        @Nullable PsiFile psiFile = PsiDocumentManager.getInstance(project)
                .getPsiFile(overlayContainer.getEditor().getDocument());
        if (psiFile == null) {
            overlayContainer.removeOverlayElement(getOverlayElement());
            return;
        }
        //remove this item for all other container too
        //TODO: maybe execute in different thread?
        SketchlinkOverlayManager.getInstance(project).removeItem(getOverlayElement(), psiFile);
    }


    /**
     * Forces this component to reevaluate its Location and Size,
     * may cause this component to become unregistered/removed from the editor overlay at next redrawing
     */
    public void updateLocationAndSize() {
        overlayDrawing.updateLocationAndSize(this);

    }


    /**
     * Default implementation of drawing the icon above an editor and calculating its location and size
     */
    private class MyDefaultOverlayElementDrawable implements OverlayElementCustomDrawable {
        private final int skipIfDistanceGreaterAs;
        private final int maxNumberOfSkips;
        /**
         * Counter for Location updates that where skipped, because the new location resulted in a to
         * big a jump.
         */
        int skippedLocationUpdates = 0;
        private float scale = 1.f;
        private Point lastUpdateLocation = new Point(0, 0);

        /**
         * Constructor
         *
         * @param skipLocationPixelThreshold skip updating the components Location, if the new Location is farther away from the old location.
         *                                   Location is immediately updated if distance (in pixel) is less then parameter.
         * @param maxSuccessiveThresholdSkip maximal amount of times, an update in location can be skipped. Recommended: 1 or 2, to reduce lag
         *                                   in valid location updates.
         */
        public MyDefaultOverlayElementDrawable(int skipLocationPixelThreshold, int maxSuccessiveThresholdSkip) {
            skipIfDistanceGreaterAs = skipLocationPixelThreshold * skipLocationPixelThreshold;
            maxNumberOfSkips = maxSuccessiveThresholdSkip;
        }

        @Override
        public void updateLocationAndSize(@NotNull OverlayComponent component) {
            //resize the icon if editorFont is changed
            scale = setSize(component);
            //calculate the actual xy point
            setLocation(component);
        }

        @Override
        public void paintComponent(@NotNull OverlayComponent component, @NotNull Graphics g) {
            if (!component.isEnabled() || !component.isVisible()) {
                return;
            }
            @Nullable PsiElement positionalElement = component.getOverlayElement().getPositionalElement();
            if (positionalElement == null) {
                getOverlayElement().setValid(false);
                return;
            }

//            System.out.println("Location: "+getLocation().x+" "+getLocation().y+"\t"+getDistanceSQR(getLocation(),lastUpdateLocation)+"\t"+lastUpdateLocation);
            //test if the cursor is beneath the icon
            if (isCursorBeneathIcon(component)) {
                if (isCursorAtIconStart(component)) {
                    //cut off the first two pixel of the image to show the cursor beneath
                    @NotNull ImageIcon imageIcon = component.getOverlayElement().getImageIcon();
                    //depend on cursor being 2 pixel wide. If the scaling is 2-times the size, only cut off the first horizontal pixel of the image.
                    int sx1 = scale > 0.f ? Math.round(2.f / scale) : 0;
                    g.drawImage(imageIcon.getImage(), 2, 0, getWidth(), getHeight(), sx1, 0, imageIcon.getIconWidth(), imageIcon.getIconHeight(), component);
                }
                return;
            }
            //finally draw the image
            @NotNull ImageIcon imageIcon = getOverlayElement().getImageIcon();
            g.drawImage(imageIcon.getImage(), 0, 0, getWidth(), getHeight(), component);

        }

        private float setSize(@NotNull OverlayComponent component) {
            int lineHeight = component.overlayContainer.getEditor().getLineHeight(); //default: 17
            @NotNull Dimension preferredSize = getOverlayElement().getPreferredImageSize();
            float scaleFactor = 1.f * lineHeight / 17;
            int width = Math.round(scaleFactor * preferredSize.width);
            int height = Math.round(scaleFactor * preferredSize.height);
            component.setSize(width, height);
            return scaleFactor;
        }

        private boolean isCursorAtIconStart(@NotNull OverlayComponent component) {
            @NotNull Editor editor = component.overlayContainer.getEditor();
            @NotNull CaretModel caretModel = editor.getCaretModel();
            @NotNull VisualPosition caretVisual = caretModel.getVisualPosition();
            @NotNull Point point = editor.visualPositionToXY(caretVisual);
            point.x++;
            point.y++;
            @NotNull VisualPosition oneStepLeft = new VisualPosition(caretVisual.line, caretVisual.column > 0 ? caretVisual.column - 1 : caretVisual.column);
            @NotNull Point stepLeftPoint = editor.visualPositionToXY(oneStepLeft);
            stepLeftPoint.x++;
            stepLeftPoint.y++;
            return !component.getBounds().contains(stepLeftPoint) && component.getBounds().contains(point);
        }

        private boolean isCursorBeneathIcon(@NotNull OverlayComponent component) {
            @NotNull Editor editor = component.overlayContainer.getEditor();
            @NotNull CaretModel caretModel = editor.getCaretModel();
            @NotNull VisualPosition caretVisual = caretModel.getVisualPosition();
            @NotNull Point point = editor.visualPositionToXY(caretVisual);
            point.x++;
            point.y++;
            return component.getBounds().contains(point);
        }

        void setLocation(@NotNull OverlayComponent component) {
            if (!getOverlayElement().isValid()) {
                return;
            }
            @NotNull Editor editor = component.overlayContainer.getEditor();
            @NotNull ImageIcon imageIcon = getOverlayElement().getImageIcon();
            int offset = getOverlayElement().getOffset(editor);
            Document document = editor.getDocument();
            if (offset < 0 || document.getTextLength() < offset) {
                throw new IndexOutOfBoundsException(getOverlayElement().getClass().getName() + "#getOffset returned an offset outside of the length of the document."
                        + "\n\tReturned offset: " + offset + " for document length: " + document.getTextLength() + " of: " + document
                        + "\n\tProblematic OverlayElement implementation: " + getOverlayElement() + " (isValid() returns " + getOverlayElement().isValid() + ") ");
            }
            @NotNull VisualPosition iconVisualPosition =
                    editor.offsetToVisualPosition(offset);
            @NotNull Point iconPoint = editor.visualPositionToXY(iconVisualPosition);
            iconPoint.x = iconPoint.x
                    + getOverlayElement().getIconAlignment().getAdditionalVisualOffset(imageIcon.getIconWidth());
            //outer_left alignment and position at the beginning on a line could move the x position outside of the editor
            if (iconPoint.x < 0) {
                iconPoint.x = 0;
            } else {
            /*the editor's "peek" function can cause the current icon position to be set outside of the normal editor
             component. Since it is outside of the visible area, the icon will not be redrawn again, which normally causes
             the position update.
             Return now and do not set the location - unneccessary in these cases anyway.
              */
                if (iconPoint.x + getWidth() > component.overlayContainer.getCurrentContainerSize().width) {
                    return;
                }
            }
            Point location = getLocation();
            if (!location.equals(iconPoint)) {
                int delta = getDistanceSQR(location, iconPoint);
                //only change the location on large differences, if this location was calculated two times in a row.
                if (delta < skipIfDistanceGreaterAs || (iconPoint.equals(lastUpdateLocation) && skippedLocationUpdates > maxNumberOfSkips) || (skippedLocationUpdates > (maxNumberOfSkips + 10))) {
                    component.setLocation(iconPoint);
                    skippedLocationUpdates = 0;
                } else {
                    skippedLocationUpdates++;
                }
            } else {
                skippedLocationUpdates = 0;
            }
            lastUpdateLocation = iconPoint;
        }

        private int getDistanceSQR(@NotNull Point p1, @NotNull Point p2) {
            int deltaX = p1.x - p2.x;
            int deltaY = p1.y - p2.y;
            return deltaX * deltaX + deltaY * deltaY;
        }

    }

}
