package plg.overlay;/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

import com.intellij.openapi.actionSystem.DataKey;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.util.TextRange;
import com.intellij.psi.PsiElement;
import com.intellij.psi.SmartPointerManager;
import com.intellij.psi.SmartPsiElementPointer;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import java.awt.*;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragSourceAdapter;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by Peter on 05.07.2014.
 * Default implementation of an Image that can be Overlayed above the editor.
 *
 */
public class OverlayElementImpl<T extends PsiElement> implements OverlayElement<T> {
    private boolean valid = true;
    private SmartPsiElementPointer<T> smartPsiElementPointer;
    @NotNull
    private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);
    private ImageIcon imageIcon;
    @Nullable
    private MouseListener mouseListener;
    @Nullable
    private MouseMotionListener mouseMotionListener;
    @Nullable
    private TransferHandler transferHandler;
    @Nullable
    private DragSourceAdapter dragsourceAdapter;
    private IconAlignment iconAlignment;
    @Nullable
    private DragGestureListener dragGestureListener;
    private Cursor mouseCursor = Cursor.getPredefinedCursor(Cursor.HAND_CURSOR);
    @NotNull
    private Map<String, Object> dataKeyObjectMap = new ConcurrentHashMap<>();
    private Dimension preferredImageSize;
    private boolean allowLineWrapsForPosition;


    /**
     * Creates a new OverlayElementImpl.
     *
     * @param positionElement           the PsiElement which offsets are used to determine the location of the icon.
     * @param icon                      the icon to add to the editor
     * @param allowLineWrapsForPosition if the PsiElement contains newline characters: false to use the end of the first line as for "right"-alignments
     */
    public OverlayElementImpl(@NotNull T positionElement, @NotNull ImageIcon icon, boolean allowLineWrapsForPosition) {
        smartPsiElementPointer =
                SmartPointerManager.getInstance(positionElement.getProject()).createSmartPsiElementPointer(positionElement, positionElement.getContainingFile());
        setImageIcon(icon);
        setIconAlignment(IconAlignment.INNER_LEFT);
        setAllowLineWrapsForPosition(allowLineWrapsForPosition);
    }

    public OverlayElementImpl(@NotNull T positionElement, @NotNull ImageIcon icon) {
        this(positionElement, icon, true);
    }

    /**
     * If false, interprets the first newline-character "\n" in the text of the PositionalElement as
     * the end of this Element.
     * Mostly important for right-alignments.
     * e.g. the icon position marked by [boolean], when using a PsiMethod as multiline-position element and outer-right:
     * <p/>
     * public void method() {[true]
     * //text
     * }[false]
     */
    public boolean isAllowLineWrapsForPosition() {
        return allowLineWrapsForPosition;
    }

    /**
     * If false and a right-alignment is used, forces the Icon to appear at the end of the first line, if
     * the PositionalElement spans several lines.
     * No effect if true, PosEl spanning a single line or an left-alignment is used
     */
    public void setAllowLineWrapsForPosition(boolean allowLineWrapsForPosition) {
        this.allowLineWrapsForPosition = allowLineWrapsForPosition;
    }

    @Override
    @NotNull
    public IconAlignment getIconAlignment() {
        return iconAlignment;
    }

    @Override
    public void setIconAlignment(@NotNull IconAlignment iconAlignment) {
        if (iconAlignment == this.iconAlignment) {
            return;
        }
        IconAlignment oldValue = this.iconAlignment;
        this.iconAlignment = iconAlignment;
        propertyChangeSupport.firePropertyChange("iconAlignment", oldValue, this.iconAlignment);
    }

    @Override
    @Nullable
    public DragSourceAdapter getDragsourceAdapter() {
        return dragsourceAdapter;
    }


    @Override
    public void setDragsourceAdapter(@Nullable DragSourceAdapter dragsourceAdapter) {
        if (dragsourceAdapter == this.dragsourceAdapter) {
            return;
        }
        @Nullable DragSourceAdapter oldValue = this.dragsourceAdapter;
        this.dragsourceAdapter = dragsourceAdapter;
        propertyChangeSupport.firePropertyChange("dragSourceAdapter", oldValue, this.dragsourceAdapter);
    }

    @Override
    @Nullable
    public TransferHandler getTransferHandler() {
        return transferHandler;
    }

    /**
     * Sets the Transferhandler, important to handle the DROP in dragNdrop actions.
     */
    @Override
    public void setTransferHandler(@Nullable TransferHandler transferHandler) {
        if (transferHandler == this.transferHandler) {
            return;
        }
        @Nullable TransferHandler oldValue = this.transferHandler;
        this.transferHandler = transferHandler;
        propertyChangeSupport.firePropertyChange("transferHandler", oldValue, this.transferHandler);
    }

    @Override
    @Nullable
    public DragGestureListener getDragGestureListener() {
        return dragGestureListener;
    }

    /**
     * Sets the Listener, used when dragNDrop STARTS from this Icon
     */
    @Override
    public void setDragGestureListener(@Nullable DragGestureListener dragGestureListener) {
        if (dragGestureListener == this.dragGestureListener) {
            return;
        }
        @Nullable DragGestureListener oldValue = this.dragGestureListener;
        this.dragGestureListener = dragGestureListener;
        propertyChangeSupport.firePropertyChange("dragSourceAdapter", oldValue, this.dragGestureListener);
    }

    @Override
    @NotNull
    public Cursor getMouseCursor() {
        return mouseCursor;
    }

    /**
     * Set the cursor that is used if the mouse is over the image.
     *
     * @param mouseCursor
     */
    @Override
    public void setMouseCursor(@NotNull Cursor mouseCursor) {
        if (mouseCursor.equals(this.mouseCursor)) {
            return;
        }
        Cursor oldValue = this.mouseCursor;
        this.mouseCursor = mouseCursor;
        propertyChangeSupport.firePropertyChange("mouseCursor", oldValue, this.mouseCursor);
    }

    @Override
    @Nullable
    public MouseListener getMouseListener() {
        return mouseListener;
    }

    @Override
    public void setMouseListener(@Nullable MouseListener mouseListener) {
        if (this.mouseListener == mouseListener) {
            return;
        }
        @Nullable MouseListener oldValue = this.mouseListener;
        this.mouseListener = mouseListener;
        propertyChangeSupport.firePropertyChange("mouseListener", oldValue, this.mouseListener);
    }

    /**
     * Gets the MouseMotionListener active for this Icon
     */
    @Override
    @Nullable
    public MouseMotionListener getMouseMotionListener() {
        return mouseMotionListener;
    }

    @Override
    public void setMouseMotionListener(@Nullable MouseMotionListener mouseMotionListener) {
        if (this.mouseMotionListener == mouseMotionListener) {
            return;
        }
        @Nullable MouseMotionListener oldValue = this.mouseMotionListener;
        this.mouseMotionListener = mouseMotionListener;
        propertyChangeSupport.firePropertyChange("mouseMotionListener", oldValue, this.mouseMotionListener);
    }

    /**
     * Gets the Element that is used to calculate the position of this Overlay Element.
     */
    @Override
    @Nullable
    public T getPositionalElement() {
        return smartPsiElementPointer.getElement();
    }

    @Override
    public void setPositionalElement(@NotNull T element) {
        @Nullable T oldValue = getPositionalElement();
        if (element.equals(oldValue)) {
            return;
        }
        smartPsiElementPointer =
                SmartPointerManager.getInstance(element.getProject()).createSmartPsiElementPointer(element, element.getContainingFile());
        propertyChangeSupport.firePropertyChange("positionalElement", oldValue, element);
    }


    /**
     * Gets the ImageIcon that will be drawn. Might be resized earlier.
     */
    @Override
    @NotNull
    public ImageIcon getImageIcon() {
        return imageIcon;
    }

    @Override
    public void setImageIcon(@NotNull ImageIcon icon) {
        if (icon.equals(imageIcon)) {
            return;
        }
        ImageIcon oldIcon = imageIcon;
        imageIcon = icon;
        propertyChangeSupport.firePropertyChange("imageIcon", oldIcon, imageIcon);
        setPreferredImageSize(new Dimension(imageIcon.getIconWidth(), imageIcon.getIconHeight()));


    }


    @Override
    public void addPropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener(listener);
    }

    @Override
    public void removePropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener(listener);
    }

    /**
     * Gets the validity of this object. If false, this object will be removed from the Overlay.
     * If true, this plg.overlay.OverlayElementImpl will be drawn.
     */
    @Override
    public boolean isValid() {
        if (smartPsiElementPointer.getElement() == null || !smartPsiElementPointer.getElement().isValid()) {
            setValid(false);
        }
        return valid;
    }

    @Override
    public void setValid(boolean valid) {
        if (this.valid == valid) {
            return;
        }
        this.valid = valid;
        propertyChangeSupport.firePropertyChange("valid", !valid, this.valid);
    }


    /**
     * Returns the offset in the file above where to draw the icon.
     */
    @Override
    public int getOffset(@NotNull Editor editor) {
        @Nullable T positionalElement = getPositionalElement();
        if (positionalElement == null || !positionalElement.isValid()) {
            return 0;
        }
        TextRange range = positionalElement.getTextRange();
        if (!isAllowLineWrapsForPosition() && positionalElement.getText().contains("\n") && getIconAlignment().isLineWrapDependend()) {
            int lengthToNewLine = positionalElement.getText().indexOf("\n");
            lengthToNewLine = lengthToNewLine > 0 ? lengthToNewLine : 0;
            range = new TextRange(range.getStartOffset(), range.getStartOffset() + lengthToNewLine);
        }
        int offset = getIconAlignment().getOffset(range);
        return Math.min(offset, editor.getDocument().getTextLength());
    }

    @Override
    public boolean equals(@Nullable Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        @NotNull OverlayElementImpl that = (OverlayElementImpl) o;

        if (valid != that.valid) return false;
        if (dragGestureListener != null ? !dragGestureListener.equals(that.dragGestureListener) : that.dragGestureListener != null)
            return false;
        if (dragsourceAdapter != null ? !dragsourceAdapter.equals(that.dragsourceAdapter) : that.dragsourceAdapter != null)
            return false;
        if (iconAlignment != that.iconAlignment) return false;
        if (!imageIcon.equals(that.imageIcon)) return false;
        if (mouseListener != null ? !mouseListener.equals(that.mouseListener) : that.mouseListener != null)
            return false;
        if (mouseMotionListener != null ? !mouseMotionListener.equals(that.mouseMotionListener) : that.mouseMotionListener != null)
            return false;
        if (propertyChangeSupport != null ? !propertyChangeSupport.equals(that.propertyChangeSupport) : that.propertyChangeSupport != null)
            return false;
        if (!smartPsiElementPointer.equals(that.smartPsiElementPointer)) return false;
        if (transferHandler != null ? !transferHandler.equals(that.transferHandler) : that.transferHandler != null)
            return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = (valid ? 1 : 0);
        result = 31 * result + smartPsiElementPointer.hashCode();
        result = 31 * result + (propertyChangeSupport != null ? propertyChangeSupport.hashCode() : 0);
        result = 31 * result + imageIcon.hashCode();
        result = 31 * result + (mouseListener != null ? mouseListener.hashCode() : 0);
        result = 31 * result + (mouseMotionListener != null ? mouseMotionListener.hashCode() : 0);
        result = 31 * result + (transferHandler != null ? transferHandler.hashCode() : 0);
        result = 31 * result + (dragsourceAdapter != null ? dragsourceAdapter.hashCode() : 0);
        result = 31 * result + iconAlignment.hashCode();
        result = 31 * result + (dragGestureListener != null ? dragGestureListener.hashCode() : 0);
        return result;
    }

    @Nullable
    @Override
    public Object getData(@Nullable @NonNls String dataId) {
        if (dataId == null) {
            return null;
        }
        return dataKeyObjectMap.get(dataId);
    }

    public void putData(@NotNull DataKey key, @Nullable Object entry) {
        if (entry == null) {
            dataKeyObjectMap.remove(key.getName());
        } else {
            dataKeyObjectMap.put(key.getName(), entry);
        }
    }


    @Override
    @NotNull
    public Dimension getPreferredImageSize() {
        return preferredImageSize;
    }

    @Override
    public void setPreferredImageSize(@NotNull Dimension preferredImageSize) {
        if (preferredImageSize.equals(this.preferredImageSize)) {
            return;
        }
        Dimension oldValue = this.preferredImageSize;
        this.preferredImageSize = preferredImageSize;
        propertyChangeSupport.firePropertyChange("preferredImageSize", oldValue, preferredImageSize);
    }
}
