package plg.overlay;/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

import com.intellij.openapi.Disposable;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.components.ServiceManager;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.EditorFactory;
import com.intellij.openapi.editor.event.EditorFactoryAdapter;
import com.intellij.openapi.editor.event.EditorFactoryEvent;
import com.intellij.openapi.project.Project;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.util.containers.ConcurrentHashSet;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.awt.*;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by noxin on 28.07.14.
 */
public class SketchlinkOverlayManager implements ProjectComponent, Disposable {
    @NotNull
    private final Project project;
    private final Map<PsiFile, Set<OverlayContainer>> registeredContainer = new ConcurrentHashMap<>(30);

    public SketchlinkOverlayManager(@NotNull final Project project) {
        this.project = project;
        EditorFactory.getInstance().addEditorFactoryListener(new EditorFactoryAdapter() {
            @Override
            public void editorCreated(@NotNull EditorFactoryEvent event) {
                @NotNull Editor editor = event.getEditor();
                if (editor.isViewer() || editor.getProject() != project) {
                    return;
                }
                registerOverlay(editor);
            }

            @Override
            public void editorReleased(@NotNull EditorFactoryEvent event) {
                if (event.getEditor().getProject() != project) {
                    return;
                }
                unregisterOverlay(event.getEditor());
            }
        }, this);
    }

    public static SketchlinkOverlayManager getInstance(@NotNull Project project) {
        return ServiceManager.getService(project, SketchlinkOverlayManager.class);
    }

    public void initComponent() {
        // TODO: insert component initialization logic here
    }

    public void disposeComponent() {
        // TODO: insert component disposal logic here
    }

    @NotNull
    public String getComponentName() {
        return "plg.overlay.SketchlinkOverlayManager";
    }

    public void projectOpened() {
        // called when project is opened
    }

    public void projectClosed() {
        registeredContainer.clear();
    }

    private void registerOverlay(@NotNull Editor editor) {
        //only register for editors with files
        //vcs push entry for comments is an editor, but without a psifile
        @Nullable PsiFile file = PsiDocumentManager.getInstance(project).getPsiFile(editor.getDocument());
        if (file == null) {
            return;
        }
        @Nullable OverlayContainerImpl overlay = null;
        for (Component component : editor.getContentComponent().getComponents()) {
            if (component instanceof OverlayContainerImpl) {
                overlay = (OverlayContainerImpl) component;
                break;
            }
        }
        if (overlay == null) {
            overlay = new OverlayContainerImpl(editor);
        }
        //copy all OverlayElements
        Set<OverlayContainer> overlayContainers = registeredContainer.get(file);

        if (overlayContainers != null && overlayContainers.size() > 0) {
            //file already exists
            @NotNull Iterator<OverlayContainer> iterator = overlayContainers.iterator();
            OverlayContainer next = iterator.next();
            if (next != null) {
                for (@NotNull OverlayElement overlayElement : next.getOverlayElements()) {
                    overlay.addOverlayElement(overlayElement);
                }
            }
            overlayContainers.add(overlay);
        } else {
            //file is new
            overlayContainers = new ConcurrentHashSet<>(3);
            overlayContainers.add(overlay);
            registeredContainer.put(file, overlayContainers);
        }
        //editor.getContentComponent().add(overlay);
        //overlay.setVisible(true);
    }

    /**
     * Tries to add a new Item to the Overlay.
     *
     * @param element the plg.overlay.OverlayElementImpl to add
     * @return true if successful, false if there is no editor for this file opened
     */
    public boolean addItem(@NotNull OverlayElement element) {
        @Nullable PsiElement positionalElement = element.getPositionalElement();
        if (positionalElement == null || !positionalElement.isValid()) {
            return false;
        }
        PsiFile file = positionalElement.getContainingFile();
        if (!registeredContainer.containsKey(file)) {
            registeredContainer.put(file, new HashSet<OverlayContainer>());
        }
        Set<OverlayContainer> overlayContainers = registeredContainer.get(file);

        boolean added = false;

        for (@NotNull OverlayContainer container : overlayContainers) {
            container.addOverlayElement(element);
            added = true;
        }
        return added;
    }

    public void removeItem(@NotNull OverlayElement element, @NotNull PsiFile file) {
        for (@NotNull OverlayContainer overlayContainer : registeredContainer.get(file)) {
            overlayContainer.removeOverlayElement(element);
        }
        if (element.isValid()) {
            element.setValid(false);
        }
    }


    private void unregisterOverlay(@NotNull Editor editor) {
        @Nullable PsiFile file = PsiDocumentManager.getInstance(project).getPsiFile(editor.getDocument());
        if (file == null) {
            return;
        }
        @NotNull java.util.List<OverlayContainerImpl> componentList = new ArrayList<>();
        for (Component component : editor.getContentComponent().getComponents()) {
            if (component instanceof OverlayContainerImpl) {
                componentList.add((OverlayContainerImpl) component);
            }
        }
        Set<OverlayContainer> overlayContainers = registeredContainer.get(file);
        if (overlayContainers != null) {
            overlayContainers.removeAll(componentList);
        }
        for (@NotNull OverlayContainerImpl overlayContainer : componentList) {
            editor.getContentComponent().remove(overlayContainer);

            overlayContainer.dispose();
        }


    }

    @Override
    public void dispose() {
        registeredContainer.clear();
    }

}
