/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg;

import com.intellij.ide.browsers.BrowserFamily;
import com.intellij.ide.browsers.WebBrowser;
import com.intellij.ide.browsers.WebBrowserManager;
import com.intellij.openapi.application.ApplicationInfo;
import com.intellij.util.xmlb.annotations.Transient;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.ui.SketchlinkColors;

import java.awt.*;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by Peter on 10.11.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class SketchlinkState {

    private String serverURL;
    private boolean connectOnStartup;
    private boolean allowRemoteNavigation;
    private String clientName;
    private boolean showFindForNavigationAtTop;
    private Map<SketchlinkColors, String> colorMap = new ConcurrentHashMap<>();
    private boolean serverLoggingEnabled;

    private boolean sketchAnchorAlwaysFolded;
    private boolean sketchAnchorsProtected;
    private boolean sketchAnchorsInMethodsUseFoldingStyle;
    private int previewPopupTime;
    private int foldingStyle;
    private String defaultBrowser;


    public SketchlinkState() {
        loadDefaultValues();
    }

    public SketchlinkState(@NotNull SketchlinkState state) {
        copyFrom(state);
    }

    public void copyFrom(@NotNull SketchlinkState state) {
        this.serverURL = state.serverURL;
        this.connectOnStartup = state.connectOnStartup;
        allowRemoteNavigation = state.allowRemoteNavigation;
        showFindForNavigationAtTop = state.showFindForNavigationAtTop;
        serverLoggingEnabled = state.serverLoggingEnabled;
        sketchAnchorAlwaysFolded = state.sketchAnchorAlwaysFolded;
        sketchAnchorsProtected = state.sketchAnchorsProtected;
        sketchAnchorsInMethodsUseFoldingStyle = state.sketchAnchorsInMethodsUseFoldingStyle;
        previewPopupTime = state.previewPopupTime;
        foldingStyle = state.foldingStyle; //postfix, icon after @sketchlink
        colorMap.putAll(state.colorMap);
        clientName = state.clientName;
        defaultBrowser = state.defaultBrowser;
    }

    public void loadDefaultValues() {
        serverURL = "";
        connectOnStartup = false;
        allowRemoteNavigation = true;
        showFindForNavigationAtTop = true;
        serverLoggingEnabled = false;
        sketchAnchorAlwaysFolded = false;
        sketchAnchorsProtected = false;
        sketchAnchorsInMethodsUseFoldingStyle = false;
        previewPopupTime = 300;
        foldingStyle = 2; //postfix, icon after @sketchlink
        colorMap.put(SketchlinkColors.PREVIEW_HIGHLIGHTED_STROKE, "#000000FF");
        colorMap.put(SketchlinkColors.PREVIEW_HIGHLIGHTED_FILL, "#00E61058");
        colorMap.put(SketchlinkColors.PREVIEW_LINKED_STROKE, "#5E5E5EFF");
        colorMap.put(SketchlinkColors.PREVIEW_LINKED_FILL, "#C6FFB35A");
        colorMap.put(SketchlinkColors.PREVIEW_NORMAL_STROKE, "#5E5E5EFF");
        colorMap.put(SketchlinkColors.PREVIEW_NORMAL_FILL, "#CCCCCC00");
        colorMap.put(SketchlinkColors.OVERLAY_TARGET, "#C6FFB3FF");
        colorMap.put(SketchlinkColors.OVERLAY_TARGET_STROKE, "#CCCCCC00");
        colorMap.put(SketchlinkColors.PREVIEW_CODE_LINKED_BACKGROUND, "#ccccccff");
        colorMap.put(SketchlinkColors.PREVIEW_CODE_LINKED_STROKE, "#ccccccff");
        colorMap.put(SketchlinkColors.PREVIEW_CODE_AVAILABLE_BACKGROUND, "#ccccccff");
        colorMap.put(SketchlinkColors.PREVIEW_CODE_AVAILABLE_STROKE, "#ccccccff");
        clientName = "IntelliJ " + ApplicationInfo.getInstance().getFullVersion();
        defaultBrowser = WebBrowserManager.getInstance().getFirstBrowser(BrowserFamily.FIREFOX).getId().toString();
    }

    @Nullable
    public String getBrowserID() {
        return defaultBrowser;
    }

    public void setBrowserID(@NotNull String currentBrowserID) {
        //System.out.println("setBrowserID " + currentBrowserID);
        defaultBrowser = currentBrowserID;
    }

    @Nullable
    @Transient
    public WebBrowser getBrowser() {
        String browserID = getBrowserID();
        if (browserID == null)
            return null;
        return WebBrowserManager.getInstance().findBrowserById(browserID);
    }

    @NotNull
    public String getColorHTMLRGBA(@NotNull SketchlinkColors type) {
        return colorMap.get(type);
    }

    @NotNull
    public Color getColor(@NotNull SketchlinkColors type) {
        String colorStr = colorMap.get(type);
        return new Color(
                Integer.valueOf(colorStr.substring(1, 3), 16),
                Integer.valueOf(colorStr.substring(3, 5), 16),
                Integer.valueOf(colorStr.substring(5, 7), 16),
                Integer.valueOf(colorStr.substring(7, 9), 16));
    }

    public void setColor(@NotNull SketchlinkColors type, @NotNull Color newColor) {
        @NotNull String color = "#"
                + String.format("%02X", newColor.getRed())
                + String.format("%02X", newColor.getGreen())
                + String.format("%02X", newColor.getBlue())
                + String.format("%02X", newColor.getAlpha());
        colorMap.put(type, color);
    }

    public String getServerURL() {
//        System.out.println("getServerURL "+serverURL+" "+this.getClass().getName());
        return serverURL;
    }

    public void setServerURL(String serverURL) {
        this.serverURL = serverURL;
    }

    public boolean isConnectOnStartup() {
        return connectOnStartup;
    }


    public void setConnectOnStartup(boolean connectOnStartup) {
        this.connectOnStartup = connectOnStartup;
    }

    public boolean isAllowRemoteNavigation() {
        return allowRemoteNavigation;
    }

    public void setAllowRemoteNavigation(boolean allowRemoteNavigation) {
        this.allowRemoteNavigation = allowRemoteNavigation;
    }

    public String getClientName() {
        return clientName;
    }

    public void setClientName(String clientName) {
        this.clientName = clientName;
    }

    public boolean isShowFindForNavigationAtTop() {
        return showFindForNavigationAtTop;
    }

    public void setShowFindForNavigationAtTop(boolean showFindForNavigationAtTop) {
        this.showFindForNavigationAtTop = showFindForNavigationAtTop;
    }

    public boolean isServerLoggingEnabled() {
        return serverLoggingEnabled;
    }

    public void setServerLoggingEnabled(boolean serverLoggingEnabled) {
        this.serverLoggingEnabled = serverLoggingEnabled;
    }

    public boolean isSketchAnchorAlwaysFolded() {
        return sketchAnchorAlwaysFolded;
    }

    public void setSketchAnchorAlwaysFolded(boolean sketchAnchorAlwaysFolded) {
        this.sketchAnchorAlwaysFolded = sketchAnchorAlwaysFolded;
    }

    public boolean isSketchAnchorsProtected() {
        return sketchAnchorsProtected;
    }

    public void setSketchAnchorsProtected(boolean sketchAnchorsProtected) {
        this.sketchAnchorsProtected = sketchAnchorsProtected;
    }

    public boolean isSketchAnchorsInMethodsUseFoldingStyle() {
        return sketchAnchorsInMethodsUseFoldingStyle;
    }

    public void setSketchAnchorsInMethodsUseFoldingStyle(boolean sketchAnchorsInMethodsUseFoldingStyle) {
        this.sketchAnchorsInMethodsUseFoldingStyle = sketchAnchorsInMethodsUseFoldingStyle;
    }

    public int getPreviewPopupTime() {
        return previewPopupTime;
    }

    public void setPreviewPopupTime(int previewPopupTime) {
        this.previewPopupTime = previewPopupTime;
    }

    public int getFoldingStyle() {
        return foldingStyle;
    }

    public void setFoldingStyle(int foldingStyle) {
        this.foldingStyle = foldingStyle;
    }

}
