/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.actions.quicktesting;

import com.intellij.openapi.Disposable;
import com.intellij.openapi.actionSystem.AnAction;
import com.intellij.openapi.actionSystem.AnActionEvent;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.EditorFactory;
import com.intellij.openapi.editor.LogicalPosition;
import com.intellij.openapi.editor.event.EditorFactoryAdapter;
import com.intellij.openapi.editor.event.EditorFactoryEvent;
import com.intellij.openapi.editor.event.EditorFactoryListener;
import com.intellij.openapi.editor.markup.HighlighterLayer;
import com.intellij.openapi.editor.markup.HighlighterTargetArea;
import com.intellij.openapi.editor.markup.RangeHighlighter;
import com.intellij.openapi.editor.markup.TextAttributes;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.IconLoader;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.PsiWhiteSpace;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.util.PsiUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.SketchlinkManager;
import plg.slCurrentProjectSettings;
import plg.ui.SketchlinkColors;
import plg.websocket.slWebsocketManager;

import javax.swing.*;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by Peter on 19.10.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class CreateNewCodeAnchorDnDAction extends AnAction implements Disposable {
    private static final Icon myEnabledIcon = IconLoader.getIcon("/plg/ui/icons/createAnchor_enabled3.png");
    private static final Icon myDisAbledIcon = IconLoader.getIcon("/plg/ui/icons/createAnchor_disabled.png");
    private final EditorFactoryListener factoryListener;
    private Map<Editor, MyMouseListener> registeredListeners = new ConcurrentHashMap<>();
    private Map<Project, Boolean> active = new ConcurrentHashMap<>();

    public CreateNewCodeAnchorDnDAction() {
        factoryListener = new EditorFactoryAdapter() {
            @Override
            public void editorCreated(@NotNull EditorFactoryEvent event) {
                addMouseListener(event.getEditor());
            }

            @Override
            public void editorReleased(@NotNull EditorFactoryEvent event) {
                removeMouseListener(event.getEditor());
            }
        };
        EditorFactory.getInstance().addEditorFactoryListener(factoryListener, this);
    }

    @Override
    public void update(AnActionEvent e) {
        Project project = e.getProject();
        if (project == null) {
            return;
        }
        if (active.get(project) == null) {
            active.put(project, false);
            return;
        }
        if (active.get(project)) {
            e.getPresentation().setIcon(myEnabledIcon);
        } else {
            e.getPresentation().setIcon(myDisAbledIcon);
        }

    }

    private void addMouseListener(@NotNull Editor editor) {
        Project project = editor.getProject();
        if (project == null) {
            return;
        }
        if (active.get(project) == null) {
            active.put(project, false);
            return;
        }
        final PsiFile psiFile = PsiDocumentManager.getInstance(project).getPsiFile(editor.getDocument());
        if (!active.get(project) || psiFile == null) {
            return;
        }
        MyMouseListener mouseListener = new MyMouseListener(editor, project, psiFile);
        registeredListeners.put(editor, mouseListener);
        editor.getContentComponent().addMouseListener(mouseListener);
        editor.getContentComponent().addMouseMotionListener(mouseListener);

    }

    private void removeMouseListener(@NotNull Editor editor) {
        MyMouseListener mouseListener = registeredListeners.get(editor);
        if (mouseListener == null) {
            return;
        }
        mouseListener.unMark();
        editor.getContentComponent().removeMouseMotionListener(mouseListener);
        editor.getContentComponent().removeMouseListener(mouseListener);
        registeredListeners.remove(editor);
    }

    private void checkAllCurrentEditors(@NotNull Project project, boolean enable) {
        Editor[] allEditors = EditorFactory.getInstance().getAllEditors();
        for (Editor allEditor : allEditors) {
            if (project.equals(allEditor.getProject())) {
                if (enable) {
                    addMouseListener(allEditor);
                } else {
                    removeMouseListener(allEditor);
                }
            }
        }
    }


    public void actionPerformed(AnActionEvent e) {
        Project project = e.getProject();
        if (project == null) {
            return;
        }
        boolean enable;
        if (active.get(project) == null || !active.get(project)) {
            enable = true;
            e.getPresentation().setIcon(myEnabledIcon);
        } else {
            enable = false;
            e.getPresentation().setIcon(myDisAbledIcon);
        }
        active.put(project, enable);
        checkAllCurrentEditors(project, enable);
        slWebsocketManager.getInstance(project).setUpdateServerOnCreate(!enable);
    }

    @Override
    public void dispose() {
        for (Project project : active.keySet()) {
            checkAllCurrentEditors(project, false);
        }
        active.clear();
        registeredListeners.clear();
    }


    private class MyMouseListener implements MouseMotionListener, MouseListener {
        private final Editor editor;
        private final PsiFile file;
        private final Project project;
        private PsiElement currentLinkElement;
        private RangeHighlighter currentRangeHightlighter;

        public MyMouseListener(@NotNull Editor editor, @NotNull Project project, @NotNull PsiFile file) {
            this.editor = editor;
            this.project = project;
            this.file = file;
        }

        @Override
        public void mouseClicked(MouseEvent e) {
            Point point = e.getPoint();
            if (e.getButton() != MouseEvent.BUTTON1) {
                return;
            }
            if (currentLinkElement == null)
                return;
            @Nullable String sketchlink = SketchlinkManager.getInstance(project).createSketchlink(currentLinkElement);
            unMark();

        }

        private void unMark() {
            if (currentRangeHightlighter != null) {
                editor.getMarkupModel().removeHighlighter(currentRangeHightlighter);
                currentRangeHightlighter = null;
                currentLinkElement = null;
            }
        }

        private void mark(@Nullable PsiElement element) {
            if (element == null) {
                unMark();
                return;
            }
            Color color = slCurrentProjectSettings.getInstance(project).getColor(SketchlinkColors.OVERLAY_TARGET);

            TextAttributes attributes = new TextAttributes(null, color, null, null, Font.BOLD);

            currentRangeHightlighter = editor.getMarkupModel().addRangeHighlighter(element.getTextRange().getStartOffset(),
                    element.getTextRange().getEndOffset(), HighlighterLayer.LAST, attributes, HighlighterTargetArea.EXACT_RANGE);
            currentLinkElement = element;
        }

        @Override
        public void mouseMoved(MouseEvent e) {
            Point point = e.getPoint();
            LogicalPosition logicalPosition = editor.xyToLogicalPosition(point);
            int offset = editor.logicalPositionToOffset(logicalPosition);

            @NotNull PsiElement element = PsiUtil.getElementAtOffset(file, offset);

            if (element instanceof PsiWhiteSpace) {
                int linestartoffset =
                        editor.getDocument().getLineStartOffset(editor.getDocument().getLineNumber(element.getTextRange().getEndOffset()));
                @Nullable PsiElement sibling;
                if (linestartoffset <= offset) {
                    sibling = PsiTreeUtil.nextLeaf(element, true);
                } else {
                    sibling = PsiTreeUtil.prevLeaf(element, true);
                }
                if (sibling != null) {
                    element = sibling;
                }
            }
            @NotNull PsiElement topmostLinkableElement = SketchlinkManager.TOOLS.getTopmostLinkableElement(element);

            if (topmostLinkableElement.equals(currentLinkElement)) {
                return;
            }
//            System.out.println("topLink: "+topmostLinkableElement+"\t"+element+" Already Linked: "+SketchlinkManager.TOOLS.hasSketchlink(topmostLinkableElement));
            unMark();
            if (SketchlinkManager.TOOLS.hasSketchlink(topmostLinkableElement)) {
//                System.out.println("Has Sketchlink");
                return;
            }
            mark(topmostLinkableElement);

        }

        @Override
        public void mouseEntered(MouseEvent e) {

        }

        @Override
        public void mouseExited(MouseEvent e) {

        }

        @Override
        public void mousePressed(MouseEvent e) {

        }

        @Override
        public void mouseReleased(MouseEvent e) {

        }


        @Override
        public void mouseDragged(MouseEvent e) {

        }


    }

}
