/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.sketchlinks;

import com.intellij.openapi.editor.Document;
import com.intellij.openapi.util.TextRange;
import com.intellij.psi.PsiElement;
import com.intellij.psi.javadoc.PsiDocTag;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.overlay.OverlayElement;

/**
 * Created by Peter on 09.03.14.
 * <p/>
 * Class describing a Sketch Anchor which can be found in the currently active Project and navigated to directly inside
 * the IDE
 */
public interface LocalLinkAnchor extends LinkAnchor, LinkAnchorParentOwner, MetadataOwner<SketchMetadataTypeC>, PropertyChangeSupported, Comparable<LocalLinkAnchor> {
    /**
     * Returns the Sketchlink-Anchor defining this LocalLinkAnchor.
     *
     * @return the PsiDocTag containing the <code>@sketchlink {id}</code>
     * null if the PsiDocTag has been deleted
     */
    @Nullable
    public PsiDocTag getAnchorTag();

    /**
     * Returns the TextRange of the DocTag, does not include parts after the linebreak.
     * If the docTag is invalid, the Textrange is from 0 to 0
     */
    @NotNull
    public TextRange getAnchorRange();

    /**
     * Returns the TextRange of the target Element(s).
     * If the docTag is invalid, the Textrange is from 0 to 0
     */
    @NotNull
    public TextRange getTargetRange();

    /**
     * Gets the document this anchor is part of.
     * The Document is unique for all running instances of IntelliJ and serves as the basis
     * for the PSI tree in this project.
     */
    @Nullable
    public Document getDocument();

    /**
     * Gets the MD5 part that is saved in the PsiDocTag value/comment section in the file.
     * It is the md5 checksum of the path to the file (accessible via the current metadata)
     */
    @NotNull
    public String getSavedMD5();

    public void updateMD5();

    /**
     * Changes the PsiDocTag that represents the anchor in the PSI tree.
     * The PsiDocTag is only saved on a best-effort basis using SmartPsiElement and might accidentally
     * change after inputting new source code.
     */
    public void updateAnchorTag(@NotNull PsiDocTag newAnchor);

    /**
     * Returns the previously calculated Target element this Sketchlink references or tries to calculate the new
     * Position if the old doesn't exist anymore. Call isTargetUnchanged(false) to check for existence without
     * updating the target.
     * <p/>
     * Always returns a target unless getAnchorTag returns null, too.
     *
     * @return the PsiIdentifier for classes,methods and field or the beginning of the statement otherwise
     * null if the target has been deleted and not been set to a new location
     */
    @Nullable
    public PsiElement getTargetElement();

    /**
     * Rechecks the currently (calculated) target Element with the saved target element.
     * If previous target element is null, the effect will be the same als using parameter <code>true</code>
     * No new position will be calculated, if <code>getAnchorTag</code> is null.<p/>
     * <img src="http://www.spiegel.de/static/sys/v10/logo/spiegel_online_logo_460_64.png" />
     * <frame src="http://de.selfhtml.org/html/frames/eigenschaften.htm" name="Navigation" scrolling="no">
     *
     * @param update if true, sets the current target as the target element of this Entry
     *               if false, only checks vor identity between the current and saved target element
     * @return true if the calculation returns the saved element,
     * false if they differ or isValid would return false
     */
    public boolean isTargetUnchanged(boolean update);

    /**
     * Checks if the SketchlinkAnchor is not null and it's id matches the general location it's supposed to be in.
     */
    public boolean isValid();

    /**
     * Sets this as an invalid Entry, for example after a refactoring changed the SketchId
     */
    public void invalidate();


    public boolean isAnchorEqualTo(@NotNull PsiDocTag docTag);

    /**
     * Calculate the current metadata of the code anchor
     */
    @Nullable
    public SketchMetadataTypeC getCurrentMetaData();

//    public SketchMetadataTypeC getMetadata();
//
//    void setMetadata(@NotNull SketchMetadataTypeC metadata);

    @NotNull
    public OverlayElement getOverlayElement();

    @Nullable
    SketchMetadataTypeC isCurrentMetaDataUpdated();
}
