/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.sketchlinks;

import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.fileEditor.FileDocumentManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Computable;
import com.intellij.openapi.util.TextRange;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.*;
import com.intellij.psi.javadoc.PsiDocComment;
import com.intellij.psi.javadoc.PsiDocTag;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.overlay.OverlayElement;
import plg.tools.slGuiTools;
import plg.tools.slSearch;
import plg.tools.slToolBox;
import plg.ui.overlay.SketchlinkOverlayElementImpl;
import plg.ui.overlay.slOverlayElementBuilder;

import javax.swing.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.security.InvalidParameterException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Created by Peter on 22.03.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class LocalLinkAnchorImpl implements LocalLinkAnchor {
    private final static Logger LOG = Logger.getInstance(LocalLinkAnchorImpl.class);
    @NotNull
    private final Project project;
    private SmartPsiElementPointer<PsiDocTag> myDocTagAnchor;
    private SmartPsiElementPointer<PsiElement> myTargetElement;
    private LinkAnchor myLinkAnchor;
    @NotNull
    private String myMD5hash;
    private SketchMetadataTypeC myServerMetadata;
    private SketchlinkOverlayElementImpl overlayElement;
    //property support
    @NotNull
    private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);

    public LocalLinkAnchorImpl(@NotNull PsiDocTag anchor, @NotNull PsiElement referencedElement) {
        @Nullable String id = SketchlinkManager.TOOLS.getSketchlinkTagID(anchor);
        if (id == null) {
            throw new InvalidParameterException("Tried creating a LocalLinkAnchor without a valid PsiDocTag " + anchor.getText());
        }
        myLinkAnchor = new LinkAnchorImpl(id, anchor.getProject());
        updateAnchorTag(anchor);
        setTargetElement(referencedElement);
        String hash = SketchlinkManager.TOOLS.getSketchlinkTagMD5(anchor);
        myMD5hash = hash == null ? "0" : hash;
        this.project = anchor.getProject();
        overlayElement = new slOverlayElementBuilder(this, anchor).build();

    }

    @Override
    @NotNull
    public Project getProject() {
        return project;
    }

    @Override
    @Nullable
    public PsiDocTag getAnchorTag() {
        PsiDocTag element = myDocTagAnchor.getElement();
        if (element == null)
            return null;
        //detect when the "best effort" SmartPsiElement switched targets and only return a valid anchor  tag with the same ID
        String sketchlinkTagID = SketchlinkManager.TOOLS.getSketchlinkTagID(element);
        if (sketchlinkTagID == null) {
            PsiDocComment containingComment = element.getContainingComment();
            if (containingComment == null)
                return null;
            for (PsiDocTag docTag : SketchlinkManager.TOOLS.getSketchlinkTags(containingComment)) {
                sketchlinkTagID = SketchlinkManager.TOOLS.getSketchlinkTagID(docTag);
                if (getAnchorId().equals(sketchlinkTagID)) {
                    updateAnchorTag(docTag);
                    return docTag;
                }
            }
            return null;
        }
        if (!sketchlinkTagID.equals(getAnchorId())) {
            return null;
        }
        return element;
    }

    @NotNull
    @Override
    public TextRange getAnchorRange() {
        @Nullable PsiDocTag anchorTag = getAnchorTag();
        if (anchorTag == null) {
            return new TextRange(0, 0);
        }
        @NotNull String[] split = anchorTag.getText().split("\n");
        if (split.length < 2) {
            return anchorTag.getTextRange();
        }
        return new TextRange(anchorTag.getTextRange().getStartOffset(), anchorTag.getTextRange().getStartOffset() + split[0].length());
    }

    @NotNull
    @Override
    public TextRange getTargetRange() {
        @Nullable PsiElement targetElement = getTargetElement();
        if (targetElement == null) {
            return new TextRange(0, 0);
        }
        //only highlight the identifier for classes, methods, field and packages.
        if (targetElement instanceof PsiNameIdentifierOwner) {
            @Nullable PsiElement nameIdentifier = ((PsiNameIdentifierOwner) targetElement).getNameIdentifier();
            if (nameIdentifier != null) {
                return nameIdentifier.getTextRange();
            }
        }
        return targetElement.getTextRange();
    }


    @Nullable
    private PsiFile getPsiFile() {
        @Nullable PsiDocTag anchorTag = getAnchorTag();
        if (anchorTag != null && anchorTag.isValid()) {
            return anchorTag.getContainingFile();
        }
        @Nullable PsiElement targetElement = getTargetElement();
        if (targetElement != null && targetElement.isValid()) {
            return targetElement.getContainingFile();
        }
        return null;
    }

    @Nullable
    @Override
    public Document getDocument() {
        @Nullable PsiFile psiFile = getPsiFile();
        if (psiFile == null) {
            return null;
        }
        return FileDocumentManager.getInstance().getDocument(psiFile.getVirtualFile());
    }

    @NotNull
    @Override
    public String getSavedMD5() {
        return myMD5hash;
    }

    @Override
    public void updateMD5() {
        @Nullable PsiDocTag docTag = myDocTagAnchor.getElement();
        if (docTag != null && SketchlinkManager.TOOLS.isValidID(docTag)) {
            String hash = SketchlinkManager.TOOLS.getSketchlinkTagMD5(docTag);
            myMD5hash = hash != null ? hash : "0";
        }
    }

    @Override
    public void updateAnchorTag(@NotNull PsiDocTag newAnchor) {
        myDocTagAnchor = SmartPointerManager.getInstance(newAnchor.getProject()).createSmartPsiElementPointer(newAnchor, newAnchor.getContainingFile());
        if (overlayElement == null) {
            overlayElement = new slOverlayElementBuilder(this, newAnchor).build();
        } else {
            overlayElement.setPositionalElement(newAnchor);
        }
    }

    @Override
    @Nullable
    public PsiElement getTargetElement() {
        @Nullable PsiElement element = myTargetElement.getElement();
        if (element != null)
            return myTargetElement.getElement();
        @Nullable PsiDocTag docTag = myDocTagAnchor.getElement();
        if (docTag == null)
            return null;
        setTargetElement(SketchlinkManager.TOOLS.getReferencedPsiElement(docTag));
        return myTargetElement.getElement();
    }

    public void setTargetElement(@NotNull PsiElement referencedElement) {
        myTargetElement = SmartPointerManager.getInstance(referencedElement.getProject()).createSmartPsiElementPointer(referencedElement,
                referencedElement.getContainingFile());
    }

    @NotNull
    @Override
    public String getAnchorId() {
        return myLinkAnchor.getAnchorId();
    }

    @Override
    public void setAnchorId(@NotNull String id) {
        myLinkAnchor.setAnchorId(id);
    }

    @NotNull
    private String getTypeName(@NotNull PsiElement element) {
        String typeName = element.getClass().getSimpleName();
        if (typeName.startsWith("Psi"))
            typeName = typeName.substring(3);
        if (typeName.endsWith("Impl")) //NON-NLS
            typeName = typeName.substring(0, typeName.length() - 4);
        return typeName;
    }

    @NotNull
    @Override
    public String getDescription() {
        @Nullable PsiElement element = getTargetElement();
        if (element == null) {
            return "Invalid anchor: " + getAnchorId();
        }
        @NotNull String name = "";
        @Nullable PsiElement parentElement = element;
        while (!(parentElement instanceof PsiNamedElement) && (parentElement = parentElement.getParent()) != null) {
            //iteration is in while conditional element
            if (name.isEmpty()) {
                name = getTypeName(element) + " in ";
            }
        }
        if (parentElement != null) {
            name += getTypeName(parentElement) + " \"" + ((PsiNamedElement) parentElement).getName() + "\"";
        }
        return name;
    }


    @Override
    @NotNull
    public AnchorTypes getAnchorType() {
        return myLinkAnchor.getAnchorType();
    }

    @Override
    @NotNull
    public ImageIcon getImageIcon() {
        return myLinkAnchor.getImageIcon();
    }

    @Override
    public void navigateTo() {
        @Nullable PsiElement element = getTargetElement();
        if (element == null) {
            element = getAnchorTag();
            if (element == null) {
                slGuiTools.ShowNotificationAtBottom("Couldn't navigate to anchor " + myLinkAnchor.getAnchorId() + " because location is unknown.", project);
                return;
            }
        }
        slSearch.browseToLocal(element);
    }

    @Override
    public boolean isTargetUnchanged(boolean update) {
        @Nullable PsiDocTag docTag = myDocTagAnchor.getElement();
        if (docTag == null || !docTag.isValid()) {
            return false;
        }
        @Nullable PsiElement savedElement = myTargetElement.getElement();
        @NotNull PsiElement currentTarget = SketchlinkManager.TOOLS.getReferencedPsiElement(docTag);
        if (update || savedElement == null) {
            setTargetElement(currentTarget);
        }
        return currentTarget.isEquivalentTo(savedElement);
    }

    @Override
    public boolean isValid() {
        @Nullable PsiDocTag docTag = myDocTagAnchor.getElement();
        return docTag != null && docTag.isValid() && SketchlinkManager.TOOLS.isValidID(docTag);
    }

    @Override
    public void invalidate() {
        setAnchorId("");


    }

    @Override
    public boolean equals(@Nullable Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        @NotNull LocalLinkAnchorImpl that = (LocalLinkAnchorImpl) o;

        if (myDocTagAnchor.getElement() != null ? !myDocTagAnchor.equals(that.myDocTagAnchor) : that.myDocTagAnchor.getElement() != null)
            return false;

        return true;
    }

    @Override
    public int hashCode() {
        int i = ApplicationManager.getApplication().runReadAction(new Computable<Integer>() {
            @NotNull
            @Override
            public Integer compute() {
                return myDocTagAnchor.getElement() != null ? myDocTagAnchor.getElement().hashCode() : 0;
            }
        });
        return i;
//        return myDocTagAnchor.getElement() != null ? myDocTagAnchor.getElement().hashCode() : 0;
    }

    @Override
    public boolean isAnchorEqualTo(@NotNull PsiDocTag docTag) {
        @Nullable PsiDocTag myTag = getAnchorTag();
        @Nullable String id = SketchlinkManager.TOOLS.getSketchlinkTagID(docTag);
        if (myTag == null) {
            if (id == null) {
                return false;
            }
            if (id.equals(getAnchorId())) {
                updateAnchorTag(docTag);
                return true;
            }
        }
        return docTag.isEquivalentTo(myTag);
    }

    @Override
    @Nullable
    public SketchMetadataTypeC getCurrentMetaData() {
        @NotNull Computable<SketchMetadataTypeC> calculateResult = new Computable<SketchMetadataTypeC>() {
            @Nullable
            @Override
            public SketchMetadataTypeC compute() {
                @Nullable PsiElement target = getTargetElement();
                if (!isValid() || target == null) {
                    return null;
                }
                @NotNull SketchMetadataTypeC result = new SketchMetadataTypeC();

                //TYPE: convert the Implementation name to the type name
                result.setTargetType(target.getClass().getSimpleName());
                if (result.getTargetType().isEmpty()) {
                    result.setTargetType("unknown");
                }
                int start = result.getTargetType().startsWith("Psi") ? 3 : 0;
                int end = result.getTargetType().endsWith("Impl") ? result.getTargetType().length() - 4 : result.getTargetType().length();
                end = result.getTargetType().endsWith("PackageStatementImpl") ? end - 9 : end;
                result.setTargetType(result.getTargetType().substring(start, end));
                //NAME: try to find the name of this element, if available
                @Nullable String name = null;
                if (target instanceof PsiNameIdentifierOwner) {
                    name = ((PsiNameIdentifierOwner) target).getName();
                } else {
                    if (target instanceof PsiDeclarationStatement && target.getFirstChild() instanceof PsiLocalVariable) {
                        name = ((PsiLocalVariable) target.getFirstChild()).getName();
                    }
                }
                result.setTargetName(name);
                //PATH: insert the path to the file containing this element
                @Nullable String relativePath = slToolBox.getRelativePath(target);
                result.setPath(relativePath != null ? relativePath : "");
                //PARENT: traverse the tree up to find a possible parent
                @Nullable LinkAnchor parent = getParent();
                if (parent != null) {
                    result.setParentID(parent.getAnchorId());
                } else {
                    result.setParentID(null);
                }

                //PROJECTNAME
                result.setProjectName(target.getProject().getName());
                return result;
            }
        };
        return ApplicationManager.getApplication().runReadAction(calculateResult);
    }

    @Override
    public SketchMetadataTypeC getMetadata() {
        return myServerMetadata;
    }

    @Override
    public void setMetadata(@NotNull SketchMetadataTypeC metadata) {
        if (metadata.equals(this.myServerMetadata))
            return;
        SketchMetadataTypeC oldValue = this.myServerMetadata;
        this.myServerMetadata = metadata;
        propertyChangeSupport.firePropertyChange(SketchMetadata.METADATA_Property, oldValue, this.myServerMetadata);
    }

    @NotNull
    @Override
    public OverlayElement getOverlayElement() {
        return overlayElement;
    }

    @Override
    @Nullable
    public SketchMetadataTypeC isCurrentMetaDataUpdated() {
        @Nullable SketchMetadataTypeC current = getCurrentMetaData();
        if (current != null) {
            if (current.equals(myServerMetadata)) {
                return null;
            } else {
                return current;
            }
        }
        return null;
    }

    @Nullable
    @Override
    public LinkAnchor getParent() {
        @Nullable PsiElement myTargetElementElement = myTargetElement.getElement();
        if (myTargetElementElement == null || !myTargetElementElement.isValid()) {
            return null;
        }
        PsiElement parent = myTargetElementElement.getParent();
        while (!((parent == null) || parent instanceof PsiDocCommentOwner)) {
            parent = parent.getParent();
        }

        if (parent != null) {
            @Nullable PsiDocComment docComment = ((PsiDocCommentOwner) parent).getDocComment();
            if (docComment != null) {
                @Nullable PsiDocTag docTag = SketchlinkManager.TOOLS.getFirstSketchlinkTag(docComment);

                if (docTag != null) {
                    return SketchlinkManager.getAnchorOf(docTag);

                }
            }
        }
        return null;
    }


//        @Override
//        public boolean equals(Object o) {
//            return o instanceof LocalLinkAnchorImpl && (super.equals(o) || Objects.equals(myDocTagAnchor.getElement(), ((LocalLinkAnchorImpl) o).myDocTagAnchor.getElement()));
//        }
//
//        @Override
//        public int hashCode() {
//            return Objects.hash(myDocTagAnchor.getElement());
//        }

    @Override
    public void addPropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener(listener);
    }

    @Override
    public void removePropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener(listener);
    }

    @NotNull
    @Override
    public List<PropertyChangeListener> getPropertyChangeListeners() {
        //seems redundant, but Arrays.asList returns an immutable list throwing errors when removing items manually
        return new ArrayList<>(Arrays.asList(propertyChangeSupport.getPropertyChangeListeners()));
    }

    /**
     * Sort by containing file name. If equal, sort by offset in file.
     * If on element is invalid, sort using the id as fallback.
     */
    @Override
    public int compareTo(@NotNull LocalLinkAnchor o) {
        if (this.equals(o)) {
            return 0;
        }
        Document thereDocument = o.getDocument();
        Document hereDocument = getDocument();
        if (thereDocument == null || hereDocument == null) {
            return getAnchorId().compareTo(o.getAnchorId());
        }
        VirtualFile file = FileDocumentManager.getInstance().getFile(hereDocument);
        VirtualFile file1 = FileDocumentManager.getInstance().getFile(thereDocument);
        String hereString = file != null ? file.getCanonicalPath() : "";
        String thereString = file1 != null ? file1.getCanonicalPath() : "";
        if (hereString == null || thereString == null) {
            return getAnchorId().compareTo(o.getAnchorId());
        }
        if (hereString.equals(thereString)) {
            TextRange here = getAnchorRange();
            TextRange there = o.getAnchorRange();
            return here.getStartOffset() - there.getStartOffset();
        }
        return hereString.compareTo(thereString);
    }
}
