/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.sketchlinks;

import com.intellij.openapi.editor.Document;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.IconLoader;
import com.intellij.openapi.util.TextRange;
import com.intellij.psi.PsiElement;
import com.intellij.psi.javadoc.PsiDocTag;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.overlay.OverlayElement;
import plg.overlay.SketchlinkOverlayManager;
import plg.tools.slToolBox;
import plg.ui.overlay.SketchlinkOverlayElementImpl;
import plg.ui.overlay.slOverlayElementBuilder;

import javax.swing.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.*;

/**
 * Created by Peter on 15.06.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class LocalLinkAnchorRange implements LocalLinkAnchor {
    private static final ImageIcon myIcon = new ImageIcon(slToolBox.iconToImage(IconLoader.getIcon("/plg/ui/icons/iconRangeAnchor_left2.png")));
    private static final ImageIcon myIconRight = new ImageIcon(slToolBox.iconToImage(IconLoader.getIcon("/plg/ui/icons/iconRangeAnchor_right2.png")));

    @NotNull
    private LocalLinkAnchor startAnchor;
    @Nullable
    private LocalLinkAnchor endAnchor;
    private SketchlinkOverlayElementImpl overlayElement_left;
    private SketchlinkOverlayElementImpl overlayElement_right;
    //property support
    @NotNull
    private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);

    public LocalLinkAnchorRange(@NotNull LocalLinkAnchor startAnchor, @NotNull LocalLinkAnchor endAnchor) {
        Set<PropertyChangeListener> listeners = new HashSet<>(startAnchor.getPropertyChangeListeners());
        listeners.addAll(endAnchor.getPropertyChangeListeners());
        for (PropertyChangeListener listener : listeners) {
            propertyChangeSupport.addPropertyChangeListener(listener);
            startAnchor.removePropertyChangeListener(listener);
            endAnchor.removePropertyChangeListener(listener);
        }
        boolean keepOrder = startAnchor.getAnchorRange().getStartOffset() <= endAnchor.getAnchorRange().getEndOffset();
        this.startAnchor = keepOrder ? startAnchor : endAnchor;
        this.endAnchor = couldMatch(keepOrder ? endAnchor : startAnchor);
        @Nullable PsiDocTag anchorTag = startAnchor.getAnchorTag();
        @Nullable PsiDocTag endAnchorTag = endAnchor.getAnchorTag();
        if (anchorTag != null && endAnchorTag != null) {
            overlayElement_left = new slOverlayElementBuilder(this, anchorTag).build();
            overlayElement_right = new slOverlayElementBuilder(this, endAnchorTag, myIconRight, endAnchor).build();
            overlayElement_right.setImageIcon(myIconRight);
            SketchlinkOverlayManager.getInstance(getProject()).addItem(overlayElement_right);
        } else {
            System.out.println("Error: LocalLinkAnchorRange was created with invalid startAnchor! " + startAnchor.getAnchorId());
        }
    }


    @Nullable
    private LocalLinkAnchor couldMatch(@NotNull LocalLinkAnchor endAnchor) {
        @Nullable Document document = startAnchor.getDocument();
        if (document == null) {
            return null;
        }
        return (document.equals(endAnchor.getDocument()) ? endAnchor : null);
    }

    public void addAnchor(@NotNull LocalLinkAnchor anchor) {
        if (!startAnchor.isValid()) {
            startAnchor = anchor;
            return;
        }
        @Nullable LocalLinkAnchor sideAnchor = couldMatch(anchor);
        if (sideAnchor == null || !anchor.isValid()) {
            return;
        }
        if (endAnchor == null || endAnchor.getAnchorRange().getStartOffset() < anchor.getAnchorRange().getStartOffset()) {
            endAnchor = anchor;
            overlayElement_right.setAnchorForPosition(endAnchor);
        }
        if (startAnchor.getAnchorRange().getStartOffset() > anchor.getAnchorRange().getStartOffset()) {
            startAnchor = anchor;
            overlayElement_left.setAnchorForPosition(startAnchor);
        }
    }

    @Override
    @Nullable
    public PsiDocTag getAnchorTag() {
        return startAnchor.getAnchorTag();
    }

    @Override
    @NotNull
    public TextRange getAnchorRange() {
        return startAnchor.getAnchorRange();
    }

    @Override
    @NotNull
    public TextRange getTargetRange() {
        if (endAnchor != null) {
            return new TextRange(startAnchor.getTargetRange().getStartOffset(), endAnchor.getTargetRange().getEndOffset());
        }
        return startAnchor.getTargetRange();
    }

    @Override
    @Nullable
    public Document getDocument() {
        return startAnchor.getDocument();
    }

    @Override
    @NotNull
    public String getSavedMD5() {
        return startAnchor.getSavedMD5();
    }

    @Override
    public void updateMD5() {
        startAnchor.updateMD5();
    }

    @Override
    public void updateAnchorTag(@NotNull PsiDocTag newAnchor) {
        @Nullable PsiDocTag endTag;
        if (endAnchor != null && (endTag = endAnchor.getAnchorTag()) != null) {
            if (!endTag.getContainingFile().equals(newAnchor)) {
                return;
            }
        }
        startAnchor.updateAnchorTag(newAnchor);
        overlayElement_left.setPositionalElement(newAnchor.getContainingComment());
    }

    @Override
    @Nullable
    public PsiElement getTargetElement() {
        return startAnchor.getTargetElement();
    }

    @Override
    public boolean isTargetUnchanged(boolean update) {
        return startAnchor.isTargetUnchanged(update) && (endAnchor == null || endAnchor.isTargetUnchanged(update));
    }

    @Override
    public boolean isValid() {
        return startAnchor.isValid();
    }

    @Override
    public void invalidate() {
        startAnchor.invalidate();
    }

    @Override
    public boolean isAnchorEqualTo(@NotNull PsiDocTag docTag) {
        return startAnchor.isAnchorEqualTo(docTag) || (endAnchor != null && endAnchor.isAnchorEqualTo(docTag));
    }

    @Nullable
    @Override
    public SketchMetadataTypeC getCurrentMetaData() {
        @Nullable SketchMetadataTypeC currentMetaData = startAnchor.getCurrentMetaData();
        if (currentMetaData != null && endAnchor != null) {
            currentMetaData.setTargetType("range");
        }
        return currentMetaData;
    }

    @Override
    public SketchMetadataTypeC getMetadata() {
        return startAnchor.getMetadata();
    }

    @Override
    public void setMetadata(@NotNull SketchMetadataTypeC metadata) {
        if (metadata.equals(startAnchor.getMetadata()))
            return;
        SketchMetadataTypeC oldValue = startAnchor.getMetadata();
        startAnchor.setMetadata(metadata);
        if (endAnchor != null) {
            endAnchor.setMetadata(metadata);
        }
        propertyChangeSupport.firePropertyChange(SketchMetadata.METADATA_Property, oldValue, startAnchor.getMetadata());
    }

    @NotNull
    @Override
    public OverlayElement getOverlayElement() {
        return overlayElement_left;
    }

    @Nullable
    @Override
    public SketchMetadataTypeC isCurrentMetaDataUpdated() {
        @Nullable SketchMetadataTypeC current = getCurrentMetaData();
        if (current != null) {
            if (current.equals(getMetadata())) {
                return null;
            } else {
                return current;
            }
        }
        return null;

    }

    @Override
    @NotNull
    public Project getProject() {
        return startAnchor.getProject();
    }

    @Override
    @NotNull
    public String getAnchorId() {
        return startAnchor.getAnchorId();
    }

    @Override
    public void setAnchorId(@NotNull String id) {
        startAnchor.setAnchorId(id);
        if (endAnchor != null) {
            endAnchor.setAnchorId(id);
        }
    }

    @Override
    @NotNull
    public String getDescription() {
        return startAnchor.getDescription();
    }

    @Override
    @NotNull
    public AnchorTypes getAnchorType() {
        return startAnchor.getAnchorType();
    }

    @Override
    @NotNull
    public ImageIcon getImageIcon() {
        return myIcon;
    }

    @Override
    public void navigateTo() {
        startAnchor.navigateTo();
    }

    @Override
    @Nullable
    public LinkAnchor getParent() {
        return startAnchor.getParent();
    }

    @Override
    public void addPropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener(listener);
    }

    @Override
    public void removePropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener(listener);
    }

    @NotNull
    @Override
    public List<PropertyChangeListener> getPropertyChangeListeners() {
        //seems redundant, but Arrays.asList returns an immutable list throwing errors when removing items manually
        return new ArrayList<>(Arrays.asList(propertyChangeSupport.getPropertyChangeListeners()));
    }

    @Override
    public int compareTo(LocalLinkAnchor o) {
        if (this.equals(o) || startAnchor.equals(o)) {
            return 0;
        }
        return startAnchor.compareTo(o);
    }
}
