/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.sketchlinks;

import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.project.DumbService;
import com.intellij.openapi.project.Project;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.tools.slSearch;
import plg.ui.preview.PreviewItem;
import plg.websocket.commands.builder.GetArtifactMetaDataRequestBuilder;

import javax.swing.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Created by Peter on 03.04.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class RemoteLinkAnchorC implements RemoteLinkAnchor<SketchMetadataTypeC> {
    @NotNull
    private final LinkAnchor myLinkAnchor;

    private SketchMetadataTypeC myMetadata;
    @Nullable
    private PreviewItem myPreviewItem;
    private boolean hasSearchedLocally = false;
    //property support
    @NotNull
    private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);

    public RemoteLinkAnchorC(@NotNull String anchorId, @NotNull Project project) {
        this.myLinkAnchor = new LinkAnchorImpl(anchorId, project);
        ApplicationManager.getApplication().executeOnPooledThread(new Runnable() {
            @Override
            public void run() {
                if (getMetadata() == null) {
                    new GetArtifactMetaDataRequestBuilder(RemoteLinkAnchorC.this).send(getProject(), null);

                }
            }
        });
    }

    @NotNull
    @Override
    public Project getProject() {
        return myLinkAnchor.getProject();
    }

    @Override
    @NotNull
    public String getAnchorId() {
        return myLinkAnchor.getAnchorId();
    }

    @Override
    public void setAnchorId(@NotNull String id) {
        myLinkAnchor.setAnchorId(id);
    }

    @Override
    @Nullable
    public SketchMetadataTypeC getMetadata() {
        return myMetadata;
    }

    @Override
    public void setMetadata(@NotNull SketchMetadataTypeC metadata) {
        if (!hasSearchedLocally && getProject().getName().equals(metadata.getProjectName())) {
            //is probably a local anchor
            DumbService.getInstance(getProject()).smartInvokeLater(new Runnable() {
                @Override
                public void run() {
                    new slSearch(getProject()).runSearchForIDs(new String[]{getAnchorId()}, true, false);
                }
            });
        }
        if (metadata.equals(this.myMetadata))
            return;
        SketchMetadataTypeC oldValue = this.myMetadata;
        this.myMetadata = metadata;
        propertyChangeSupport.firePropertyChange(SketchMetadata.METADATA_Property, oldValue, this.myMetadata);

    }

    @Override
    @NotNull
    public String getDescription() {
        if (myMetadata == null) {
            return "Non local source code: no information available";
        }
        String name = (myMetadata.getTargetName() != null && !myMetadata.getTargetName().trim().isEmpty()) ? " \"" + myMetadata.getTargetName() + "\"" : "";
        return "Remote: " + myMetadata.getTargetType() + name + " in Project \"" + myMetadata.getProjectName() + "\"";
    }

    @Override
    @NotNull
    public AnchorTypes getAnchorType() {
        return myLinkAnchor.getAnchorType();
    }

    @Override
    @NotNull
    public ImageIcon getImageIcon() {
        return myLinkAnchor.getImageIcon();
    }

    @Override
    public void navigateTo() {
        @NotNull String pushPullMessage = "";
        if (myMetadata != null && getProject().getName().equals(myMetadata.getProjectName())) {
            pushPullMessage = "\nPossible Reasons: "
                    + "\n\tThe Anchor describing the Target exists in the repository, but no yet locally. "
                    + "\n\t\tSuggested Solution: Create a pull request for the correct branch"
                    + "\n\tThe Anchor has been created by a different User, who didn't yet push the corresponding code."
                    + "\n\t\tSuggested Solution: Ask your Colleagues to commit their code";
        }
        if (hasSearchedLocally) {
            JOptionPane.showMessageDialog(null, "The Target doesn't exist in the currently opened Project files or could not be found here." +
                    pushPullMessage +
                    "\nSketchlink currently doesn't support browsing to Repositories like Github.com , but might in the future." +
                    "\n\nThis ID is used inside a Project called: " + (myMetadata != null ? myMetadata.getProjectName() : "(no metainformation available") +
                    "\nin a file called: " + (myMetadata != null ? myMetadata.getPath() : "(no metainformation available")
                    + "\nand is a " + (myMetadata != null ? myMetadata.getTargetType() + (myMetadata.getTargetName() != null ? " named " + myMetadata.getTargetName() : "") : "(no metainformation available"));
        } else {
            DumbService.getInstance(getProject()).smartInvokeLater(new Runnable() {
                @Override
                public void run() {
                    new slSearch(getProject()).runSearchForIDs(new String[]{getAnchorId()}, true, true);
                }
            });
            hasSearchedLocally = true;
        }
    }

//    @Nullable
////    @Override
//    public URL getPreviewLocation() {
//        return null;
//    }
//
////    @Override
//    public void setPreviewLocation(@NotNull URL previewURL) {
//
//    }
//
//    @NotNull
////    @Override
//    public JPanel getPreview() {
//        preparePreview();
//        return myPreviewItem.getPanel();
//    }
//
//    @NotNull
////    @Override
//    public PreviewItem getPreviewItem() {
//        if (myPreviewItem == null) {
//            preparePreview();
//        }
//        return myPreviewItem;
//    }
//
////    @Override
//    public boolean clearPreviewDataIfOld(boolean forceUnload) {
//        long curTime = System.currentTimeMillis();
//        if (myPreviewItem == null) {
//            return true;
//        }
//        if (forceUnload || (curTime - myPreviewItem.getTimeOfLastAccess() > 6000000)) {
//            if (myPreviewItem.getPanel().isShowing()) {
//                return false;
//            }
//            myPreviewItem.clear();
//            myPreviewItem = null;
//            return true;
//        }
//        return false;
//    }
//
////    @Override
//    public void preparePreview() {
//        if (getMetadata() == null) {
//            ApplicationManager.getApplication().executeOnPooledThread(new Runnable() {
//                @Override
//                public void run() {
//                    if (getMetadata() == null) {
//                        new GetArtifactMetaDataRequestBuilder(RemoteLinkAnchorC.this).send(getProject(), null);
//
//                    }
//                }
//            });
//
//        }
//        if (myPreviewItem == null) {
//            myPreviewItem = new PreviewSketch(PreviewManager.getInstance(getProject()), this);
//        }
//
//    }


    @Override
    public void addPropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener(listener);
    }

    @Override
    public void removePropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener(listener);
    }

    @NotNull
    @Override
    public List<PropertyChangeListener> getPropertyChangeListeners() {
        //seems redundant, but Arrays.asList returns an immutable list throwing errors when removing items manually
        return new ArrayList<>(Arrays.asList(propertyChangeSupport.getPropertyChangeListeners()));
    }
}
