/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.sketchlinks;

import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.project.Project;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.ui.preview.PreviewItem;
import plg.ui.preview.PreviewManager;
import plg.websocket.SketchlinkChannel;
import plg.websocket.commands.builder.GetArtifactMetaDataRequestBuilder;
import plg.websocket.slWebsocketManager;

import javax.swing.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Created by Peter on 03.04.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class RemoteLinkAnchorS implements RemoteLinkAnchor<SketchMetadataTypeS>, PreviewableAnchor {
    @NotNull
    private final LinkAnchor myLinkAnchor;
    @NotNull
    private final Project project;
    private URL myURL;
    @Nullable
    private PreviewItem myPreviewItem;
    private SketchMetadataTypeS myMetadata;
    //property support
    @NotNull
    private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);

    public RemoteLinkAnchorS(@NotNull String anchorId, @NotNull Project project) {
        this.myLinkAnchor = new LinkAnchorImpl(anchorId, project);
        this.project = project;
        ApplicationManager.getApplication().executeOnPooledThread(new Runnable() {
            @Override
            public void run() {
                if (getMetadata() == null) {
                    new GetArtifactMetaDataRequestBuilder(RemoteLinkAnchorS.this).send(getProject(), null);
                }
                preparePreview();
            }
        });
    }

    @NotNull
    @Override
    public Project getProject() {
        return project;
    }

    @Override
    @NotNull
    public String getAnchorId() {
        return myLinkAnchor.getAnchorId();
    }

    @Override
    public void setAnchorId(@NotNull String id) {
        myLinkAnchor.setAnchorId(id);
    }

    @Override
    @NotNull
    public String getDescription() {

        return (myMetadata != null) ? myMetadata.getAnnotation() : myLinkAnchor.getDescription();

    }

    @Override
    @NotNull
    public AnchorTypes getAnchorType() {
        return myLinkAnchor.getAnchorType();
    }

    @Override
    @NotNull
    public ImageIcon getImageIcon() {
        return myLinkAnchor.getImageIcon();
    }

    @Override
    public void navigateTo() {
        PreviewManager.getInstance(project).showAnchorInWebView(this.getAnchorId(), true);
    }

    @Nullable
    @Override
    public URL getPreviewLocation() {
        if (myURL != null)
            return myURL;
        if (!slWebsocketManager.getInstance(project).isConnected()) {
            return myURL;
        }
        SketchlinkChannel sketchlinkChannel = slWebsocketManager.getInstance(project).getChannelManager().getAllChannels().get(0);
        @Nullable URL serverUrl = sketchlinkChannel.getEndpointAsHTTP();
        if (serverUrl == null)
            return myURL;
        try {
            myURL = new URL(serverUrl.getProtocol(), serverUrl.getHost(), serverUrl.getPort(),
                    serverUrl.getFile() + "/sketches/svg/" + myLinkAnchor.getAnchorId()); //NON-NLS
        } catch (MalformedURLException e) {
            JOptionPane.showMessageDialog(null, "Malformed url exception\n" + e.getMessage());
            e.printStackTrace();
        }
        return myURL;
    }

    @Override
    public void setPreviewLocation(@NotNull URL previewURL) {
        myURL = previewURL;
    }

    @Nullable
    @Override
    public SketchMetadataTypeS getMetadata() {
        return myMetadata;
    }

    @Override
    public void setMetadata(@NotNull SketchMetadataTypeS metadata) {
        if (metadata.equals(this.myMetadata))
            return;
        SketchMetadataTypeS oldValue = this.myMetadata;
        this.myMetadata = metadata;
        propertyChangeSupport.firePropertyChange(SketchMetadata.METADATA_Property, oldValue, this.myMetadata);
    }

    @NotNull
    @Override
    public JPanel getPreview() {
        preparePreview();
        return myPreviewItem.getPanel();
    }

    @NotNull
    @Override
    public PreviewItem getPreviewItem() {
        preparePreview();
        return this.myPreviewItem;
    }

    @Override
    public boolean clearPreviewDataIfOld(boolean forceUnload) {
        long curTime = System.currentTimeMillis();
        if (myPreviewItem == null) {
            return true;
        }

        if (forceUnload || (curTime - myPreviewItem.getTimeOfLastAccess() > 6000000)) {
            if (myPreviewItem.getPanel().isShowing()) {
                return false;
            }
            myPreviewItem.clear();
            myPreviewItem = null;
            return true;
        }
        return false;
    }

    @Override
    public void preparePreview() {
        if (getMetadata() == null) {
            ApplicationManager.getApplication().executeOnPooledThread(new Runnable() {
                @Override
                public void run() {
                    if (getMetadata() == null) {
                        new GetArtifactMetaDataRequestBuilder(RemoteLinkAnchorS.this).send(getProject(), null);
                    }
                }
            });
        }
        if (myPreviewItem == null) {
            myPreviewItem = PreviewManager.getInstance(getProject()).createSVGPreview(this, null);
        } else {
            if (getPreviewLocation() != null) {
                myPreviewItem.setRemoteLocation(getPreviewLocation());
            }
        }
    }

    @Override
    public void addPropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener(listener);
    }

    @Override
    public void removePropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener(listener);
    }

    @NotNull
    @Override
    public List<PropertyChangeListener> getPropertyChangeListeners() {
        //seems redundant, but Arrays.asList returns an immutable list throwing errors when removing items manually
        return new ArrayList<>(Arrays.asList(propertyChangeSupport.getPropertyChangeListeners()));
    }

}
