/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
package plg.sketchlinks;

import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.command.CommandProcessor;
import com.intellij.openapi.command.WriteCommandAction;
import com.intellij.openapi.components.ServiceManager;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Pair;
import com.intellij.psi.*;
import com.intellij.psi.javadoc.PsiDocComment;
import com.intellij.psi.javadoc.PsiDocTag;
import com.intellij.psi.util.PsiUtil;
import com.intellij.ui.awt.RelativePoint;
import io.netty.channel.Channel;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.overlay.SketchlinkOverlayManager;
import plg.slApplicationState;
import plg.slProjectComp;
import plg.tools.SketchIDRepair;
import plg.tools.slGuiTools;
import plg.tools.slSearch;
import plg.tools.slToolBox;
import plg.ui.preview.LocalNavigationPopup;
import plg.websocket.commands.builder.ChangeArtifactMetaDataRequestBuilder;
import plg.websocket.commands.builder.GetArtifactMetaDataRequestBuilder;
import plg.websocket.commands.builder.GetLinkedArtifactsRequestBuilder;
import plg.websocket.commands.builder.LinkRequestBuilder;
import plg.websocket.slWebsocketManager;

import java.security.InvalidParameterException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by Peter on 22.03.2014.
 * Part of Sketchlink Plugin for IntelliJ
 * @sketchlink c-f3fc8c54-f5eb-4449-abb8-577c24e6f093 381f0a8d45dd2831ab238f9160e5d019
 */
public class SketchlinkManager {
    public final static SketchlinkTools TOOLS = new SketchlinkTools();
    //    private final static SketchlinkManager instance = new SketchlinkManager();
    private final static Logger LOG = Logger.getInstance(SketchlinkManager.class);

    private final Map<Pair<LinkAnchor, LinkAnchor>, Boolean> myKnownSketchlinks = new ConcurrentHashMap<>();
    private final Map<String, LocalLinkAnchor> myLocalSketchAnchors = new ConcurrentHashMap<>();
    private final Map<String, RemoteLinkAnchor> myRemoteSketchAnchors = new ConcurrentHashMap<>();
    @NotNull
    private final Project project;

    public SketchlinkManager(@NotNull Project project) {
        this.project = project;
    }

    @NotNull
    public static SketchlinkManager getInstance(@NotNull Project project) {
        return ServiceManager.getService(project, SketchlinkManager.class);
//        return instance;
    }

    @NotNull
    private static Pair<LinkAnchor, LinkAnchor> getLinkIndexPair(@NotNull LinkAnchor sideA, @NotNull LinkAnchor sideB) {
        LinkAnchor first;
        LinkAnchor second;
        //order lexicographically, because Pair.equals doesn't detect duplicates with switched positions
        if (sideA.getAnchorId().compareTo(sideB.getAnchorId()) <= 0) {
            first = sideA;
            second = sideB;
        } else {
            first = sideB;
            second = sideA;
        }
        return Pair.create(first, second);
    }

    /**
     * The same as SketchlinkManager.getInstance(docTag.getProject()).addlocal(docTag);
     */
    @Nullable
    public static LocalLinkAnchor getAnchorOf(@NotNull final PsiDocTag docTag) {
        return getInstance(docTag.getProject()).addLocal(docTag);
    }

    /**
     * Returns all known SketchAnchors known in the Database, whose file equals the supplied file
     *
     * @param file
     * @return
     */
    @NotNull
    public List<LocalLinkAnchor> getLocalAnchors(@NotNull PsiFile file) {
        @NotNull List<LocalLinkAnchor> result = new ArrayList<>();
        for (@NotNull LocalLinkAnchor local : myLocalSketchAnchors.values()) {
            @Nullable PsiDocTag docTag = local.getAnchorTag();
            if (docTag == null)
                continue;
            if (file == docTag.getContainingFile()) {
                result.add(local);
                //TODO: only ask for elements that didn't create a request before
                requestAnchorList(local, docTag.getProject(), null);

            }
        }
        return result;
    }

    private void requestAnchorList(@NotNull LinkAnchor anchor, @NotNull Project project, @Nullable Channel channel) {
        if (slWebsocketManager.getInstance(project).isConnected()) {
            new GetLinkedArtifactsRequestBuilder(anchor).send(project, channel);
        }
    }

    /**
     * Looks up all Anchors linked to this anchor
     *
     * @param anchor  the anchor whose links to look up
     * @param channel the preferred channel to use. If null, the request is broadcasted
     */
    public void lookupAnchor(@NotNull final LinkAnchor anchor, @Nullable final Channel channel) {
        ApplicationManager.getApplication().runReadAction(new Runnable() {
            @Override
            public void run() {
                requestAnchorList(anchor, anchor.getProject(), channel);
            }
        });
    }

    /**
     * Iterates through the database and sends a websocket request for linked artifacts of known IDs
     *
     * @param includeRemote
     * @param project
     * @param preferredChannel
     */
    public void lookupAnchors(final boolean includeRemote, @NotNull final Project project, @Nullable final Channel preferredChannel) {
        ApplicationManager.getApplication().runReadAction(new Runnable() {
            @Override
            public void run() {
                for (@NotNull LocalLinkAnchor localSketchAnchor : myLocalSketchAnchors.values()) {
                    if (!localSketchAnchor.isValid()) {
                        continue;
                    }
                    requestAnchorList(localSketchAnchor, project, preferredChannel);
                }
                if (!includeRemote) {
                    return;
                }
                for (@NotNull RemoteLinkAnchor anchor : myRemoteSketchAnchors.values()) {
                    requestAnchorList(anchor, project, preferredChannel);
                }
            }
        });

    }

    private void addAnchorToOverlay(@NotNull LocalLinkAnchor anchor) {
        SketchlinkOverlayManager instance = SketchlinkOverlayManager.getInstance(anchor.getProject());
        boolean b = instance.addItem(anchor.getOverlayElement());
//        System.out.println("AddItem "+b+" "+anchor.getAnchorId());
    }

    /**
     * Adds a found DocTag to the database and creates a new LinkAnchor Object
     *
     * @param docTag the DocTag containing a sketchlink name and ID
     * @return the created LinkAnchor Object with this ID or null, if the ID is invalid
     */
    @Nullable
    public LocalLinkAnchor addLocal(@NotNull final PsiDocTag docTag) {
        if (!SketchlinkManager.TOOLS.isSketchlink(docTag)) {
            return null;
        }
        @Nullable final String id = TOOLS.getSketchlinkTagID(docTag);
        if (id == null) {
            return null;
        }
        //check if the now found local ID has been found earlier as remote-ID,
        //move this entry to the local database
        boolean isUpdate = false;
        if (myRemoteSketchAnchors.containsKey(id)) {
            myRemoteSketchAnchors.remove(id);
            isUpdate = true;
        }
        if (!myLocalSketchAnchors.containsKey(id)) {
            @NotNull LocalLinkAnchor anchor = new LocalLinkAnchorImpl(docTag, TOOLS.getReferencedPsiElement(docTag));
            myLocalSketchAnchors.put(id, anchor);
            //try to look up all Links for this id
            requestAnchorList(anchor, docTag.getProject(), null);
            addAnchorToOverlay(anchor);
            new GetArtifactMetaDataRequestBuilder(anchor).send(project, null);
//            new ChangeArtifactMetaDataRequestBuilder(anchor).queue(docTag.getProject(), null, isUpdate);
            return anchor;
        }
        //check for id collisions
        LocalLinkAnchor knownAnchor = myLocalSketchAnchors.get(id);
        if (!knownAnchor.isValid()) {
            knownAnchor.updateAnchorTag(docTag);
            addAnchorToOverlay(knownAnchor);
            if (null != knownAnchor.isCurrentMetaDataUpdated()) {
                new ChangeArtifactMetaDataRequestBuilder(knownAnchor).queue(docTag.getProject(), null, true);
            }
            return knownAnchor;
        }
        if (knownAnchor.isAnchorEqualTo(docTag)) {
            addAnchorToOverlay(knownAnchor);
            return knownAnchor;
        }
        //TODO: check here for duplicate IDs, possible invalid MD5-Hashes
        @NotNull LocalLinkAnchor duplicate = new LocalLinkAnchorImpl(docTag, TOOLS.getReferencedPsiElement(docTag));
//        addAnchorToOverlay(duplicate);
        if (knownAnchor instanceof LocalLinkAnchorRange) {
            ((LocalLinkAnchorRange) knownAnchor).addAnchor(duplicate);
        } else {
            @Nullable Document document = knownAnchor.getDocument();
            if (document != null && document.equals(duplicate.getDocument())) {
                SketchlinkOverlayManager.getInstance(knownAnchor.getProject()).removeItem(knownAnchor.getOverlayElement(), docTag.getContainingFile());
                knownAnchor = new LocalLinkAnchorRange(knownAnchor, duplicate);
//                System.out.println("Duplicate added "+knownAnchor.getAnchorId());
                myLocalSketchAnchors.put(id, knownAnchor);
            }
        }
        //ignore duplicates for now
//        Pair<LocalLinkAnchor, LocalLinkAnchor> doublePair = new Pair<>(knownAnchor,duplicate);
//        TOOLS.fixDuplicate(doublePair);
//
//        myLocalSketchAnchors.remove(id);
//        myLocalSketchAnchors.put(knownAnchor.getAnchorId(), knownAnchor);
//        myLocalSketchAnchors.put(duplicate.getAnchorId(), duplicate);
        addAnchorToOverlay(knownAnchor);
        return knownAnchor;
    }

    /**
     * Adds a Link between two SketchAnchors to the database
     *
     * @param sideA        First part of the link
     * @param sideB        Second part of the link
     * @param isDeprecated true if the created link is deprecated
     */
    public void addLink(@NotNull LinkAnchor sideA, @NotNull LinkAnchor sideB, boolean isDeprecated, boolean updateServer) {
//        @NotNull slWebsocketManager websocketManager = slWebsocketManager.getInstance(project);
//        if (!websocketManager.isConnected()) {
//            websocketManager.connect();
//        }
        if (!slProjectComp.getInstance(project).isConnectedToServer()) {
            slProjectComp.getInstance(project).setConnectedToServer(true);
        }
        if (sideA.getAnchorId().equals(sideB.getAnchorId())) {
            return; //no link to itself
        }

        @NotNull Pair<LinkAnchor, LinkAnchor> link = getLinkIndexPair(sideA, sideB);
        //if true, send update to server as update, instead as create
        boolean exists = isLinked(sideA, sideB, false);
        synchronized (myKnownSketchlinks) {
            myKnownSketchlinks.put(link, isDeprecated);
        }
        LinkRequestBuilder builder = new LinkRequestBuilder(sideA, sideB, isDeprecated, exists);
        if (updateServer) {
            builder.queue(sideA.getProject(), null);
        }
    }


    /**
     * Only return linked Anchors whose link has not been deprecated
     */
    @NotNull
    public List<LinkAnchor> getLinksNotDeprecated(@NotNull LinkAnchor anchor) {
        @NotNull List<LinkAnchor> result = new ArrayList<>();
        synchronized (myKnownSketchlinks) {
            for (@NotNull Pair<LinkAnchor, LinkAnchor> anchorPair : myKnownSketchlinks.keySet()) {
                Boolean aBoolean = myKnownSketchlinks.get(anchorPair);
                if (aBoolean == null || aBoolean) {
                    continue;
                }
                if (anchorPair.first.equals(anchor)) {
                    result.add(anchorPair.second);
                    continue;
                }
                if (anchorPair.second.equals(anchor)) {
                    result.add(anchorPair.first);
                }
            }
        }
        return result;
    }

    /**
     * Looks up all linked code Anchors available in this Project and browses to them
     */
    @NotNull
    public Set<LocalLinkAnchor> browseToLocalAnchorsLinkedTo(@NotNull Collection<LinkAnchor> anchors, @Nullable RelativePoint pointForPopup, @Nullable RemoteLinkAnchorS sketchAnchor) {
        @NotNull Set<PsiElement> anchorTargets = new HashSet<>();
        @NotNull Set<LocalLinkAnchor> localAnchor = new HashSet<>();

        for (@NotNull LinkAnchor anchor : anchors) {
            @NotNull List<LinkAnchor> anchorList = getLinksNotDeprecated(anchor);
            for (LinkAnchor linkAnchor : anchorList) {
                if (linkAnchor instanceof LocalLinkAnchor) {
                    @Nullable PsiElement targetElement = ((LocalLinkAnchor) linkAnchor).getTargetElement();
                    if (targetElement != null) {
                        anchorTargets.add(targetElement);
                        localAnchor.add((LocalLinkAnchor) linkAnchor);
                    }
                }
            }
        }
        if (anchorTargets.isEmpty()) {
            slGuiTools.ShowNotificationAtBottom(slApplicationState.RESOURCE_BUNDLE.getString("no.locally.linked.artifacts.found"), project);
            return localAnchor;
        }
        if (localAnchor.size() > 1 && pointForPopup != null && sketchAnchor != null) {
            new LocalNavigationPopup(project, pointForPopup, sketchAnchor, localAnchor, anchors);
            return new HashSet<>();
        } else {
            //update the highlighting on click
            slSearch.browseToLocal(anchorTargets);
        }
        return localAnchor;
    }

    /**
     * Retrieve all SketchAnchors linked to this anchor including deprecation state
     *
     * @param anchor the anchor whose links to look up
     * @return List of all Links, first being the linked Anchor, second being the isDeprecated value
     */
    @NotNull
    public List<Pair<LinkAnchor, Boolean>> getLinksFor(@NotNull LinkAnchor anchor) {
        @NotNull List<Pair<LinkAnchor, Boolean>> result = new ArrayList<>();
        synchronized (myKnownSketchlinks) {
            for (@NotNull Pair<LinkAnchor, LinkAnchor> anchorPair : myKnownSketchlinks.keySet()) {
                Boolean deprecated = myKnownSketchlinks.get(anchorPair);
                LinkAnchor otherSide = null;
                if (anchorPair.first.equals(anchor)) {
                    otherSide = anchorPair.second;
//                    result.add(new Pair<>(anchorPair.second, deprecated));
//                    continue;
                } else {
                    if (anchorPair.second.equals(anchor)) {
                        otherSide = anchorPair.first;
//                        result.add(new Pair<>(anchorPair.first, deprecated));
                    }
                }
                //should NOT be possible, if null means there is an invalid state in the DB
                if (deprecated == null) {
                    LOG.warn("Invalid Entry in myKnownSketchlinks: IDs are linked but state is null" + anchor.getAnchorId() + " & " + (otherSide == null ? "null" : otherSide.getAnchorId()));
                    continue;
                }
                //link is not about the looked up anchor
                if (otherSide == null) {
                    continue;
                }
                //Link is about the looked up anchor
                result.add(Pair.create(otherSide, deprecated));
            }
        }
        return result;
    }


    /**
     * Looks up an anchorID supplied id and returns its LinkAnchor Object if found, null otherwise
     */
    @Nullable
    public LinkAnchor getKnownAnchor(@NotNull String id) {
        if (myLocalSketchAnchors.containsKey(id)) {
            //System.out.println("SketchlinkMan#addRemote in LOCAL: "+id);
            return myLocalSketchAnchors.get(id);
        }
        if (myRemoteSketchAnchors.containsKey(id)) {
            //System.out.println("SketchlinkMan#addRemote in REMOTE: "+id);
            return myRemoteSketchAnchors.get(id);
        }
        return null;
    }

    /**
     * Adds a new Sketchlink not available inside the current project
     *
     * @param id
     * @param parentID
     * @param project
     * @return the newly created Anchor object or the old object if duplicate
     */
    @NotNull
    public LinkAnchor addRemote(@NotNull String id, @Nullable String parentID, @NotNull Project project) {
        @Nullable LinkAnchor knownAnchor = getKnownAnchor(id);
        if (knownAnchor != null) {
            return knownAnchor;
        }
        //System.out.println("SketchlinkMan#addRemote CALLING url: "+id);

        RemoteLinkAnchor anchor;
        char i = id.toLowerCase().charAt(0);
        if (i == 'c') {
            anchor = new RemoteLinkAnchorC(id, project);
        } else if (i == 's') {
            anchor = new RemoteLinkAnchorS(id, project);
        } else if (i == 'm') {
            if (parentID == null)
                throw new InvalidParameterException("unknown artifactID starts with 'm', but no parentID is supplied " + id);
            @NotNull LinkAnchor parentAnchor = addRemote(parentID, null, project);
            if (!(parentAnchor instanceof RemoteLinkAnchorS))
                throw new InvalidParameterException("unknown artifactID starts with 'm', but parentID is not an sAnchor " + id + " " + parentID);
            anchor = new RemoteLinkAnchorM(id, (RemoteLinkAnchorS) parentAnchor, project);

        } else {
            throw new InvalidParameterException("unknown artifactID format. Only s-,m- and c-Anchors supported");
        }
        myRemoteSketchAnchors.put(id, anchor);

        //new getSketchRequestBuilder(id).send(project);
        return anchor;
    }

    /**
     * Check all remote SketchAnchors and clear their preview data if possible.
     * By default, only Previews older than a certain time get cleaned up.
     *
     * @param forceUnload force unloading of preview data
     * @return the amount of anchors, who actually removed their previews.
     */
    public int cleanOldPreviews(boolean forceUnload) {
        int cleaned = 0;
        for (@NotNull RemoteLinkAnchor remotes : myRemoteSketchAnchors.values()) {
            if (remotes instanceof PreviewableAnchor) {
                cleaned += ((PreviewableAnchor) remotes).clearPreviewDataIfOld(forceUnload) ? 1 : 0;
            }
        }
        return cleaned;
    }

    public void clearAllData() {
        cleanOldPreviews(true);
        myRemoteSketchAnchors.clear();
        myKnownSketchlinks.clear();
        myLocalSketchAnchors.clear();
    }


    /**
     * Creates a new Sketchlink-Anchor for targetElement, if none exists yet.
     *
     * @param targetElement the artifact to create an anchor to
     * @return The created ID or null, if ID already exists or couldn't be created
     */
    @Nullable
    public String createSketchlink(@NotNull PsiElement targetElement) {
        //try to find an already available sketchlink-anchor for this element.
        int insertOffset = TOOLS.getDocCommentOffset(targetElement);


        @NotNull PsiElement element = PsiUtil.getElementAtOffset(targetElement.getContainingFile(), insertOffset);
        element = TOOLS.getTopmostLinkableElement(element);
        if (element instanceof PsiDocComment) {
            if (TOOLS.checkCommentContainsSketchlink((PsiDocComment) element)) {
                LOG.info("Location already contains a sketchlink. Target: " + targetElement
                        + " SketchlinkId: " + TOOLS.getFirstSketchlinkTag((PsiDocComment) element));
                return null;
            }
            return insertSketchlinkDocTag((PsiDocComment) element);
        } else {
            return insertSketchlinkDocComment(element, insertOffset);
        }

    }

    /**
     * Generates a new Sketchlink cAnchor ID (first) and it's corresponding md5-hash (second)
     *
     * @param element
     * @return cAnchor-ID in first, md5 in second; null, if the MD5 couldn't be calculated
     */
    @Nullable
    private Pair<String, String> calculateSketchlinkID(@NotNull PsiElement element) {
        //the md5 part

        @Nullable String md5 = slToolBox.calculateMD5(slToolBox.getRelativePath(element));
        if (md5 == null || md5.equals("0")) {
            LOG.debug("Couldn't create Sketchlink, the MD5 or relative Path for the referenced Element couldn't be calculated." +
                    "\nSupplied Element: " + element + " Result of MD5: " + md5);
            return null;
        }
        //the UUID-part
        UUID uuid = UUID.randomUUID();

        //        sb.append(" ");
//        sb.append(md5);
//        sb.append(" ");

        //LOG.info("Created new SketchlinkID for " + element + " with ID: " + sb.toString());
        //TODO: check for existing UUIDs in the same file needed? Chance for duplicates are very low
        return Pair.create("c-" + uuid.toString(), md5);
    }

    /**
     * Creates and Inserts a new Sketchlink DocComment at the given insertOffset
     *
     * @param referencedElement The topmost Element to create the comment for
     * @param insertOffset      the actual offset in the document, where the comment will be inserted
     */
    @Nullable
    private String insertSketchlinkDocComment(@NotNull final PsiElement referencedElement, final int insertOffset) {
        if (referencedElement instanceof PsiDocComment) {
            if (TOOLS.checkCommentContainsSketchlink((PsiDocComment) referencedElement))
                return null;
            return insertSketchlinkDocTag((PsiDocComment) referencedElement);

        }
        //Don't doing it for now, because the comment might be code
//        if (referencedElement instanceof PsiComment) {
//            convertCommentAndInsertSketchlink((PsiComment) referencedElement);
//            return;
//        }
        @Nullable Pair<String, String> sketchlinkID = calculateSketchlinkID(referencedElement);
        if (sketchlinkID == null) {
            return null;
        }
        final boolean asMultiLine = (referencedElement instanceof PsiClass
                || referencedElement instanceof PsiMethod
                || referencedElement instanceof PsiField);
        @Nullable final Document document =
                PsiDocumentManager.getInstance(referencedElement.getProject()).getDocument(referencedElement.getContainingFile());
        @NotNull final StringBuilder whitespace = new StringBuilder("");
        if (document == null) {
            return null;
        }
        for (int i = document.getLineStartOffset(document.getLineNumber(referencedElement.getTextRange().getStartOffset()))
                     ; i < referencedElement.getTextRange().getStartOffset(); i++) {
            whitespace.append(" ");
        }
        @NotNull final StringBuilder sb = new StringBuilder("/**");
        if (asMultiLine) {
            sb.append("\n");
            sb.append(whitespace);
            sb.append(" * ");
        } else {
            sb.insert(0, "  "); //add 2 whitespaces between the element and the docComment
            sb.append(" ");
        }
        sb.append("@");
        sb.append(slApplicationState.SKETCHLINK_TAGNAME);
        sb.append(" ");
        sb.append(sketchlinkID.first); //here goes the created sketchlink ID
        sb.append(" ");
        sb.append(sketchlinkID.second);
        sb.append(" ");
        if (asMultiLine) {
            sb.append("\n");
            sb.append(whitespace);
            sb.append(" ");
        } else {
            sb.append(" ");
        }
        sb.append("*/");
        if (asMultiLine) {
            sb.append("\n");
            sb.append(whitespace);
        }
        CommandProcessor.getInstance().executeCommand(referencedElement.getProject(),
                new Runnable() {
                    @Override
                    public void run() {
                        WriteCommandAction.runWriteCommandAction(null, new Runnable() {
                            @Override
                            public void run() {
                                document.insertString(insertOffset, sb.toString());
                            }
                        });

                    }
                }, slApplicationState.RESOURCE_BUNDLE.getString("create.sketchlink.comment"), null
        );
        return sketchlinkID.first;
    }


    /**
     * Adds a Sketchlink Anchor to an existing PsiDocComment if none exists
     *
     * @param comment where to add the Sketchlink anchor
     */
    @Nullable
    private String insertSketchlinkDocTag(@NotNull final PsiDocComment comment) {
        if (SketchlinkManager.TOOLS.checkCommentContainsSketchlink(comment)) {
            return null;
        }
        @Nullable Pair<String, String> sketchlinkID = calculateSketchlinkID(comment);
        if (sketchlinkID == null) {
            return null;
        }
        //contains the final data
        @NotNull final StringBuilder sb = new StringBuilder("");
        //if comment will be multiline, whitespace contains enough spaces to keep indentation intact
        @NotNull final StringBuilder whitespace = new StringBuilder(" ");
        sb.append("@");
        sb.append(slApplicationState.SKETCHLINK_TAGNAME);
        sb.append(" ");
        sb.append(sketchlinkID.first);  //here goes the created Sketchlink ID
        sb.append(" ");
        sb.append(sketchlinkID.second);
        sb.append(" ");
        @Nullable final Document document = PsiDocumentManager.getInstance(comment.getProject()).getDocument(comment.getContainingFile());
        if (document != null) {
            // sb.append("\n");
            for (int i = document.getLineStartOffset(document.getLineNumber(comment.getTextRange().getStartOffset()))
                         ; i < comment.getTextRange().getStartOffset(); i++) {
                whitespace.append(" ");
            }
        } else {
            return null;
        }
        PsiElement checkWhiteSpace = comment.getLastChild();
        while (checkWhiteSpace.getPrevSibling() instanceof PsiWhiteSpace) {
            checkWhiteSpace = checkWhiteSpace.getPrevSibling();
        }
        final PsiElement beforeHere = checkWhiteSpace;
        sb.append(whitespace);
        if (beforeHere.getNextSibling() instanceof PsiDocTag) {
            sb.append("* ");
        } else {
            sb.insert(0, "\n* ");
            sb.insert(1, whitespace);
        }
        CommandProcessor.getInstance().executeCommand(beforeHere.getProject(),
                new Runnable() {
                    @Override
                    public void run() {
                        WriteCommandAction.runWriteCommandAction(null, new Runnable() {
                            @Override
                            public void run() {
                                document.insertString(beforeHere.getTextRange().getStartOffset(), sb.toString());
                            }
                        });

                    }
                }, slApplicationState.RESOURCE_BUNDLE.getString("insert.sketchlink.anchor"), null
        );
        return sketchlinkID.first;
    }

    /**
     * Rechecks the locally existing SketchAnchors for changes and repairs their indexing
     */
    public void recheckLocalIndizes() {
        ApplicationManager.getApplication().runReadAction(new Runnable() {
            @Override
            public void run() {
                @NotNull Iterator<String> iterator = myLocalSketchAnchors.keySet().iterator();
                @NotNull List<LocalLinkAnchor> askUserAboutList = new ArrayList<>();
                @NotNull Map<String, LocalLinkAnchor> fixed = new HashMap<>();
                while (iterator.hasNext()) {
                    String currentIndexID = iterator.next();
                    LocalLinkAnchor localSketchAnchor = myLocalSketchAnchors.get(currentIndexID);
                    @Nullable PsiDocTag docTag = localSketchAnchor.getAnchorTag();
                    if (docTag == null || !docTag.isValid()) {
                        //invalid sketchanchor
                        iterator.remove();
                        continue;
                    }
                    @Nullable String currentDogTagID = TOOLS.getSketchlinkTagID(docTag);
                    if (currentDogTagID == null || !TOOLS.isValidID(docTag)) {
                        //System.out.println("recheckLocalIndizes Part1 " + currentDogTagID);
                        askUserAboutList.add(localSketchAnchor);
                        continue;
                    }
                    if (currentIndexID.equals(currentDogTagID)) {
                        continue;
                    }
                    iterator.remove();
                    localSketchAnchor.setAnchorId(currentDogTagID);
                    if (myLocalSketchAnchors.containsKey(currentDogTagID)) {
                        LocalLinkAnchor possibleDuplicate = myLocalSketchAnchors.get(currentDogTagID);
                        if (possibleDuplicate.equals(localSketchAnchor)) {
                            continue;
                        }
                        if (!possibleDuplicate.isValid()) {
                            fixed.put(currentDogTagID, localSketchAnchor);
                        }
                        @NotNull Pair<LocalLinkAnchor, LocalLinkAnchor> duplicate = Pair.create(localSketchAnchor, possibleDuplicate);
                        myLocalSketchAnchors.remove(currentDogTagID);
                        //System.out.println("recheckLocalIndizes Part2 " + currentDogTagID);
                        TOOLS.fixDuplicate(duplicate);
                        @NotNull String possibleDuplicateID = duplicate.second.getAnchorId();
                        currentDogTagID = duplicate.first.getAnchorId();
                        fixed.put(possibleDuplicateID, possibleDuplicate);
                    }
                    fixed.put(currentDogTagID, localSketchAnchor);
                }
                //add fixed items to the final set
                myLocalSketchAnchors.putAll(fixed);
                //System.out.println("recheckLocalIndizes Part3 ");
                SketchIDRepair.RepairSketchAnchors(askUserAboutList);

            }
        });
    }

    /**
     * Delete the supplied anchorID from the local database and delete all links this ID is part of in the local & remote database
     *
     * @param anchorID           the id to delete
     * @param deprecateLinksOnly delete the id, but only deprecate the links it's part of
     */
    public void deleteAnchor(@NotNull String anchorID, boolean deprecateLinksOnly) {
        @Nullable LinkAnchor knownAnchor = getKnownAnchor(anchorID);
        if (knownAnchor == null) {
            return;
        }
        myLocalSketchAnchors.remove(anchorID);
        myRemoteSketchAnchors.remove(anchorID);

        synchronized (myKnownSketchlinks) {
            Iterator<Pair<LinkAnchor, LinkAnchor>> iterator = myKnownSketchlinks.keySet().iterator();
            while (iterator.hasNext()) {
                Pair<LinkAnchor, LinkAnchor> pair = iterator.next();
                if (anchorID.equals(pair.first.getAnchorId()) || anchorID.equals(pair.second.getAnchorId())) {
                    if (deprecateLinksOnly) {
                        addLink(pair.first, pair.second, true, true);
                    } else {
                        new LinkRequestBuilder(pair.first, pair.second, true, true, true);
                        iterator.remove();
                    }
                }
            }
        }
//        for (@NotNull Pair<LinkAnchor, LinkAnchor> anchorPair : myKnownSketchlinks.keySet()) {
//            if (anchorID.equals(anchorPair.first.getAnchorId()) || anchorID.equals(anchorPair.second.getAnchorId())) {
//                if (deprecateLinksOnly) {
//                    addLink(anchorPair.first, anchorPair.second, true, true);
//                } else {
//                    removeLink(anchorPair.first, anchorPair.second, true);
//                }
//            }
//        }

    }


    /**
     * Iterates through all locally available SketchAnchors and updates their Metadata on the server
     */
    public void updateServerMetaDataOnLocalData() {
        for (LocalLinkAnchor localSketchAnchor : myLocalSketchAnchors.values()) {
            if (localSketchAnchor instanceof LocalLinkAnchorImpl) {
                @Nullable SketchMetadataTypeC currentMetaDataUpdated = localSketchAnchor.isCurrentMetaDataUpdated();
                if (currentMetaDataUpdated != null) {
                    new ChangeArtifactMetaDataRequestBuilder(localSketchAnchor).queue(localSketchAnchor.getProject(), null, true);
                }
            }
        }

    }

    public boolean isLinked(@NotNull LinkAnchor sideA, @NotNull LinkAnchor sideB, boolean ignoreDeprecate) {
        @NotNull Pair<LinkAnchor, LinkAnchor> indexPair = getLinkIndexPair(sideA, sideB);
        Boolean isDeprecated = myKnownSketchlinks.get(indexPair);
        return !(isDeprecated == null || (isDeprecated && !ignoreDeprecate));
    }

    @NotNull
    public RemoteLinkAnchorM createMarker(@NotNull String id, @NotNull RemoteLinkAnchorS parentSketch, @NotNull SketchMetadataTypeM metadata, boolean toServer) {
        @NotNull RemoteLinkAnchorM result = new RemoteLinkAnchorM(id, parentSketch, parentSketch.getProject());
        result.setMetadata(metadata);
        myRemoteSketchAnchors.put(id, result);
        if (toServer) {
            new ChangeArtifactMetaDataRequestBuilder(result).queue(result.getProject(), null, false);
        }
        return result;
    }


    @NotNull
    @SuppressWarnings("unchecked")
    public <T extends LinkAnchor> Collection<T> getAllAnchors(Class<T> clazz) {
        Collection<T> result = new HashSet<>();
        for (RemoteLinkAnchor linkAnchor : myRemoteSketchAnchors.values()) {
            if (clazz.isInstance(linkAnchor)) {
                result.add((T) linkAnchor);
            }
        }
        for (LocalLinkAnchor linkAnchor : myLocalSketchAnchors.values()) {
            if (clazz.isInstance(linkAnchor)) {
                result.add((T) linkAnchor);
            }
        }
        return result;
    }

    public void removeAnchorFromCache(@NotNull LocalLinkAnchor localAnchor) {

        myLocalSketchAnchors.remove(localAnchor.getAnchorId());

    }
}
