/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
package plg.sketchlinks;

import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.command.CommandProcessor;
import com.intellij.openapi.command.UndoConfirmationPolicy;
import com.intellij.openapi.command.WriteCommandAction;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Computable;
import com.intellij.openapi.util.Pair;
import com.intellij.openapi.util.TextRange;
import com.intellij.openapi.vcs.VcsShowConfirmationOption;
import com.intellij.psi.*;
import com.intellij.psi.javadoc.PsiDocComment;
import com.intellij.psi.javadoc.PsiDocTag;
import com.intellij.psi.javadoc.PsiDocToken;
import com.intellij.psi.tree.IElementType;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.util.PsiUtil;
import com.intellij.util.ui.ConfirmationDialog;
import com.intellij.util.ui.UIUtil;
import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Hex;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.slApplicationState;
import plg.tools.slToolBox;
import plg.websocket.commands.builder.ChangeArtifactMetaDataRequestBuilder;

import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.RunnableFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * Tool class for handling and calculation of different Aspects regarding Sketchlink Anchors.
 * Using non-static Methods to improve flexibility, the default way of accessing this would be through
 * <code>SketchlinkManager.TOOLS</code>
 */
public class SketchlinkTools {
    private final static Logger LOG = Logger.getInstance(SketchlinkTools.class);


    @Nullable
    @Contract("null -> null")
    public LinkAnchor.AnchorTypes checkAnchorType(@Nullable String anchorID) {
        if (anchorID == null || !isValidID(anchorID)) {
            return null;
        }
        if (anchorID.toLowerCase().charAt(0) == 'm')
            return LinkAnchor.AnchorTypes.SVG_Rectangle_Anchor;
        if (anchorID.toLowerCase().charAt(0) == 's')
            return LinkAnchor.AnchorTypes.SVG_Anchor;
        if (anchorID.toLowerCase().charAt(0) == 'c')
            return LinkAnchor.AnchorTypes.Code_Anchor;
        return null;
    }


    /**
     * Calculates the probably referenced PsiElement for the PsiDocTag. Since PsiDocTags/Anchors have no inherently saved
     * location, the result depends on the state of the PsiTree at the moment of calling.
     * <p/>
     */
    @NotNull
    public PsiElement getReferencedPsiElement(@NotNull PsiDocTag sketchlinkAnchor) {
        PsiElement result = sketchlinkAnchor.getParent().getParent();
        //Classes, Methods and Class Variables
        if (result instanceof PsiNameIdentifierOwner) {
            return result;
        }
        @Nullable Document document = PsiDocumentManager.getInstance(sketchlinkAnchor.getProject()).getDocument(sketchlinkAnchor.getContainingFile());
        if (document == null) {
            //old and faulty message of looking up, ignores several possible sketchlink-locations
            LOG.debug("getReferencedPsiElement: document==null, using the old and buggy way to calculate the reference");
            result = sketchlinkAnchor.getParent().getNextSibling();
            while (result != null && (result instanceof PsiWhiteSpace)) {
                result = result.getNextSibling();
            }
            if (result == null) {
                //somehow nothing found, return the parent of this DocComment
                //returns null if the psidocTag is above the class declaration
                return sketchlinkAnchor.getParent();
            }
            if (result instanceof PsiCodeBlock) {
                //comment is last item inside a message,
                return sketchlinkAnchor.getParent().getParent();
            }
            return result;
        }
        result = sketchlinkAnchor.getParent();
        int linestartoffset = document.getLineStartOffset(document.getLineNumber(result.getTextRange().getStartOffset()));
        result = PsiUtil.getElementAtOffset(sketchlinkAnchor.getContainingFile(), linestartoffset); //probably PsiWhitespace at beginning of the line
        @Nullable PsiElement nextLeaf = PsiTreeUtil.nextLeaf(result, true);
        if (nextLeaf != null) {
            result = getTopmostLinkableElement(nextLeaf);
        }
        if (nextLeaf != null && !(result instanceof PsiDocComment)) {
            return result;
        }
        //either nextLeaf is still null or the DocComment of the sketchlinkanchor is in it's own line. Choose the following element
        result = sketchlinkAnchor.getParent();
        if (result.getNextSibling() != null) {
            result = result.getNextSibling();
            if (result instanceof PsiWhiteSpace) {
                result = result.getNextSibling();
            }
        } else {
            if (result.getParent().getNextSibling() != null) {
                result = result.getParent().getNextSibling();
            }
        }
        return result;
    }

    /**
     * @see plg.sketchlinks.SketchlinkTools#checkCommentContainsSketchlink(com.intellij.psi.javadoc.PsiDocComment, boolean)
     * and parameter false
     */
    public boolean checkCommentContainsSketchlink(@NotNull PsiDocComment comment) {
        return checkCommentContainsSketchlink(comment, false);
    }

    /**
     * Checks if the current DocComment contains a @sketchlink-DocTag.
     *
     * @param checkPure true: the current DocComment should only contain @sketchlink-DocTags and nothing else
     * @return true, if it contains a (only) sketchlink-DocTags, false: it doesn't contain this DocTag
     */
    public boolean checkCommentContainsSketchlink(@NotNull PsiDocComment comment, boolean checkPure) {
        @NotNull PsiElement[] children = comment.getChildren();
        boolean isPure = true;
        boolean isMagicComment = false;
        for (PsiElement child : children) {
            if (child instanceof PsiDocTag) {
                //if the DocComment contains a non-sketchlink DocTag
                if (!slApplicationState.SKETCHLINK_TAGNAME.equals(((PsiDocTag) child).getName())) {
                    isPure = false;
                    continue;
                } else {
                    if (getSketchlinkTagID((PsiDocTag) child) != null)
                        isMagicComment = true;
                }
            }
            if (child instanceof PsiDocToken) {
                @NotNull PsiDocToken psiDocToken = (PsiDocToken) child;
                IElementType iElementType = psiDocToken.getTokenType();
                //Don't allow additional Comments inside a magic comment
                if (iElementType.toString().equalsIgnoreCase("DOC_COMMENT_DATA") && psiDocToken.getTextLength() > 1) {
                    isPure = false;
                }
            }
        }
        //returns true if element contains a @sketchlink tag AND isPure->checkPure
        return isMagicComment && (isPure || !checkPure);
    }

    /**
     * Extracts the ID from the supplied PsiDocTag
     *
     * @param docTag the Tag containing a sketchlink-id
     * @return null if the docTag is malformed or not a sketchlink-tag,
     * a String representation of the ID otherwise
     */
    @Nullable
    public String getSketchlinkTagID(@NotNull final PsiDocTag docTag) {
        return ApplicationManager.getApplication().runReadAction(new Computable<String>() {
            @Override
            public String compute() {
                @NotNull String[] split = docTag.getText().split(" |\n|\t");
                for (int i = 1; i < split.length; i++) {
                    if (split[i - 1].equals("@" + slApplicationState.SKETCHLINK_TAGNAME)) {
                        return isValidID(split[i]) ? split[i] : null;
                    }
                }
                return null;
            }
        });
    }

    @Nullable
    public String getSketchlinkTagMD5(@NotNull final PsiDocTag docTag) {
        return ApplicationManager.getApplication().runReadAction(new Computable<String>() {
            @Override
            public String compute() {
                @NotNull String[] split = docTag.getText().split(" |\n|\t");
                for (int i = 2; i < split.length; i++) {
                    if (split[i - 2].equals("@" + slApplicationState.SKETCHLINK_TAGNAME)) {
                        try {
                            Hex.decodeHex(split[i].toCharArray());
                        } catch (DecoderException e) {
                            return null;
                        }
                        return split[i];
                    }
                }
                return null;
            }
        });

    }

    @Contract("null -> false")
    public boolean isValidID(@Nullable PsiDocTag docTag) {
        if (docTag == null) {
            return false;
        }
        @NotNull String[] idParts = getIdParts(docTag);
        return (idParts[0].length() == 1
                && idParts[1].length() == 36);
    }

    /**
     * Checks the formatting of the supplied id String
     *
     * @param id the String to check for valid formatting
     * @return true if valid, false otherwise
     */
    @Contract("null -> false")
    public boolean isValidID(@Nullable String id) {
        if (id == null)
            return false;
        if (id.contains(" ")) {
            LOG.info("Found invalid sketchlink id format, containing a whitespace: " + id);
            return false;
        }
        if (id.length() != 38) {
            return false;
        }
        @NotNull String[] idParts = getIdParts(id, null);
        return !(idParts[0].isEmpty()
                || idParts[1].isEmpty());
    }


    /**
     * Generates a List with all DocTags 'sketchlink' inside the supplied comment
     *
     * @param comment the DocComment to parse
     * @return List containing all Doctags.
     */
    @NotNull
    public List<PsiDocTag> getSketchlinkTags(@NotNull PsiDocComment comment) {
        @NotNull List<PsiDocTag> result = new ArrayList<>();
        for (PsiElement child : comment.getChildren()) {
            if ((child instanceof PsiDocTag) && getSketchlinkTagID((PsiDocTag) child) != null) {
                result.add((PsiDocTag) child);
            }
        }
        return result;
    }

    @Nullable
    public PsiDocTag getFirstSketchlinkTag(@NotNull PsiDocComment comment) {
        for (PsiElement child : comment.getChildren()) {
            if ((child instanceof PsiDocTag) && getSketchlinkTagID((PsiDocTag) child) != null) {
                return (PsiDocTag) child;
            }
        }
        return null;
    }

    /**
     * Checks whether the supplied PsiDocTag is really a Sketchlink. Returns true if it's name is correct.
     * Does NOT check the internal validity of the Anchor.
     *
     * @return true if it's a sketchlink-anchor, false otherwise.
     */
    public boolean isSketchlink(@NotNull PsiDocTag element) {
        return element.getName().equals(slApplicationState.SKETCHLINK_TAGNAME);
    }

    /**
     * Checks whether the given DocComment is in an officially supported location.
     * <p/>See <a href="http://www.oracle.com/technetwork/java/javase/documentation/index-137868.html">Javadoc Documentation</a>
     * for more information
     */
    public boolean isDefaultDocCommentLocation(@NotNull PsiDocComment psiDocComment) {
        PsiElement parent = psiDocComment.getParent();
        return ((parent instanceof PsiClass) && !(parent instanceof PsiAnonymousClass))
                || (parent instanceof PsiMethod)
                || (parent instanceof PsiField);
    }

    /**
     * Gets the actual LocalLinkAnchor for the supplied PsiElement.
     * The supplied element doesn't have to be same as {@link LocalLinkAnchor#getTargetElement()}
     *
     * @param element the PsiElement to lookup the Sketchlink Anchor for
     * @return null if no Sketchlink anchor is found,
     * the anchor otherwise
     */
    @Nullable
    public PsiDocTag getSketchlinkForElement(final @NotNull PsiElement element) {
        @NotNull Computable<PsiDocTag> result = new Computable<PsiDocTag>() {
            @Nullable
            @Override
            public PsiDocTag compute() {
                @NotNull PsiElement target = PsiUtil.getElementAtOffset(element.getContainingFile(), getDocCommentOffset(element));
                if (target instanceof PsiDocToken) {
                    target = target.getParent();
                }
                if (target instanceof PsiDocComment) {
                    return getFirstSketchlinkTag((PsiDocComment) target);
                }
                return null;
            }
        };
        return ApplicationManager.getApplication().runReadAction(result);
    }


    /**
     * Checks if there is a DocComment containing a Sketchlink prior the given PsiElement.
     */
    public boolean hasSketchlink(@NotNull PsiElement element) {
        return getSketchlinkForElement(element) != null;
    }

    @Nullable
    private PsiElement getNextSiblingSkipWhiteSpace(@NotNull PsiElement psiElement) {
        PsiElement sibling = psiElement.getNextSibling();
        if (sibling instanceof PsiWhiteSpace)
            return sibling.getNextSibling();
        return sibling;
    }

    @Nullable
    @Contract("null -> null")
    private PsiElement getPreviousSiblingSkipWhiteSpace(@Nullable PsiElement psiElement) {
        if (psiElement == null)
            return null;
        PsiElement sibling = psiElement.getPrevSibling();
        if (sibling instanceof PsiWhiteSpace)
            return sibling.getPrevSibling();
        return sibling;
    }

    private int offsetOfPreviousSiblingIfDocComment(@NotNull PsiElement element) {
        @Nullable PsiElement sibling = getPreviousSiblingSkipWhiteSpace(element);
        if (sibling instanceof PsiDocComment)
            return sibling.getTextRange().getStartOffset();
        return element.getTextRange().getStartOffset();
    }

    /**
     * Calculates the topmost parent for which it makes sense to create a Sketchlink Anchor
     *
     * @param element
     * @return the topmost parent or the supplied element, if it's already the best location or none else could be found
     */
    @NotNull
    public PsiElement getTopmostLinkableElement(@NotNull PsiElement element) {
        @Nullable PsiElement target = element;
        if (target instanceof PsiWhiteSpace && getPreviousSiblingSkipWhiteSpace(target) != null) {
            target = getPreviousSiblingSkipWhiteSpace(target);
        }
        @Nullable PsiElement temp;
        //traverse the PsiTree and check at common locations
        while (target != null) {


            if (target instanceof PsiClass && !(target instanceof PsiAnonymousClass)) {
                return target;
            }
            if (target instanceof PsiField) {
                return target;
            }
            if (target instanceof PsiMethod) {
                return target;
            }
//            if (target.getParent() instanceof PsiClass || target.getParent() instanceof PsiMethod || target.getParent() instanceof PsiField) {
//                target = target.getParent();
//            }
            if (target instanceof PsiSwitchLabelStatement) {
                //the "case"-part inside a switch statement
                return target;
            }
            if (target instanceof PsiComment && !(target instanceof PsiDocComment)) {
                PsiElement prevSibling = target.getPrevSibling();
                if (prevSibling != null && !prevSibling.getText().contains("\n")) {
                    PsiElement psiElement = getPreviousSiblingSkipWhiteSpace(target);
                    if (psiElement != null) {
                        target = psiElement;
                    }
                } else {
                    PsiElement psiElement = getNextSiblingSkipWhiteSpace(target);
                    if (psiElement != null) {
                        target = psiElement;
                    }
                }
            }

            if (target instanceof PsiDocComment) {
                return target;
            }
            if (getPreviousSiblingSkipWhiteSpace(target) instanceof PsiDocComment
                    && !(target.getParent() instanceof PsiClass
                    || target.getParent() instanceof PsiMethod
                    || target.getParent() instanceof PsiField)) {
                return target;
            }
            if (target.getParent() instanceof PsiIfStatement) {
                //target can be either the "if (true)" part or a line of code when no brackets where used
                temp = getNextSiblingSkipWhiteSpace(target);
                if (temp == null || (temp instanceof PsiKeyword && temp.getText().equals("else"))) {
                    //the keyword can only be "else"; check if there's a DocComment between the if(true) part and the codeline
                    return target;
                }
            }
            if (target.getParent() instanceof PsiLabeledStatement) {
                return target.getParent();
            }
            if (target instanceof PsiIfStatement) {
                return target;
            }
            if (target instanceof PsiLoopStatement) {
                return target;
            }
            if (target.getParent() instanceof PsiFile) {
                //element was either the package declaration or import statement
                return target;
            }
            if (target.getParent() instanceof PsiCodeBlock && !(target instanceof PsiJavaToken)) {
                //for PsiSwitchStatements, all "case:" and PsiElements are
                //most common case for Sketchlink anchors inside a Method
                return target;
            }
            if (target.getParent() instanceof PsiLoopStatement) {
                if (target.getParent() instanceof PsiDoWhileStatement) {
                    /*  do-while Loops have a different order inside the PsiTree, the statement is in between.
                        Most common do-while Loops are caught by the PsiCodeBlock-check,
                        however check for those loops without Brackets
                     */
                    int i = 0;
                    do i++; while (i < 4);
                    temp = target.getPrevSibling();
                    while (temp != null && !(temp instanceof PsiKeyword)) {
                        temp = temp.getPrevSibling();
                    }
                    if (temp != null && !temp.getText().equals(PsiKeyword.WHILE)) {
                        return target;
                    }
                } else {
                    //normal while, for and foreach Statements
                    if (target.equals(target.getParent().getLastChild())) {
                        //actual looping code is always the last child for those statements
                        return target;
                    }
                }
            }
            //traverse upwards and ignore all other PsiElement types
            target = target.getParent();
        }
        //nothing found
        System.out.println("SketchlinkTools#getTopMostLinkableElement - nothing found");
        return element;
    }

    /**
     * Finds the last closing parenthese in the child, or null if none available
     */
    @Nullable
    private PsiElement getChildWithLastClosingParenthese(@NotNull PsiElement element) {
        PsiElement child = element.getFirstChild();
        int parentheseOpen = 0;
        while (child != null) {
            if (child instanceof PsiJavaToken) {
                if (child.getText().equals("(")) {
                    parentheseOpen++;
                }
                if (child.getText().equals(")")) {
                    parentheseOpen--;
                    if (parentheseOpen == 0) {
                        return child;
                    }
                }
            }
            child = child.getNextSibling();
        }
        return null;
    }


    /**
     * Calculates the Offset where to look for or create sketchlinks for this element.
     * If element is of Instance Whitespace, the offset for the previous sibling is calculated.
     *
     * @param element the element to calculate the sketchlink position for
     * @return the beginning offset of the DocComment OR where one would be expected to be
     */
    public int getDocCommentOffset(@NotNull PsiElement element) {
        @Nullable PsiElement target = getTopmostLinkableElement(element);
        if (target instanceof PsiWhiteSpace && getPreviousSiblingSkipWhiteSpace(target) != null) {
            target = getPreviousSiblingSkipWhiteSpace(target);
        }
        //TODO: variable probably no longer needed
//        int previousOffset = -1;
        //always prefer to add the comment near the beginning of the element, not the end
        boolean preferCommentBefore = false;
        //traverse the PsiTree and check at common locations
        if (target instanceof PsiClass && !(target instanceof PsiAnonymousClass)) {
            return target.getTextRange().getStartOffset();
        }
        if (target instanceof PsiField) {
            return target.getTextRange().getStartOffset();
        }
        if (target instanceof PsiMethod) {
            return target.getTextRange().getStartOffset();
        }
        if (target instanceof PsiSwitchStatement) {
            preferCommentBefore = true;
//            previousOffset = offsetOfPreviousSiblingIfDocComment(target);
        }
//        if (target instanceof PsiSwitchLabelStatement) {
//            //the "case"-part inside a switch statement
//            previousOffset = offsetOfPreviousSiblingIfDocComment(target);
//        }
        if (target instanceof PsiDocComment) {
            return target.getTextRange().getStartOffset();
        }
        if (target instanceof PsiIfStatement) {
            preferCommentBefore = true;
//            previousOffset = offsetOfPreviousSiblingIfDocComment(target);
        }
//        if (target != null && target.getParent() instanceof PsiFile) {
//            //element was either the package declaration or import statement
//            previousOffset = offsetOfPreviousSiblingIfDocComment(target);
//        }
        if (target instanceof PsiLoopStatement) {
            preferCommentBefore = true;
//            previousOffset = offsetOfPreviousSiblingIfDocComment(target);
        }
//        if (previousOffset < 0) {
//            //for PsiSwitchStatements, all "case:" and PsiElements are
//            //most common case for Sketchlink anchors inside a Method
//            previousOffset = offsetOfPreviousSiblingIfDocComment(target);
//        }
        int offset = (target != null) ? target.getTextRange().getStartOffset() : element.getTextRange().getEndOffset();
        @Nullable Document document = PsiDocumentManager.getInstance(element.getProject()).getDocument(element.getContainingFile());
        if (document == null || target == null)
            return offset;

        //if a DocComment has been found as the previous Sibling, check whether it's mine or belongs to the previous code line
        @Nullable PsiElement ele = getPreviousSiblingSkipWhiteSpace(target);
        @Nullable PsiElement ele2 = getPreviousSiblingSkipWhiteSpace(ele);

        int targetLineStart = document.getLineNumber(target.getTextRange().getStartOffset());
        int targetLineEnd = document.getLineNumber(target.getTextRange().getEndOffset());

        int eleLine = ele != null ? document.getLineNumber(ele.getTextRange().getStartOffset()) : -1;
        int ele2Line = ele2 != null ? document.getLineNumber(ele2.getTextRange().getEndOffset()) : -1;
        if (ele instanceof PsiDocComment && eleLine != ele2Line) {
            return ele.getTextRange().getStartOffset();
        }
        //no DocComment exists, return the preferred location to create one
        //for,if, etc statements: should be in the same line as the closing ")"
        @Nullable PsiElement parenthesesChild = getChildWithLastClosingParenthese(target);
        if (parenthesesChild != null && preferCommentBefore) {
            targetLineStart = document.getLineNumber(parenthesesChild.getTextRange().getStartOffset());
        }
        PsiElement lastElement = PsiUtil.getElementAtOffset(target.getContainingFile(), document.getLineEndOffset(targetLineStart)).getPrevSibling();
        if (lastElement instanceof PsiDocComment) {
            return lastElement.getTextRange().getStartOffset();
        }
        lastElement = PsiUtil.getElementAtOffset(element.getContainingFile(), document.getLineEndOffset(targetLineEnd)).getPrevSibling();
        if (lastElement instanceof PsiDocComment && !preferCommentBefore) {
            return lastElement.getTextRange().getStartOffset(); //return
        }
        //use this to insert before the describing element
        //return offset;

        //check if there is a normal comment at the end of the line
        int offsetCommentTrailing = target.getTextRange().getEndOffset();
        PsiElement lastChild = target.getLastChild();
        if (lastChild instanceof PsiComment) {
            offsetCommentTrailing = lastChild.getTextRange().getStartOffset();
        }

        //return the end of the describing element
        return preferCommentBefore ? document.getLineEndOffset(targetLineStart) : offsetCommentTrailing;
        //target.getTextRange().getEndOffset();//lastElement.getTextRange().getEndOffset();
    }

    @NotNull
    public String[] getIdParts(@NotNull PsiDocTag docTag) {
        return getIdParts(getSketchlinkTagID(docTag), getSketchlinkTagMD5(docTag));
    }

    /**
     * Splits the Sketchlink ID into it's 2 or 3 parts.
     * Index 0: Type, Supposed length==1
     * Index 1: UUID of the Anchor, supposed length==
     * Index 2: if type=='c' and locally available, the md5 hash saved with the uuid, empty String otherwise
     *
     * @param sketchlinkTagID the id to parse
     * @param sketchlinkMD5   if local Code anchor, optionally, the md5 part.
     * @return an Array with length==3, if the data couldn't be parsed correctly, it contains only empty strings
     */
    @NotNull
    public String[] getIdParts(@Nullable String sketchlinkTagID, @Nullable String sketchlinkMD5) {
        if (sketchlinkTagID == null) {
            return new String[]{"", "", ""};
        }
        //valid example: c-3e3394a3-a3af-46ea-ae61-80918f470146 89d352acffbc8691730f9945b4f03a2e
        @NotNull String[] result = new String[]{"", "", ""};
        @NotNull String[] split = sketchlinkTagID.split("-|\n|\t");
        if (split.length != 6) {
            LOG.info("Found invalid sketchlink id format, id length doesn't match a known format: " + sketchlinkTagID);
            return result;
        }
        if (!sketchlinkTagID.toLowerCase().startsWith("c-") && sketchlinkMD5 != null && !sketchlinkMD5.isEmpty()) {
            //invalid parameter exception
            //LOG.warn("Found invalid sketchlink id format, has MD5 data but doesn't start with c-Type Identifier " + sketchlinkTagID + " md5 data:" + sketchlinkMD5);
            return result;
        }
        if (!split[0].equalsIgnoreCase("c") && !split[0].equalsIgnoreCase("s") && !split[0].equalsIgnoreCase("m") && !split[0].equalsIgnoreCase("u")) {
            //LOG.warn("Found invalid sketchlink id format, doesn't start with known Type Identifier " + sketchlinkTagID);
            //throw new InvalidParameterException("Found invalid sketchlink id format, doesn't start with identifer");
            return result;
        }
        try {
            UUID uuid = UUID.fromString(sketchlinkTagID.substring(2));
        } catch (IllegalArgumentException e) {
            //LOG.warn("Found invalid Sketchlink id"+sketchlinkTagID+" "+e.getMessage());

            return result;
        }
        if (sketchlinkMD5 != null && sketchlinkMD5.length() != 32) {
            @NotNull String[] splitMD5 = sketchlinkMD5.split("-|\n|\t");
            try {
                Hex.decodeHex(splitMD5[0].toCharArray());
            } catch (DecoderException e) {
                //LOG.warn(String.format("Found invalid MD5 for %s", sketchlinkTagID));
                return result;
            }
        }
        result[0] = split[0];
        result[1] = sketchlinkTagID.substring(2);
        result[2] = (sketchlinkMD5 != null) ? sketchlinkMD5 : "";
        return result;
    }

    /**
     * Splits the Sketchlink ID into it's 3 parts:
     * Index 0: Type
     * Index 1: MD5 Hash (for code anchors only), empty string for none code anchors
     * Index 2: UUID of the Link
     *
     * @param sketchlinkTagID the id to parse
     * @return always an array with length==3, if the id couldn't be parsed, the array contains partially empty strings
     */
    @NotNull
    @Deprecated
    private String[] getIdParts(@Nullable String sketchlinkTagID) {
        @NotNull String[] result = new String[]{"", "", ""};
        if (sketchlinkTagID == null)
            return result;
        //String[] split = sketchlinkTagID.split("-");
        //valid example: c-89d352acffbc8691730f9945b4f03a2e-3e3394a3-a3af-46ea-ae61-80918f470146
        @NotNull String[] split = sketchlinkTagID.split("-|\n|\t");
        if (split.length < 6 || split.length > 7) {
            LOG.info("Found invalid sketchlink id format, id length doesn't match a known format: " + sketchlinkTagID);
            return result;
        }
        if (split.length == 7 && !split[0].equalsIgnoreCase("c")) {
            LOG.warn("Found invalid sketchlink id format, doesn't start with c-Type Identifier " + sketchlinkTagID);
            return result;
        } else {
            if (split.length == 6 && !(split[0].equalsIgnoreCase("s") || split[0].equalsIgnoreCase("r") || split[0].equalsIgnoreCase("u"))) {
                LOG.warn("Found invalid sketchlink id format, doesn't start with known Type Identifier " + sketchlinkTagID);
                return result;
            }
        }

        @NotNull StringBuilder sb = new StringBuilder();
        for (int i = 1; i < split.length; i++) {
            try {
//                if (split[i].charAt(split[i].length()-1) == '\n' ) {
//                    if (split[i].length()<4) {
//                        return result;
//                    }
//                    char[] data = split[i].substring(0,split[i].length()-2).toCharArray();
//                    Hex.decodeHex(data);
//                    continue;
//                }
                Hex.decodeHex(split[i].trim().toCharArray());
            } catch (DecoderException e) {
                LOG.warn(String.format("Found invalid sketchlink id %s  in Part %d Partlength: %d EndChar: %d\n%s", sketchlinkTagID, i, split[i].length(), (int) split[i].charAt(split[i].length() - 1), e.getMessage()));
//                for (char val : split[i].toCharArray())
//                    System.out.println("Value: "+val+" "+(int)val);
//                System.out.println("\\n is "+ (int)'\n');
                return result;
            }
            if (i == 1 && split.length == 7) {
                continue;
            }
            sb.append(split[i]);
            sb.append("-");
        }
        sb.deleteCharAt(sb.length() - 1);

        result[0] = split[0];
        if (split.length == 7)
            result[1] = split[1]; //md5 part for code anchors
        result[2] = sb.toString();
        return result;
    }

    public void fixDuplicate(@NotNull final Pair<LocalLinkAnchor, LocalLinkAnchor> duplicatePair) {
        @Nullable final PsiDocTag docTagA = duplicatePair.first.getAnchorTag();
        @Nullable final PsiDocTag docTagB = duplicatePair.second.getAnchorTag();
        if (docTagA == null || docTagB == null || docTagA == docTagB) {
            return;
        }
        boolean redoA = !isValidID(docTagA);
        boolean redoB = !isValidID(docTagB);
        @Nullable String idA = getSketchlinkTagID(docTagA);
        @Nullable String idB = getSketchlinkTagID(docTagB);

        if (redoA || idA == null) {
            boolean fix = ConfirmationDialog.requestForConfirmation(VcsShowConfirmationOption.STATIC_SHOW_CONFIRMATION,
                    docTagA.getProject(),
                    "An invalid LinkAnchor has been found in File\n" + slToolBox.getRelativePath(docTagA) + "\n\nProblematic id: " + idA + "\n\nRepair it?", "Fix LinkAnchor", null, "Generate new ID", "Cancel");
            if (!fix) {
                return;
            }
            idA = "c-" + UUID.randomUUID().toString();
            setDocTagText(idA, docTagA);
            duplicatePair.first.setAnchorId(idA);
            redoA = false;
        }
        if (redoB || idB == null) {
            boolean fix = ConfirmationDialog.requestForConfirmation(VcsShowConfirmationOption.STATIC_SHOW_CONFIRMATION,
                    docTagA.getProject(),
                    "An invalid LinkAnchor has been found in File\n" + slToolBox.getRelativePath(docTagB) + "\n\nProblematic id: " + idB + "\n\nRepair it?", "Fix LinkAnchor", null, "Generate new ID", "Cancel");
            if (!fix) {
                return;
            }
            idB = "c-" + UUID.randomUUID().toString();
            setDocTagText(idB, docTagB);
            duplicatePair.second.setAnchorId(idB);
            redoB = false;
        }
        if (!idA.equals(idB)) {
            return;
        }
        @Nullable final String md5A = getSketchlinkTagMD5(docTagA);
        @Nullable final String md5B = getSketchlinkTagMD5(docTagB);
        @Nullable final String currentMd5A = slToolBox.calculateMD5(slToolBox.getRelativePath(docTagA));
        @Nullable final String currentMd5B = slToolBox.calculateMD5(slToolBox.getRelativePath(docTagB));
        if (currentMd5A == null || currentMd5B == null) {
            return; //can't fix
        }
        redoA = !currentMd5A.equals(md5A);
        redoB = !currentMd5B.equals(md5B);
        //System.out.println("SketchlinkTools#fixDuplicate " + redoA + " " + redoB + "\n" + md5A + "\t" + md5B + "\n" + currentMd5A + "\t" + currentMd5B);
        if (redoA && !redoB) {

            idA = "c-" + UUID.randomUUID().toString();
            setDocTagText(idA, docTagA);
            duplicatePair.first.setAnchorId(idA);
            return;
        } else {
            if (!redoA && redoB) {
                idB = "c-" + UUID.randomUUID().toString();
                setDocTagText(idB, docTagB);
                duplicatePair.second.setAnchorId(idB);
                return;
            }
        }
        @Nullable final SketchMetadataTypeC metaA = duplicatePair.first.getCurrentMetaData();
        @Nullable final SketchMetadataTypeC metaB = duplicatePair.second.getCurrentMetaData();
        @Nullable final String idBfinal = idB;
        @Nullable final String idAfinal = idA;
        @NotNull RunnableFuture<Boolean> fixFirstFuture = new RunnableFuture<Boolean>() {
            private boolean fixFirst = true;
            private boolean isDone = false;

            @Override
            public void run() {
                @Nullable PsiElement target = duplicatePair.first.getTargetElement();
                @Nullable Document documentA = PsiDocumentManager.getInstance(docTagA.getProject()).getDocument(docTagA.getContainingFile());
                int targetLineA = -1;
                int targetLineB = -1;
                @Nullable String exampleTextA = null;
                @Nullable String exampleTextB = null;
                if (target != null && documentA != null) {
                    targetLineA = documentA.getLineNumber(target.getTextOffset());
                    exampleTextA = documentA.getText(new TextRange(documentA.getLineStartOffset(targetLineA), documentA.getLineEndOffset(targetLineA)))
                            .trim().replaceAll("\t", " ");
                }
                @Nullable Document documentB = PsiDocumentManager.getInstance(docTagB.getProject()).getDocument(docTagB.getContainingFile());
                target = duplicatePair.second.getTargetElement();
                if (target != null && documentB != null) {
                    targetLineB = documentB.getLineNumber(target.getTextOffset());
                    exampleTextB = documentB.getText(new TextRange(documentB.getLineStartOffset(targetLineB), documentB.getLineEndOffset(targetLineB)))
                            .trim().replaceAll("\n", " ").replaceAll("\t", " ");
                }
                fixFirst = ConfirmationDialog.requestForConfirmation(VcsShowConfirmationOption.STATIC_SHOW_CONFIRMATION,
                        docTagA.getProject(),
                        "Two identical SketchAnchors found with ID=" + getSketchlinkTagID(docTagA)
                                + "\nThe previously known Anchor is in File: " + slToolBox.getRelativePath(docTagA)
                                + "\n\tLinking to an " + metaA.getTargetType() + (metaA.getTargetName() != null ? " named " + metaA.getTargetName() : "")
                                + " in Line: " + (targetLineA + 1)
                                + (exampleTextA != null ? "\nText: " + (exampleTextA.length() < 30 ? exampleTextA.substring(0, exampleTextA.length()) : exampleTextA) : "")
                                + "\n\nThe newly found Anchor is in File: " + slToolBox.getRelativePath(docTagB)
                                + "\n\tLinking to an " + metaB.getTargetType() + (metaB.getTargetName() != null ? " named " + metaB.getTargetName() : "")
                                + " in Line: " + (targetLineB + 1)
                                + (exampleTextB != null ? "\nText: " + (exampleTextB.length() < 30 ? exampleTextB.substring(0, exampleTextB.length()) : exampleTextB) : "")
                        , "Duplicate ID found", null, "Change the second location", "Change the first location"
                );
                if (!fixFirst) {
                    @NotNull String idA = "c-" + UUID.randomUUID().toString();
                    setDocTagText(idA, docTagA);
                    duplicatePair.first.setAnchorId(idA);

                    if (!currentMd5B.equals(md5B)) {

                        setDocTagText(idBfinal, docTagB);
                    }
                } else {
                    @NotNull String idB = "c-" + UUID.randomUUID().toString();
                    setDocTagText(idB, docTagB);
                    duplicatePair.second.setAnchorId(idB);
                    if (!currentMd5A.equals(md5A)) {
                        setDocTagText(idAfinal, docTagA);
                    }
                }
            }

            @Override
            public boolean cancel(boolean mayInterruptIfRunning) {
                return false;
            }

            @Override
            public boolean isCancelled() {
                return isDone;
            }

            @Override
            public boolean isDone() {
                return isDone;
            }

            @Override
            public Boolean get() throws InterruptedException, ExecutionException {
                return fixFirst;
            }

            @Override
            public Boolean get(long timeout, @NotNull TimeUnit unit) throws InterruptedException, ExecutionException, TimeoutException {
                return fixFirst;
            }
        };

        boolean fixFirst = false;
        UIUtil.invokeAndWaitIfNeeded(fixFirstFuture);
//        try {
//            fixFirst = fixFirstFuture.get();
//        } catch (InterruptedException | ExecutionException e) {
//            LOG.warn("User interaction for fixing duplicate AnchorIDs got interrupted.\n"+e.getMessage());
//        }
//        if (fixFirst) {
//            idA = "c-"+UUID.randomUUID().toString();
//            setDocTagText(idA, docTagA);
//            duplicatePair.first.setAnchorId(idA);
//            if (!currentMd5B.equals(md5B)) {
//                setDocTagText(idB, docTagB);
//            }
//        } else {
//            idB = "c-"+UUID.randomUUID().toString();
//            setDocTagText(idB, docTagB);
//            duplicatePair.second.setAnchorId(idB);
//            if (!currentMd5A.equals(md5A)) {
//                setDocTagText(idA, docTagA);
//            }
//        }

    }

    public void updateDocTagMD5(@NotNull final List<PsiDocTag> sketchlinkTags) {
        if (sketchlinkTags.size() == 0)
            return;
        @NotNull final Project project = sketchlinkTags.get(0).getProject();
        CommandProcessor.getInstance().executeCommand(project, new Runnable() {
            @Override
            public void run() {
                WriteCommandAction.runWriteCommandAction(null, new Runnable() {
                    @Override
                    public void run() {
                        Collections.sort(sketchlinkTags, new Comparator<PsiDocTag>() {
                            @Override
                            public int compare(@NotNull PsiDocTag o1, @NotNull PsiDocTag o2) {
                                return o2.getTextOffset() - o1.getTextOffset();
                            }
                        });

                        for (@NotNull PsiDocTag docTag : sketchlinkTags) {
                            @Nullable String sketchlinkTagID = getSketchlinkTagID(docTag);
                            if (sketchlinkTagID == null) {
                                continue;
                            }
                            setDocTagText(sketchlinkTagID, docTag);
                        }
                    }
                });
            }
        }, null, null);
    }

    /**
     * Updates a doctag with the supplied replacement ID, auto calculates and updates the md5 part
     */
    public void setDocTagText(@NotNull final String replacementID, @NotNull final PsiDocTag docTag) {
        @NotNull final Project project = docTag.getProject();

        CommandProcessor.getInstance().executeCommand(project, new Runnable() {
                            @Override
                            public void run() {
                                WriteCommandAction.runWriteCommandAction(null, new Runnable() {
                                    @Override
                                    public void run() {
                                        @Nullable Document document = PsiDocumentManager.getInstance(project).getDocument(docTag.getContainingFile());
                                        if (document == null || !docTag.isValid())
                                            return;
                                        //behind the tag NAME
                                        int startOffset = docTag.getNameElement().getTextRange().getEndOffset() + 1;
                                        //offset of the line, the NAME is in
                                        int lineEndOffset = document.getLineEndOffset(document.getLineNumber(startOffset));
                                        //offset of the end of the whole doc Comment. needed for single line comments
                                        int docCommentEndOffset = docTag.getParent().getTextRange().getEndOffset() - 2;
                                        //the endoffset to use
                                        int endOffset = docCommentEndOffset < lineEndOffset ? docCommentEndOffset : lineEndOffset;
                                        if (!docTag.isValid()) {
                                            return;
                                        }
                                        document.replaceString(docTag.getTextRange().getStartOffset(), endOffset,
                                                "@" + slApplicationState.SKETCHLINK_TAGNAME + " " + replacementID + " " + slToolBox.calculateMD5(slToolBox.getRelativePath(docTag)) + " ");
                                        @Nullable LocalLinkAnchor localSketchAnchor = SketchlinkManager.getAnchorOf(docTag);
                                        if (localSketchAnchor != null) {
                                            new ChangeArtifactMetaDataRequestBuilder(localSketchAnchor).queue(docTag.getProject(), null, false);
                                        }
                                    }
                                });
                            }
                        }, "Repair Sketchanchor in File", null, UndoConfirmationPolicy.REQUEST_CONFIRMATION
                );


    }
}
