/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.tools;

import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Pair;
import com.kitfox.svg.*;
import com.kitfox.svg.animation.AnimationElement;
import com.kitfox.svg.xml.StyleAttribute;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.SketchlinkManager;
import plg.slCurrentProjectSettings;
import plg.ui.SketchlinkColors;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.Stack;

/**
 * Created by Peter on 13.03.14.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class SVGTools {
    private static final Logger LOG = Logger.getInstance(SVGTools.class);


    @NotNull
    public static List<String> getShapeElementAnchors(@NotNull SVGDiagram diagram) {
        SVGRoot root = diagram.getRoot();
        @NotNull Stack<SVGElement> stack = new Stack<>();
        stack.push(root);
        @NotNull List<String> result = new ArrayList<>();
        while (!stack.isEmpty()) {
            SVGElement element = stack.pop();
            if (element instanceof Group) {
                for (Object o : element.getChildren(null)) {
                    if (o instanceof SVGElement) {
                        stack.push((SVGElement) o);
                    }
                }
                continue;
            }
            //if it's a circle, rectangle, polygon, etc.
            if (element instanceof ShapeElement) {
                String foundId = element.getId();
                if (SketchlinkManager.TOOLS.isValidID(foundId)) {
                    result.add(foundId);
                }
            }
        }
        return result;
    }

    /**
     * Gets the first element after the SVGRoot element which defines a group.
     * <p/>
     * returns null, if the root element has no child of type Group, the first group otherwise
     */
    @Nullable
    public static Group getTopGroup(@NotNull SVGDiagram svgDiagram) {
        SVGRoot root = svgDiagram.getRoot();
        for (Object o : root.getChildren(null)) {
            if (o instanceof Group) {
                return (Group) o;
            }
        }
        return null;

    }

    /**
     * Gets the datatype for the SVGElement or it's parent, if the datatype is not defined or empty for the element
     *
     * @param element  the element of a svg-file to extract data from
     * @param datatype the type of data to extract
     * @return Stringvalue of the Datatype, null if not available
     */
    @NotNull
    public static String getDataOf(@Nullable SVGElement element, @NotNull SVGData datatype) {
        @Nullable SVGElement svgElement = element;
        while (svgElement != null) {
            StyleAttribute attribute = svgElement.getPresAbsolute(datatype.getPresentationName());

            if (attribute != null && attribute.getStringValue() != null && !attribute.getStringValue().trim().equals("")) {
                //System.out.println("svgtools getdataof "+datatype.getPresentationName()+ " "+element.getTagName()+" "+element.getId());
                return attribute.getStringValue();
            }
            svgElement = svgElement.getParent();
        }
        //System.out.println("svgtools getdataof "+datatype.getPresentationName()+ " "+element);
        return "";
    }

    @NotNull
    public static String getDataOf(@Nullable SVGElement element, @NotNull SVGDrawing drawingType) {
        @Nullable SVGElement svgElement = element;
        while (svgElement != null) {
            StyleAttribute attribute = svgElement.getStyleAbsolute(drawingType.getStyleName());

            if (attribute != null && attribute.getStringValue() != null && !attribute.getStringValue().trim().equals("")) {
                //System.out.println("svgtools getdataof "+datatype.getPresentationName()+ " "+element.getTagName()+" "+element.getId());


                return attribute.getStringValue();
            } else {
                attribute = svgElement.getPresAbsolute(drawingType.getStyleName());
                if (attribute !=null && attribute.getStringValue() !=null && !attribute.getStringValue().trim().isEmpty()) {
                    return attribute.getStringValue();
                }
            }
            svgElement = svgElement.getParent();
        }
        //System.out.println("svgtools getdataof "+datatype.getPresentationName()+ " "+element);
        return "";
    }


    public static void setDataOf(@NotNull SVGElement element, @NotNull SVGData datatype, @Nullable String value) {
        try {
            if (value != null) {
                // StyleAttribute attribute = element.getStyleAbsolute(datatype.getPresentationName());
                if (element.hasAttribute(datatype.getPresentationName(), AnimationElement.AT_AUTO)) {
                    element.setAttribute(datatype.getPresentationName(), AnimationElement.AT_AUTO, value);
                } else {
                    element.addAttribute(datatype.getPresentationName(), AnimationElement.AT_XML, value);
                }
            } else {
                element.removeAttribute(datatype.getPresentationName(), AnimationElement.AT_AUTO);
            }
        } catch (SVGElementException e) {
            LOG.error("Could not set Element Data for " + element + " in SVGTools#setDataOf \n" + e.getMessage());
        }
    }

    public static void setDataOf(@NotNull SVGElement element, @NotNull SVGDrawing drawingType, @Nullable String value) {
        try {
            if (value != null) {
                //StyleAttribute attribute = element.getStyleAbsolute(drawingType.getStyleName());
                if (element.hasAttribute(drawingType.getStyleName(), AnimationElement.AT_AUTO)) {
                    element.setAttribute(drawingType.getStyleName(), AnimationElement.AT_AUTO, value);
                } else {
                    element.addAttribute(drawingType.getStyleName(), AnimationElement.AT_XML, value);
                }
            } else {
                element.removeAttribute(drawingType.getStyleName(), AnimationElement.AT_AUTO);
            }
        } catch (SVGElementException e) {
            LOG.error("Could not set Element Data for " + element + " in SVGTools#setDataOf \n" + e.getMessage());
        }
    }

    /**
     * Returns the Color currently specified in slColormanager and indexed by type as a Pair of Strings usable as values in svg tags.
     * The first Element is the RGB Color in hexadecimal/html format, the second is the String representation of the opacity with a value between 0 and 1
     */
    @NotNull
    public static Pair<String, String> getColorCodeAsHTML(@NotNull SketchlinkColors type, @NotNull Project project) {
        @NotNull String color2 = slCurrentProjectSettings.getInstance(project).getColorHTMLRGBA(type);
        float alpha = 1.f * Integer.valueOf(color2.substring(7, 9), 16) / 255;
        return Pair.create(color2.substring(0, 7), String.valueOf(alpha));
    }


    private static void setShapeElementToColor(@NotNull ShapeElement element, @NotNull Pair<String, String> strokeColorAndOpacity, @NotNull Pair<String, String> fillColorAndOpacity) {
        setDataOf(element, SVGDrawing.STROKE, strokeColorAndOpacity.first);
        setDataOf(element, SVGDrawing.STROKE_OPACITY, strokeColorAndOpacity.second);
        setDataOf(element, SVGDrawing.FILL, fillColorAndOpacity.first);
    }

    /**
     * Sets the supplied element to highlight-colors
     */
    public static void ShapeElementHighlightedColors(@NotNull ShapeElement element, @NotNull Project project) {
        String foundId = element.getId();
        if (foundId==null) {
            System.out.println("Tried highlighting element without id");
            return;
        }
        try {
            float width = element.getPresAbsolute("width").getFloatValue(); //NON-NLS
            @NotNull Pair<String, String> colorCode = getColorCodeAsHTML(SketchlinkColors.PREVIEW_HIGHLIGHTED_FILL, project);
            setShapeElementToColor(element, getColorCodeAsHTML(SketchlinkColors.PREVIEW_HIGHLIGHTED_STROKE, project), colorCode);

            if (!foundId.startsWith("s")) { //NON-NLS
                setDataOf(element, SVGDrawing.FILL_OPACITY, colorCode.second);
            } else {
                setDataOf(element, SVGDrawing.STROKE, colorCode.first);
            }
            if (width <= 1.) {
                element.setAttribute("stroke-width", AnimationElement.AT_AUTO, "0.005"); //NON-NLS
            } else {
                element.setAttribute("stroke-width", AnimationElement.AT_AUTO, "3"); //NON-NLS
            }
        } catch (SVGElementException e) {
            LOG.debug("Setting Element with id " + foundId, e);
        }
    }

    /**
     * Sets the supplied element to linked-colors
     */
    public static void ShapeElementLinkedColors(@NotNull ShapeElement element, @NotNull Project project) {
        String foundId = element.getId();
        if (foundId == null) {
            return;
        }
        try {
            float width = element.getPresAbsolute("width").getFloatValue(); //NON-NLS
            @NotNull Pair<String, String> colorCode = getColorCodeAsHTML(SketchlinkColors.PREVIEW_LINKED_FILL, project);
            setShapeElementToColor(element, getColorCodeAsHTML(SketchlinkColors.PREVIEW_LINKED_STROKE, project), colorCode);
            if (width <= 1.) {
                element.setAttribute("stroke-width", AnimationElement.AT_AUTO, "0.005"); //NON-NLS
            } else {
                element.setAttribute("stroke-width", AnimationElement.AT_AUTO, "3"); //NON-NLS
            }
            if (!foundId.startsWith("s")) { //NON-NLS
                setDataOf(element, SVGDrawing.FILL_OPACITY, colorCode.second);
            } else {
                setDataOf(element, SVGDrawing.STROKE, colorCode.first);
            }

        } catch (SVGElementException e) {
            LOG.debug("Setting Element with id " + foundId, e);
        }
    }

    /**
     * Sets the supplied element to non linked, additional colors
     */
    public static void ShapeElementNormalColors(@NotNull ShapeElement element, @NotNull Project project) {
        String foundId = element.getId();
        if (foundId == null) {
            return;
        }
        try {
            float width = element.getPresAbsolute("width").getFloatValue(); //NON-NLS
            @NotNull Pair<String, String> colorCode = getColorCodeAsHTML(SketchlinkColors.PREVIEW_NORMAL_FILL, project);
            setShapeElementToColor(element, getColorCodeAsHTML(SketchlinkColors.PREVIEW_NORMAL_STROKE, project), colorCode);
            if (!foundId.startsWith("s")) { //NON-NLS
                setDataOf(element, SVGDrawing.FILL_OPACITY, colorCode.second);
            }
            if (width <= 1.) {
                element.setAttribute("stroke-width", AnimationElement.AT_AUTO, "0.005"); //NON-NLS
            } else {
                element.setAttribute("stroke-width", AnimationElement.AT_AUTO, "3"); //NON-NLS
            }
        } catch (SVGElementException e) {
            LOG.debug("Setting Element with id " + foundId, e);
        }
    }

    /**
     * Prints the formatting of the given diagram to System.out. Useful for debugging, should not be used in final code
     */
    public static void printSvgToStdOut(@NotNull SVGDiagram diagram) {
        System.out.println("Diagram: " + diagram.getWidth() + " " + diagram.getHeight() + " " + diagram.getDeviceViewport());
        System.out.println("SvgUri: " + diagram.getXMLBase());
        SVGRoot root = diagram.getRoot();
        //System.out.println("Root: " + root.getBoundingBox() + " " + root.getNumChildren() + " " + root.getId());
        System.out.println(root.getClass().getName() + ": " + root.getTagName() + " w:" + root.getDeviceWidth() + " h:" + root.getDeviceHeight() + " ChildCount:" + root.getNumChildren());
        @NotNull List<Object> liste = new ArrayList<>();
        liste.add(root);
        Set set = root.getInlineAttributes();
        System.out.println("\tInline:");
        printSet(set, root, false);
        System.out.println("\tPresentation:");
        set = root.getPresentationAttributes();
        printSet(set, root, true);
        while (liste.size() > 0) {
            Object data = liste.get(0);
            liste.remove(0); //unperformand
            List children;
            if (data instanceof Group) {
                children = ((Group) data).getChildren(null);

            } else {
                continue;
            }
            for (Object object : children) {
                liste.add(object);
                SVGElement child;
                if (!(object instanceof SVGElement)) {
                    System.out.println("Object not a SvgElement: " + object.getClass() + " toString:" + object.toString());
                    continue;
                }
                child = (SVGElement) object;

                boolean found = false;
                if (child instanceof ShapeElement) {
                    @NotNull ShapeElement shapeElement = (ShapeElement) child;

                    try {
                        System.out.println(child.getClass().getName() + ": " + shapeElement.getTagName()
                                        + " w:" + shapeElement.getShape().getBounds().width
                                        + " h:" + shapeElement.getShape().getBounds().height
                                        + " Children:" + shapeElement.getNumChildren()
                                // + " Bounding: " + shapeElement.getBoundingBox().getBounds()

                                // +"\nURI: "+shapeElement.getXMLBase().toASCIIString()
                        );
                    } catch (Exception e) {
                        LOG.error(e);
                    }
                    found = true;
                }
                if (child instanceof ImageSVG) {
                    @NotNull ImageSVG imageSVG = (ImageSVG) child;
                    System.out.println(child.getClass().getName() + ": " + imageSVG.getTagName()
                                    + " w:" + imageSVG.getWidth()
                                    + " h:" + imageSVG.getHeight()
                                    + " Children:" + imageSVG.getNumChildren()
                                    + " BoundingBox: " + imageSVG.getBoundingBox().getBounds()
                            // +"\nURI: "+imageSVG.getXMLBase().toASCIIString()
                    );
                    found = true;
                }
                if (!found) {
                    System.out.println("UNKNOWN Child: " + child.getClass() + " " + child.toString());
                }
                set = child.getInlineAttributes();
                System.out.println("\tInline:");
                printSet(set, child, false);
                System.out.println("\tPresentation:");
                set = child.getPresentationAttributes();
                printSet(set, child, true);
            }
        }
    }

    private static void printSet(@Nullable Set set, @NotNull SVGElement element, boolean presentation) {
        if (set == null || set.size() == 0) {
            System.out.println("\t\tno Attributes");
            return;
        }
        for (Object o : set) {
            if (o instanceof String) {
                StyleAttribute attribute;
                if (presentation) {
                    attribute = element.getPresAbsolute(o.toString());
                } else {
                    attribute = element.getStyleAbsolute(o.toString());
                }
                System.out.println("\t\tTag: " + o + " Value: " + attribute.getStringValue());
                for (String string : attribute.getStringList()) {
                    System.out.println("\t\t\tString: " + string);
                }
            } else {
                System.out.println("Not String: " + o.getClass().getCanonicalName() + " -- " + o.toString());
            }
        }
    }

    @Nullable
    public static ImageSVG getImageSVG(@NotNull SVGDiagram diagram) {
        SVGRoot root = diagram.getRoot();
        @NotNull Stack<Object> liste = new Stack<>();
        liste.push(root);
        while (liste.size() > 0) {
            Object data = liste.pop();
            List children;
            if (data instanceof Group) {
                children = ((Group) data).getChildren(null);
            } else {
                continue;
            }
            for (Object object : children) {
                if (object instanceof ImageSVG) {
                    return (ImageSVG) object;
                }
                if (object instanceof Group) {
                    liste.push(object);
                }
            }
        }
        return null;
    }

    /**
     * XML Attributes
     */
    public enum SVGData {
        ANNOTATION {
            @NotNull
            String getPresentationName() {
                return "data:annotation";
            }
        },
        AUTHORS {
            @NotNull
            String getPresentationName() {
                return "data:authors";
            }
        },
        TIME {
            @NotNull
            String getPresentationName() {
                return "data:time";
            }
        },
        DATE {
            @NotNull
            String getPresentationName() {
                return "data:date";
            }
        },
        ID {
            @NotNull
            String getPresentationName() {
                return "id";
            }
        },
        X {
            @NotNull
            String getPresentationName() {
                return "x";
            }
        },
        Y {
            @NotNull
            String getPresentationName() {
                return "y";
            }
        },
        WIDTH {
            @NotNull
            String getPresentationName() {
                return "width";
            }
        },
        HEIGHT {
            @NotNull
            String getPresentationName() {
                return "height";
            }
        };

        /**
         * Returns the name of the StyleAttribute inside a svg-file
         *
         * @return String name of this type
         */
        @NotNull
        abstract String getPresentationName();
    }

    /**
     * Attributes in the style tag
     */
    public enum SVGDrawing {
        STROKE {
            @NotNull
            String getStyleName() {
                return "stroke";
            }
        },
        STROKE_OPACITY {
            @NotNull
            String getStyleName() {
                return "stroke-opacity";
            }
        },
        STROKE_WIDTH {
            @NotNull
            String getStyleName() {
                return "stroke-width";
            }
        },
        FILL {
            @NotNull
            String getStyleName() {
                return "fill";
            }
        },
        FILL_OPACITY {
            @NotNull
            String getStyleName() {
                return "fill-opacity";
            }
        },
        WIDTHstyle {
            @NotNull
            String getStyleName() {
                return "width";
            }
        },
        HEIGHTstyle {
            @NotNull
            String getStyleName() {
                return "height";
            }
        },
        Xstyle {
            @NotNull
            String getStyleName() {
                return "x";
            }
        },
        Ystyle {
            @NotNull
            String getStyleName() {
                return "y";
            }
        };

        /**
         * Returns the name of the StyeAttribute inside a svg-file, regarding to
         *
         * @return String name of this type
         */
        @NotNull
        abstract String getStyleName();
    }


}
