/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.tools;

import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.DialogWrapper;
import com.intellij.openapi.ui.MessageDialogBuilder;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.util.Pair;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.javadoc.PsiDocTag;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.LocalLinkAnchor;
import plg.sketchlinks.SketchlinkManager;
import plg.slApplicationState;

import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.RunnableFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * Created by Peter on 06.04.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class SketchIDRepair {

    public static FixMessageParameter showFixMessage(@NotNull final FixMessageParameter parameter) {
        if (!parameter.getMyDoNotAskOption().isToBeShown()) {
            parameter.setExitCode(parameter.getDefaultExitCode());
            return parameter;
        }
        @NotNull final RunnableFuture<FixMessageParameter> integerFuture = new RunnableFuture<FixMessageParameter>() {

            private boolean done = false;
            private MessageDialogBuilder.YesNoCancel yesNoCancel;

            @NotNull
            private FixMessageParameter returnValue = new FixMessageParameter(parameter.getMessage(), parameter.getMyDoNotAskOption(), parameter.defaultExitCode);

            public void run() {
                int result = 0;
                yesNoCancel = MessageDialogBuilder.yesNoCancel("Fixing Sketch IDs", parameter.getMessage());
                yesNoCancel.doNotAsk(returnValue.getMyDoNotAskOption());
                result = yesNoCancel.show();
                returnValue.setExitCode(result);
                if (!returnValue.getMyDoNotAskOption().isToBeShown()) {
                    returnValue.setDefaultExitCode(result);
                }
                done = true;
            }

            @Override
            public boolean cancel(boolean mayInterruptIfRunning) {
                return false;
            }

            @Override
            public boolean isCancelled() {
                return false;
            }

            @Override
            public boolean isDone() {
                return done;
            }

            @NotNull
            @Override
            public FixMessageParameter get() throws InterruptedException, ExecutionException {
                return returnValue;
            }

            @NotNull
            @Override
            public FixMessageParameter get(long timeout, @NotNull TimeUnit unit) throws InterruptedException, ExecutionException, TimeoutException {
                return returnValue;
            }
        };

        UIUtil.invokeLaterIfNeeded(integerFuture);
        int result = -1;
        try {
            return integerFuture.get();
        } catch (@NotNull InterruptedException | ExecutionException e) {
            e.printStackTrace();
        }
        return new FixMessageParameter(parameter.getMessage(), parameter.getMyDoNotAskOption(), parameter.defaultExitCode);

    }

    public static void RepairSketchAnchors(@NotNull List<LocalLinkAnchor> localSketchAnchors) {
        @NotNull Collection<PsiDocTag> result = new ArrayList<>(localSketchAnchors.size());
        for (@NotNull LocalLinkAnchor localSketchAnchor : localSketchAnchors) {
            @Nullable PsiDocTag docTag = localSketchAnchor.getAnchorTag();
            if (docTag == null || !docTag.isValid())
                continue;
            result.add(docTag);
        }
        RepairSketchAnchors(result);
    }

    public static void RepairSketchAnchors(@NotNull PsiFile file) {
        RepairSketchAnchors(PsiTreeUtil.findChildrenOfType(file, PsiDocTag.class));
    }

    public static void RepairSketchAnchors(@NotNull Collection<PsiDocTag> invalidDocTags) {
        Iterator<PsiDocTag> iterator = invalidDocTags.iterator();
        Project project = iterator.hasNext() ? iterator.next().getProject() : null;

        DialogWrapper.DoNotAskOption myDoNotAskOption;
        int exitCode = -1;

        myDoNotAskOption = new DialogWrapper.DoNotAskOption() {
            boolean show = true;
            int exitCode;

            @Override
            public boolean isToBeShown() {

                return show;
            }

            @Override
            public void setToBeShown(boolean value, int exitCode) {
                show = value;

                this.exitCode = exitCode;
            }

            @Override
            public boolean canBeHidden() {

                return true;
            }

            @Override
            public boolean shouldSaveOptionsOnCancel() {

                return false;
            }

            @NotNull
            @Override
            public String getDoNotShowMessage() {
                return "Apply to all";
            }
        };

        @NotNull CheckAnchors checkAnchors = new CheckAnchors(invalidDocTags);

        @NotNull Map<PsiDocTag, CheckAnchors.ErrorType> faults = checkAnchors.getNormalFaults();
        int faultSize = checkAnchors.getBadAnchors().size();
        if (faultSize == 0) {
            if (project != null)
                slGuiTools.ShowNotificationAtBottom("No problematic IDs detected.", project);
            return;
        }

        int errorNo = 1;
        FixMessageParameter param = new FixMessageParameter("", myDoNotAskOption, exitCode);
        @NotNull String baseText1 = "Error ";
        @NotNull String baseText2 = " of " + faultSize;
        @NotNull StringBuilder sb = new StringBuilder(baseText1);
        @NotNull StringBuilder idText = new StringBuilder("");

        for (@NotNull PsiDocTag docTag : faults.keySet()) {
            sb.setLength(baseText1.length());
            idText.setLength(0);
            sb.append(errorNo++);
            sb.append(baseText2);
            sb.append("\nPrevious ID: ");
            for (@NotNull PsiElement element : docTag.getDataElements()) {
                @NotNull String[] split = element.getText().split(" |\n|\t");
                if (element.getText().contains(" ") && split.length > 0) {
                    idText.append(split[0]);
                    break;
                } else {
                    idText.append(element.getText());
                }
            }
            sb.append(idText);
            sb.append("\nPrevious MD5: ");
            @NotNull String[] split = docTag.getText().split(" |\n|\t");
            for (int i = 2; i < split.length; i++) {
                if (split[i - 2].equals("@" + slApplicationState.SKETCHLINK_TAGNAME)) {
                    sb.append(split[i]);
                    break;
                }
            }
            //sb.append(SketchlinkManager.TOOLS.getSketchlinkTagMD5(docTag));
            sb.append("\n\nDetected Problem:\n\t");
            sb.append(faults.get(docTag).getDescription());
            param.setMessage(sb.toString());
            param = showFixMessage(param);

            if (param.getExitCode() == Messages.YES) {
                checkAnchors.fixId(docTag);
            } else {
                if (param.getExitCode() == Messages.CANCEL) {
                    return;
                }
            }
        }
        //ignore duplicates

//        Map<String, Set<PsiDocTag>> doublettes = checkAnchors.getDoublettes2();
//        for (String indexID : doublettes.keySet()) {
//            Set<PsiDocTag> duplicateSet = doublettes.get(indexID);
//            for (PsiDocTag docTag : duplicateSet) {
//                LocalLinkAnchor anchor = SketchlinkManager.getInstance().addLocal(docTag);
//            }
//        }
//        faults = checkAnchors.getDoublettes();
//        for (PsiDocTag docTag : faults.keySet()) {
//            sb.setLength(baseText1.length());
//            sb.append(errorNo++);
//            sb.append(baseText2);
//            sb.append("Previous ID: ");
//            sb.append(SketchlinkManager.TOOLS.getSketchlinkTagID(docTag));
//            sb.append("\nPrevious MD5: ");
//            sb.append(SketchlinkManager.TOOLS.getSketchlinkTagMD5(docTag));
//            sb.append("\n\nDetected Problem:\n\t");
//            sb.append(faults.get(docTag).getDescription());
//            param.setMessage(sb.toString());
//            param = showFixMessage(param);
//            if (param.getExitCode() == Messages.YES) {
//                checkAnchors.fixId(docTag);
//            } else {
//                if (param.getExitCode() == Messages.CANCEL) {
//                    return;
//                }
//            }
//
//        }
        checkAnchors.runFix();
    }

    public static class FixMessageParameter {
        private String message;
        private DialogWrapper.DoNotAskOption myDoNotAskOption;
        private int defaultExitCode;
        private int exitCode;

        public FixMessageParameter(@NotNull String message, @NotNull DialogWrapper.DoNotAskOption doNotAskOption, int defaultExitCode) {
            this.message = message;
            this.myDoNotAskOption = doNotAskOption;
            this.defaultExitCode = defaultExitCode;
            exitCode = Messages.CANCEL;
        }

        @NotNull
        public String getMessage() {
            return message;
        }

        public void setMessage(@NotNull String message) {
            this.message = message;
        }

        @NotNull
        public DialogWrapper.DoNotAskOption getMyDoNotAskOption() {
            return myDoNotAskOption;
        }

        public void setMyDoNotAskOption(@NotNull DialogWrapper.DoNotAskOption myDoNotAskOption) {
            this.myDoNotAskOption = myDoNotAskOption;
        }

        public int getExitCode() {
            return exitCode;
        }

        public void setExitCode(int exitCode) {
            this.exitCode = exitCode;
        }

        public int getDefaultExitCode() {
            return defaultExitCode;
        }

        public void setDefaultExitCode(int defaultExitCode) {
            this.defaultExitCode = defaultExitCode;
        }
    }

    private static class CheckAnchors {
        final private Map<String, PsiDocTag> good = new HashMap<>();
        final private Map<PsiDocTag, ErrorType> faulty = new HashMap<>();
        final private Map<PsiDocTag, String> repair = new HashMap<>();

        public CheckAnchors(@NotNull Collection<PsiDocTag> docTags) {
            @Nullable String anchorId = null;
            for (@NotNull PsiDocTag docTag : docTags) {
                if (!docTag.getName().equals(slApplicationState.SKETCHLINK_TAGNAME)) {
                    continue;
                }
                checkAnchor(docTag);

            }
            for (@NotNull PsiDocTag docTag : faulty.keySet()) {
                if (faulty.get(docTag).equals(ErrorType.DOUBLE)) {
                    anchorId = SketchlinkManager.TOOLS.getSketchlinkTagID(docTag);
                    good.remove(anchorId);
                }
            }
        }

        private boolean checkAnchor(@NotNull PsiDocTag docTag) {
            @Nullable String anchorId = SketchlinkManager.TOOLS.getSketchlinkTagID(docTag);
//            if (anchorId == null) { //empty id
//                faulty.put(docTag, ErrorType.EMPTY);
//                return false;
//            }
            @NotNull String[] parts = SketchlinkManager.TOOLS.getIdParts(docTag);

            try {
                UUID uuid = UUID.fromString(parts[1]);
            } catch (IllegalArgumentException ignore) {
                faulty.put(docTag, ErrorType.UUID);
                return false;
            }
            if (good.containsKey(anchorId)) { //id already known
                faulty.put(docTag, ErrorType.DOUBLE);
                //leave the double there, in case there are more copies
                faulty.put(good.get(anchorId), ErrorType.DOUBLE);
                return false;
            }
            if (!SketchlinkManager.TOOLS.isValidID(docTag)) { //invalid id structure
                faulty.put(docTag, ErrorType.STRUCTURE);
                return false;
            }
            if (!parts[2].equals(slToolBox.calculateMD5(slToolBox.getRelativePath(docTag)))) { //invalid path md5 sum
                faulty.put(docTag, ErrorType.MD5);
                return false;
            }
            if (!parts[0].equals("c")) { //not a code anchor
                faulty.put(docTag, ErrorType.TYPE);
                return false;
            }

            good.put(anchorId, docTag);
            return true;
        }

        public boolean hasFaults() {
            return !faulty.isEmpty();
        }

        @NotNull
        public Collection<PsiDocTag> getGoodAnchors() {
            return good.values();
        }

        @NotNull
        public Collection<PsiDocTag> getBadAnchors() {
            return faulty.keySet();
        }

        @NotNull
        public Map<PsiDocTag, ErrorType> getNormalFaults() {
            @NotNull Map<PsiDocTag, ErrorType> result = new HashMap<>(faulty.size());
            for (PsiDocTag docTag : faulty.keySet()) {
                if (faulty.get(docTag).equals(ErrorType.DOUBLE))
                    continue;
                result.put(docTag, faulty.get(docTag));
            }
            return result;
        }

        /**
         * Gets all DocTags with duplicate IDs. Check getNormalFaults first, in case a normal repair
         * generates new Doubles
         */
        @NotNull
        public Map<PsiDocTag, ErrorType> getDoublettes() {
            @NotNull Map<PsiDocTag, ErrorType> result = new HashMap<>(faulty.size());

            for (PsiDocTag docTag : faulty.keySet()) {
                if (!faulty.get(docTag).equals(ErrorType.DOUBLE)) {

                    continue;
                }
                result.put(docTag, faulty.get(docTag));
            }

            return result;
        }

        /**
         * Gets all DocTags with duplicate IDs.
         * <p/>
         * (not used for now, since SketchlinkTools.fixDuplicates expects LocalSketchANchors as input)
         * TODO: Change fixDuplicates to work with raw DocTags
         *
         * @return
         */
        @NotNull
        public Map<String, Set<PsiDocTag>> getDoublettes2() {
            @NotNull Map<String, Set<PsiDocTag>> result = new HashMap<>(faulty.size());
            for (@NotNull PsiDocTag docTag : faulty.keySet()) {
                if (!faulty.get(docTag).equals(ErrorType.DOUBLE)) {
                    continue;
                }
                @Nullable String id = SketchlinkManager.TOOLS.getSketchlinkTagID(docTag);
                if (id == null) {
                    Logger.getInstance(SketchIDRepair.class).warn("Found a duplette with invalid id, please fix checking algorithm");
                    continue;
                }
                if (!result.containsKey(id)) {
                    result.put(id, new HashSet<PsiDocTag>(3));
                }
                result.get(id).add(docTag);
            }
            return result;
        }

        /**
         * Adds the following docTag to the fix-functionality.
         *
         * @param docTag the docTag to fix
         * @return the proposed new ID
         */
        @NotNull
        public String fixId(@NotNull PsiDocTag docTag) {
            @Nullable String oldId = SketchlinkManager.TOOLS.getSketchlinkTagID(docTag);
            if (oldId == null)
                oldId = "";
            @NotNull StringBuilder sb = new StringBuilder("c-");

            @NotNull String[] parts = SketchlinkManager.TOOLS.getIdParts(docTag);
            ErrorType myFault = faulty.get(docTag);
            if (myFault == null) {
                myFault = ErrorType.STRUCTURE;
            }
            UUID uuid = UUID.randomUUID();
            String result;
            if (myFault == ErrorType.MD5 || myFault == ErrorType.TYPE) {
                sb.append(parts[1]);
            } else {
                sb.append(UUID.randomUUID());
            }
//            sb.append(" ");
//            sb.append(slToolBox.calculateMD5(slToolBox.getRelativePath(docTag))); //always recalculate md5 part

            repair.put(docTag, sb.toString());
            good.put(sb.toString(), docTag);
            return sb.toString();
        }

        /**
         * Executes all proposed fixes. The Ids returned from fixId become valid after this message ran.
         */
        public void runFix() {
            if (!repair.keySet().iterator().hasNext()) {
                return;
            }


            //final Project project = repair.keySet().iterator().next().getProject();
            @NotNull final List<Pair<PsiDocTag, String>> reversedOffsets = new ArrayList<>(repair.size());
            for (PsiDocTag docTag : repair.keySet()) {
                reversedOffsets.add(new Pair<>(docTag, repair.get(docTag)));
            }
            Collections.sort(reversedOffsets, new Comparator<Pair<PsiDocTag, String>>() {
                @Override
                public int compare(@NotNull Pair<PsiDocTag, String> o1, @NotNull Pair<PsiDocTag, String> o2) {
                    if (o1.equals(o2))
                        return 0;
                    //descending order
                    return o2.first.getTextRange().getStartOffset() - o1.first.getTextRange().getStartOffset();

                }
            });
            for (@NotNull Pair<PsiDocTag, String> reversedOffset : reversedOffsets) {
                SketchlinkManager.TOOLS.setDocTagText(reversedOffset.second, reversedOffset.first);
            }


//
//            CommandProcessor.getInstance().executeCommand(project,
//                    new Runnable() {
//                        @Override
//                        public void run() {
//                            WriteCommandAction.runWriteCommandAction(null, new Runnable() {
//                                @Override
//                                public void run() {
//
//                                    for (Pair<PsiDocTag, String> pair : reversedOffsets) {
//                                        Document document = PsiDocumentManager.getInstance(project).getDocument(pair.first.getContainingFile());
//                                        if (document == null)
//                                            continue;
//                                        int startOffset = pair.first.getNameElement().getTextRange().getEndOffset() + 1;
//                                        int lineEndOffset = document.getLineEndOffset(document.getLineNumber(startOffset));
//                                        int docCommentEndOffset = pair.first.getParent().getTextRange().getEndOffset() - 2;
//                                        int endOffset = docCommentEndOffset < lineEndOffset ? docCommentEndOffset : lineEndOffset;
//                                        document.replaceString(startOffset, endOffset, pair.second);
//
//                                    }
//                                    SketchlinkManager.getInstance().recheckLocalIndizes();
//                                }
//                            });
//                        }
//                    }, "Repair Sketchanchors in File", null, UndoConfirmationPolicy.REQUEST_CONFIRMATION
//            );
        }

        protected enum ErrorType {
            MD5 {
                @NotNull
                String getDescription() {
                    return "The md5 hash of the saved path doesn't match the actual file path";
                }
            }, TYPE {
                @NotNull
                String getDescription() {
                    return "The type of this anchor is not a code anchor";
                }
            }, UUID {
                @NotNull
                String getDescription() {
                    return "The UUID part has wrong formatting";
                }
            }, DOUBLE {
                @NotNull
                String getDescription() {
                    return "An Anchor with this id already exists";
                }
            }, EMPTY {
                @NotNull
                String getDescription() {
                    return "The Anchor has no detectable id";
                }
            }, STRUCTURE {
                @NotNull
                String getDescription() {
                    return "The Structure of this anchor is invalid";
                }
            };

            @NotNull
            abstract String getDescription();
        }
    }
}
