/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.tools;


import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.mime.MultipartEntityBuilder;
import org.apache.http.entity.mime.content.StringBody;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.util.EntityUtils;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.slProjectComp;
import plg.ui.preview.svgwrapper.SVGDiagram2XML;
import plg.ui.preview.svgwrapper.SVGDiagramEx;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * Access the Sketchlink Server through its REST api to upload a new sketch or an existing sketch after changes.
 * Uses Apache's HttpClient implementation, tested with the synchronous implementation.
 *
 * @author Peter Schmitz
 * @version 1.0
 */
public class SketchlinkServerRESTUtil {
    public static final int DEFAULT_PORT = 8080;


    @org.jetbrains.annotations.NotNull
    private static String sketchUploadPostfix = "sketches";
    @org.jetbrains.annotations.NotNull
    private static String sketchSVGDataPostfix = "sketches/svg";


    @Nullable
    public static String uploadSketchAndGetResponseBody(
            @NotNull String serverUrl, @NotNull File file, @NotNull String title,
            @NotNull String authors, @Nullable String annotation, @Nullable String artifact_id) throws IOException {

        @org.jetbrains.annotations.NotNull String[] response = uploadSketch(serverUrl, file, title, authors, annotation, artifact_id);
        if ("200".equals(response[2]))
            return response[0]; //should contain sketchID
        else
            return null;
    }


    /**
     * Uploads a file to the Sketchlink Server using its REST API in a multipart http request.
     *
     * @param serverUrl   the URL of the server, {e.g. "http://frodo.uni-trier:8080/"}
     * @param file        the file of the sketch, must be either jpg or png
     * @param title       the title for the sketch
     * @param authors     the authors of the sketch
     * @param annotation  the annotation of the sketch
     * @param artifact_id the id of the artifact linked to the whole sketch, optional
     * @return an array of length 3 with the following contents:
     * 0: the id of the sketch if successful (==response message body), or the message body of the response in general
     * 1: the status line of the response;
     * 2: the http status code of the response ("200"==success)
     * @throws java.io.IOException
     */
    @NotNull
    public static String[] uploadSketch(
            @NotNull String serverUrl, @NotNull File file, @NotNull String title,
            @NotNull String authors, @Nullable String annotation, @Nullable String artifact_id) throws IOException {
        @NotNull String uploadURL = buildUploadURL(serverUrl, true, sketchUploadPostfix);
        @NotNull StringBody titleBody = new StringBody(title, ContentType.TEXT_PLAIN);
        @NotNull StringBody authorBody = new StringBody(authors, ContentType.TEXT_PLAIN);
        @NotNull StringBody annotationBody = new StringBody(annotation != null ? annotation : "", ContentType.TEXT_PLAIN);
        @NotNull StringBody artifact_id_Body = new StringBody(artifact_id != null ? artifact_id : "", ContentType.TEXT_PLAIN);
        ContentType contentType;
        if (file.getName().toLowerCase().endsWith(".jpg") || file.getName().toLowerCase().endsWith(".jpeg")) {
            contentType = ContentType.create("image/jpeg");
        } else {
            if (file.getName().toLowerCase().endsWith(".png")) {
                contentType = ContentType.create("image/png");
            } else {
                contentType = ContentType.DEFAULT_BINARY;
            }
        }
        HttpEntity reqEntity = MultipartEntityBuilder.create()
                .addPart("title", titleBody)
                .addPart("authors", authorBody)
                .addPart("annotation", annotationBody)
                .addPart("artifact-id", artifact_id_Body)
                .addBinaryBody("image", file, contentType, file.getName())
                .build();
        return POSTrequest(uploadURL, reqEntity);
    }


    @NotNull
    private static String readResponseBody(@NotNull HttpResponse response) throws IOException {
        @NotNull StringBuilder sb = new StringBuilder();
        @NotNull BufferedReader reader =
                new BufferedReader(new InputStreamReader(response.getEntity().getContent()));

        String line;
        while ((line = reader.readLine()) != null) {
            sb.append(line);
        }
        return sb.toString();
    }

    @NotNull
    protected static String[] POSTrequest(@NotNull String uploadURL, @NotNull HttpEntity reqEntity) throws IOException {

        @NotNull String[] result = new String[3];

        //automatic resource management, implicit httpClient.close() as "finally"
        try (CloseableHttpClient httpClient = HttpClients.createDefault()) {
            @NotNull HttpPost httppost = new HttpPost(uploadURL);
            httppost.setEntity(reqEntity);
            try (CloseableHttpResponse response = httpClient.execute(httppost)) {
//                System.out.println("----------------------------------------");
//                System.out.println(response.getStatusLine());
                result[1] = response.getStatusLine().toString();
                result[2] = String.valueOf(response.getStatusLine().getStatusCode());
                result[0] = readResponseBody(response);
                EntityUtils.consume(response.getEntity());
            }
        }
        return result;
    }

    @NotNull
    private static String buildUploadURL(@NotNull String serverUrl, boolean shouldEndOnSlash, @NotNull String... postfix) {
        @NotNull StringBuilder sb = new StringBuilder();
        if (!serverUrl.toLowerCase().startsWith("http://")) {
            sb.append("http://");
        }
        sb.append(serverUrl);
        if (!serverUrl.endsWith("/")) {
            sb.append("/");
        }
        for (@NotNull String s : postfix) {
            sb.append(s);
            if (!s.endsWith("/"))
                sb.append("/");
        }
        if (!shouldEndOnSlash) {
            sb.deleteCharAt(sb.length() - 1);
        }
        URL url;
        String result = sb.toString();
        try {
            url = new URL(result);
            if (url.getPort() == -1) {
                url = new URL(url.getProtocol(), url.getHost(), DEFAULT_PORT, url.getFile());
            }
            result = url.toString();
        } catch (MalformedURLException ignored) {

        }
        return result;
    }

    /**
     * Uploads an existing sketch to the server after changing
     *
     * @param serverUrl the URL of the server, {e.g. "http://frodo.uni-trier:8080/"}
     * @param sketchID  the ID of the sketch
     * @param svg       the svg-xml content
     * @param changelog the changelog, as json serialized string
     * @return the response, 0: response body, 1: response line, 2: http response code (should be 200)
     * @throws java.io.IOException
     */
    @NotNull
    public static String[] updateSVG(
            @NotNull String serverUrl, @NotNull String sketchID, @NotNull String svg, @Nullable String changelog
    ) throws IOException {
        @NotNull String uploadURL = buildUploadURL(serverUrl, false, sketchSVGDataPostfix, sketchID);
        @NotNull StringBody svgBody = new StringBody(svg, ContentType.APPLICATION_SVG_XML);
        @NotNull StringBody changelogBody = new StringBody(changelog != null ? changelog : "{}", ContentType.TEXT_PLAIN);
        @NotNull StringBody sketchIDBody = new StringBody(sketchID, ContentType.TEXT_PLAIN);
        HttpEntity reqEntity = MultipartEntityBuilder.create()
                .addPart("svg", svgBody)
                .addPart("changelog", changelogBody)
                .addPart("id", sketchIDBody)
                .build();

        return POSTrequest(uploadURL, reqEntity);
    }


    @Nullable
    public static String[] updateSVG(
            @NotNull String serverUrl, @NotNull SVGDiagramEx diagram, @Nullable String changelog
    ) throws IOException {
        @Nullable String exportSVG = SVGDiagram2XML.export(diagram);
        if (exportSVG == null) {
            return null;
        }
        return updateSVG(serverUrl, diagram.getSketchID(), exportSVG, changelog);
    }


    @Nullable
    public static String[] updateSVG(
            @NotNull SVGDiagramEx diagram, @Nullable String changelog
    ) throws IOException {
        String serverURL = slProjectComp.getInstance(diagram.getProject()).getServerURL();
        if (serverURL.isEmpty()) {
            return null;
        }
        return updateSVG(serverURL, diagram, changelog);
    }

    /**
     * Uses REST to access the SVG on the server
     *
     * @param serverUrl the URL of the server, {e.g. "http://frodo.uni-trier:8080/"}
     * @param sketchID  the ID of the sketch that should be accessed
     * @return an String array of size 3 with the following contents:
     * 0: the svg -if statuscode="200"-, the message body of the response in general/otherwise
     * 1: the statusline of the response
     * 2: the statuscode of the response,
     * @throws java.io.IOException
     */
    @NotNull
    public static String[] getSVG(
            @NotNull String serverUrl, @NotNull String sketchID
    ) throws IOException {
        @NotNull String getURL = buildUploadURL(serverUrl, false, sketchSVGDataPostfix, sketchID);
        @NotNull HttpGet httpGet = new HttpGet(getURL);
        @NotNull String[] result = new String[3];
        try (CloseableHttpClient client = HttpClients.createDefault()) {
            try (CloseableHttpResponse response = client.execute(httpGet)) {
                result[1] = String.valueOf(response.getStatusLine());
                result[2] = String.valueOf(response.getStatusLine().getStatusCode());
                result[0] = readResponseBody(response);
                EntityUtils.consume(response.getEntity());
            }
        }
        return result;
    }


}
