/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.tools;

import com.intellij.find.FindUtil;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.fileEditor.FileEditorManager;
import com.intellij.openapi.fileEditor.OpenFileDescriptor;
import com.intellij.openapi.progress.ProgressIndicator;
import com.intellij.openapi.progress.util.ProgressIndicatorBase;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Computable;
import com.intellij.psi.NavigatablePsiElement;
import com.intellij.psi.PsiElement;
import com.intellij.psi.impl.search.PsiSearchHelperImpl;
import com.intellij.psi.javadoc.PsiDocTag;
import com.intellij.psi.search.*;
import com.intellij.psi.util.PsiUtilCore;
import com.intellij.util.containers.ArrayListSet;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.LocalLinkAnchor;
import plg.sketchlinks.SketchlinkManager;
import plg.slApplicationState;

import java.util.*;

/**
 * Created by Peter on 18.03.14.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class slSearch {

    private static long lastLookup = 0;
    private static long lastCompleteLookup = 0;
    private static boolean runCompleteLookupSoon = false;
    @NotNull
    private final Project project;


    public slSearch(@NotNull Project project) {
        this.project = project;
    }

    public static long getTimeOfLastSearch() {
        return lastLookup;
    }

    public static void browseToLocal(@NotNull final PsiElement element) {
        @NotNull Set<PsiElement> mySet = new ArrayListSet<>();
        mySet.add(element);
        browseToLocal(mySet);
    }

    public static void browseToLocal(@NotNull final Set<PsiElement> resultTargets) {
        if (resultTargets.isEmpty()) {
            return;
        }

        @NotNull final Project project = resultTargets.iterator().next().getProject();
        ApplicationManager.getApplication().invokeLater(new Runnable() {
            @Override
            public void run() {
                if (resultTargets.size() > 1) {
                    @NotNull List<PsiElement> sortList = new ArrayList<PsiElement>(resultTargets);
                    Collections.sort(sortList, new Comparator<PsiElement>() {
                        @Override
                        public int compare(@NotNull PsiElement o1, @NotNull PsiElement o2) {
                            return o1.getTextOffset() - o2.getTextOffset();
                        }
                    });
                    @NotNull PsiElement[] psiElements = sortList.toArray(new PsiElement[sortList.size()]);
                    FindUtil.showInUsageView(null,
                            psiElements,
                            slApplicationState.RESOURCE_BUNDLE.getString("search.usageview.sketchlink.has.found.multiple.linked.source.code.artifacts"), resultTargets.iterator().next().getProject());
                } else {
                    final PsiElement target = resultTargets.iterator().next().getNavigationElement();
                    if (target instanceof NavigatablePsiElement) {
                        ((NavigatablePsiElement) target).navigate(true);
                    } else {
                        //set the targetfile and targetoffset
                        @NotNull OpenFileDescriptor fileDescriptor = new OpenFileDescriptor(project, target.getContainingFile().getVirtualFile()
                                , target.getTextOffset());
                        //open a new texteditor
                        @Nullable Editor editor = FileEditorManager.getInstance(project).openTextEditor(fileDescriptor, true);
                    }
                }
            }
        });
    }

    public long getLastLookup() {
        return lastLookup;
    }

    @NotNull
    PsiElement[] findIdentifierInComment(@NotNull String identifier, @NotNull SearchScope searchScope) {
        @NotNull final List<PsiElement> results = new ArrayList<>();
        PsiSearchHelper helper = PsiSearchHelperImpl.SERVICE.getInstance(project);
        @NotNull TextOccurenceProcessor processor = new TextOccurenceProcessor() {
            @Override
            public boolean execute(PsiElement element, int offsetInElement) {
                synchronized (results) {
                    results.add(element);
                }
                return true;
            }
        };
        helper.processElementsWithWord(processor, searchScope, identifier, UsageSearchContext.IN_COMMENTS, true, false);
        synchronized (results) {
            return PsiUtilCore.toPsiElementArray(results);
        }
    }

    /**
     * Tries to search all PsiDocTags inside the current Project for artifactIDs supplied in artifactIDs. Found DocTags with valid IDs are
     * automatically added to SketchlinkManager and can be accessed there in the future.
     * <p/>
     * BUG: until sketchlinkmanager can handle and solve duplicate IDs, the resulting set is never larger than the requested artifactIDs.length
     * only the first found artifactID for duplicates is returned
     *
     * @param artifactIDs      An array containing the IDs to search for inside the project files.
     * @param includeLibraries if true, the project and all libraries are searched
     * @param showResultToUser if true, all results will be shown to the user, if only one result is found, the Element is opened and shown
     * @return all found LocalSketchAnchors. if it's empty, none have been found.
     */
    @NotNull
    public Set<LocalLinkAnchor> runSearchForIDs(@Nullable final String[] artifactIDs, boolean includeLibraries, final boolean showResultToUser) {
        @NotNull final ProgressIndicator progressIndicator = new ProgressIndicatorBase();
        lastLookup = System.currentTimeMillis();
        //PsiSearchHelper helper = PsiSearchHelperImpl.SERVICE.getInstance(project);
        @NotNull GlobalSearchScope searchScope = GlobalSearchScope.projectScope(project);
        if (includeLibraries) {
            searchScope = GlobalSearchScope.allScope(project);
        }
        progressIndicator.setFraction(0.1);
        progressIndicator.setText("Looking for Anchor in Files");
        @NotNull PsiElement[] psiElements = findIdentifierInComment("@" + slApplicationState.SKETCHLINK_TAGNAME, searchScope);
        @NotNull final List<PsiDocTag> hitList = new ArrayList<>();
        double step = 0.3 / psiElements.length;
        double fraction = 0.3;
        progressIndicator.setText("Filtering Locations");
        for (final PsiElement psiElement : psiElements) {
            if (psiElement instanceof PsiDocTag) {
                hitList.add((PsiDocTag) psiElement);
            }
            fraction += step;
            progressIndicator.setFraction(fraction);
        }


        return ApplicationManager.getApplication().runReadAction(new Computable<Set<LocalLinkAnchor>>() {
            @NotNull
            @Override
            public Set<LocalLinkAnchor> compute() {
                //use a set to check for duplicate supplied ids
                @Nullable Set<String> anchorSet = null;
                if (showResultToUser && artifactIDs != null) {
                    anchorSet = new HashSet<>(artifactIDs.length);
                    Collections.addAll(anchorSet, artifactIDs);
                }

                @NotNull final Set<LocalLinkAnchor> resultEntries = new HashSet<>(hitList.size());

                progressIndicator.setText("Checking Objects");
                progressIndicator.setFraction(0.7);

                @NotNull final Set<PsiElement> resultTargets = new HashSet<>(hitList.size());

                for (@NotNull PsiDocTag docTag : hitList) {
                    //always try to add found SketchAnchors, addLocal checks if the docTag is a sketchlink tag
                    @Nullable LocalLinkAnchor entry = SketchlinkManager.getAnchorOf(docTag);
                    if (entry == null) {
                        continue;
                    }
                    if (anchorSet != null && !anchorSet.contains(SketchlinkManager.TOOLS.getSketchlinkTagID(docTag))) {
                        continue;
                    }
                    resultEntries.add(entry);
                    @Nullable PsiElement target = entry.getTargetElement();
                    if (target == null) {
                        entry.isTargetUnchanged(true);
                        target = entry.getTargetElement();
                    }
                    resultTargets.add(target);
                }

                progressIndicator.setText("Done!");
                if (showResultToUser) {
                    browseToLocal(resultTargets);
                }
                return resultEntries;
            }
        });


    }
}
