/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.overlay;

import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.EditorFactory;
import com.intellij.openapi.editor.markup.*;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.TextRange;
import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.LocalLinkAnchor;
import plg.slCurrentProjectSettings;
import plg.slProjectComp;
import plg.ui.SketchlinkColors;

import java.awt.*;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by Peter on 15.06.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class CodeHighlighter {
    final private Map<Editor, Set<RangeHighlighter>> rangeHighlighterSet = new ConcurrentHashMap<>();
    private final Project project;

    public CodeHighlighter(@NotNull Project project) {
        this.project = project;
    }

    @NotNull
    public static CodeHighlighter getInstance(@NotNull Project project) {
        return slProjectComp.getInstance(project).getCodeHighlighter();
    }


    public void markCode(@NotNull Set<LocalLinkAnchor> highlightSet) {
        slCurrentProjectSettings projectSettings = slCurrentProjectSettings.getInstance(project);
        markCode(highlightSet, projectSettings.getColor(SketchlinkColors.OVERLAY_TARGET),
                projectSettings.getColor(SketchlinkColors.OVERLAY_TARGET_STROKE));
    }

    public void markCode(@NotNull final Set<LocalLinkAnchor> highlightSet, @NotNull final Color backColor, @NotNull final Color strokeColor) {
        //disable box when alpha is below 128, because IntelliJ doesn't support alpha for this

        UIUtil.invokeLaterIfNeeded(new Runnable() {
            @Override
            public void run() {
                for (@NotNull LocalLinkAnchor anchor : highlightSet) {
                    @Nullable Document document = anchor.getDocument();
                    if (document == null) {
                        continue;
                    }
                    for (@NotNull Editor editor : EditorFactory.getInstance().getEditors(document, anchor.getProject())) {
                        markCode(anchor, backColor, strokeColor, editor);
                    }
                }
            }
        });
    }

    public void markCode(@NotNull final LocalLinkAnchor anchor, @NotNull final Color backColor, @NotNull final Color strokeColor, @NotNull final Editor editor) {
        @Nullable final Color realStrokeColor = strokeColor.getAlpha() < 128 ? null : strokeColor;
        @NotNull final TextAttributes textAttributes = new TextAttributes(null, backColor, realStrokeColor, EffectType.BOXED, Font.BOLD);
        UIUtil.invokeLaterIfNeeded(new Runnable() {
            @Override
            public void run() {
                synchronized (rangeHighlighterSet) {
                    markRange(anchor.getTargetRange(), editor, textAttributes);
//                    System.out.println("Highlight: "+anchor.getClass());
                }
            }
        });

    }

    public void unmarkAllRanges() {
        UIUtil.invokeLaterIfNeeded(new Runnable() {
            @Override
            public void run() {
                synchronized (rangeHighlighterSet) {
                    @NotNull Iterator<Editor> keyIter = rangeHighlighterSet.keySet().iterator();
                    while (keyIter.hasNext()) {
                        Editor editor = keyIter.next();
                        Set<RangeHighlighter> rangeHighlighters = rangeHighlighterSet.get(editor);
                        @NotNull Iterator<RangeHighlighter> rangeIter = rangeHighlighters.iterator();
                        while (rangeIter.hasNext()) {
                            RangeHighlighter highlighter = rangeIter.next();
                            unmarkRange(highlighter, editor);
                            rangeIter.remove();
                        }
                        keyIter.remove();
                    }
                }
            }
        });
    }

    @NotNull
    private RangeHighlighter markRange(@NotNull TextRange range, @NotNull Editor editor, @NotNull TextAttributes attributes) {
        Set<RangeHighlighter> rangeHighlighters = rangeHighlighterSet.get(editor);
        if (rangeHighlighters == null) {
            rangeHighlighters = new HashSet<>();
            rangeHighlighterSet.put(editor, rangeHighlighters);
        }
        @NotNull RangeHighlighter result = editor.getMarkupModel().addRangeHighlighter(range.getStartOffset(), range.getEndOffset(), HighlighterLayer.LAST, attributes, HighlighterTargetArea.EXACT_RANGE);
        rangeHighlighters.add(result);
        return result;
    }

    private void unmarkRange(@NotNull RangeHighlighter highlighter, @NotNull Editor editor) {
        Set<RangeHighlighter> rangeHighlighters = rangeHighlighterSet.get(editor);
//        rangeHighlighters.remove(highlighter);
        editor.getMarkupModel().removeHighlighter(highlighter);
//        if (rangeHighlighters.isEmpty()) {
//            rangeHighlighterSet.remove(editor);
//        }
    }

    @NotNull
    public CodeHighlighter getCodeHighlighter() {
        return this;
    }
}
