/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview;

import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.MessageType;
import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.*;
import plg.tools.slGuiTools;
import plg.websocket.commands.builder.ChangeArtifactMetaDataRequestBuilder;

import javax.swing.*;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.JTextComponent;
import java.awt.*;
import java.awt.event.*;

public class AnnotationEditDialog extends JDialog {
    @NotNull
    private final Project project;
    private JPanel contentPane;
    private JButton buttonOK;
    private JButton buttonCancel;
    private JTextField myTitleText;
    private JTextField myAuthorText;
    private JTextArea myAnnotationArea;
    private JTextField myIDtext;
    private JTextField myDateText;


    public AnnotationEditDialog(@NotNull RemoteLinkAnchorS anchor) {
        this(anchor.getProject());
        @Nullable SketchMetadataTypeS metadata = anchor.getMetadata();
        if (metadata == null) {
            return;
        }
        myIDtext.setText(anchor.getAnchorId());
        myAnnotationArea.setText(metadata.getAnnotation());
        setSketchInfo(metadata);
    }


    public AnnotationEditDialog(@NotNull RemoteLinkAnchorM anchor) {
        this(anchor.getProject());
        @Nullable SketchMetadataTypeM metaM = anchor.getMetadata();
        @Nullable SketchMetadata metadata = anchor.getParentAnchor().getMetadata();
        if (metaM == null || metadata == null || !(metadata instanceof SketchMetadataTypeS)) {
            return;
        }
        myIDtext.setText(anchor.getAnchorId());
        setSketchInfo((SketchMetadataTypeS) metadata);
        myTitleText.setEditable(false);
        myAuthorText.setEditable(false);
        myDateText.setEditable(false);
        myAnnotationArea.setText(metaM.getAnnotation());
    }

    private AnnotationEditDialog(@NotNull Project project) {
        this.project = project;
        setContentPane(contentPane);
        setModal(true);
        getRootPane().setDefaultButton(buttonOK);

        buttonOK.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onOK();
            }
        });

        buttonCancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onCancel();
            }
        });

// call onCancel() when cross is clicked
        setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                onCancel();
            }
        });

// call onCancel() on ESCAPE
        contentPane.registerKeyboardAction(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onCancel();
            }
        }, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
        myAnnotationArea.getDocument().addDocumentListener(createDocumentListener(myAnnotationArea));
        myAuthorText.getDocument().addDocumentListener(createDocumentListener(myAuthorText));
        myTitleText.getDocument().addDocumentListener(createDocumentListener(myTitleText));
        this.setTitle("Edit Annotation");
    }

    private static void show(@NotNull final AnnotationEditDialog dialog) {
        UIUtil.invokeLaterIfNeeded(new Runnable() {
            @Override
            public void run() {
                dialog.setLocationRelativeTo(UIUtil.getActiveWindow());
                dialog.setAlwaysOnTop(true);

//                dialog.setMinimumSize(new Dimension(350,250));
//                dialog.setSize(360,300);
                dialog.pack();
                Insets insets = dialog.getInsets();
                Dimension minSize = dialog.rootPane.getMinimumSize();
                dialog.setMinimumSize(new Dimension(minSize.width + insets.left + insets.right, minSize.height + insets.top + insets.bottom));
                dialog.setVisible(true);
            }
        });
    }

    public static void createAndShow(RemoteLinkAnchorS knownAnchor) {
        final AnnotationEditDialog dialog = new AnnotationEditDialog(knownAnchor);
        show(dialog);
    }

    public static void createAndShow(@NotNull RemoteLinkAnchorM knownAnchor) {
        final AnnotationEditDialog dialog = new AnnotationEditDialog(knownAnchor);
        show(dialog);
    }

    public boolean isValidData() {

        if (myIDtext.getText() == null || myTitleText.getText() == null || myAuthorText.getText() == null || myAnnotationArea.getText() == null) {
            return false;
        }
//        System.out.println("- "+myTitleText.getText()+" - "+myAuthorText.getText()+" - "+myAnnotationArea.getText());
        return true; //!(myTitleText.getText().trim().isEmpty() || myAuthorText.getText().trim().isEmpty() || myAnnotationArea.getText().trim().isEmpty());
    }

    private void setSketchInfo(@NotNull SketchMetadataTypeS metadata) {
        myTitleText.setText(metadata.getTitle());
        myAuthorText.setText(metadata.getAuthors());
        myDateText.setText(metadata.getTimestamp());
    }

    private void onOK() {
// add your code here
        String text = myIDtext.getText();
        @Nullable LinkAnchor knownAnchor = SketchlinkManager.getInstance(project).getKnownAnchor(text);
        boolean changed = false;
        if (knownAnchor instanceof RemoteLinkAnchorS) {
            @Nullable SketchMetadataTypeS metadata = ((RemoteLinkAnchorS) knownAnchor).getMetadata();
            if (metadata == null) {
                slGuiTools.ShowNotificationAtBottom("Could not change meta-information.", project, MessageType.ERROR);
                dispose();
                return;
            }
            if (!myAnnotationArea.getText().equals(metadata.getAnnotation())) {
                changed = true;
                metadata.setAnnotation(myAnnotationArea.getText());
            }
            if (!myTitleText.getText().equals(metadata.getTitle())) {
                changed = true;
                metadata.setTitle(myTitleText.getText());
            }
            if (!myAuthorText.getText().equals(metadata.getAuthors())) {
                changed = true;
                metadata.setAuthors(myAuthorText.getText());
            }

        } else {
            if (knownAnchor instanceof RemoteLinkAnchorM) {
                @Nullable SketchMetadataTypeM metadata = ((RemoteLinkAnchorM) knownAnchor).getMetadata();
                if (metadata == null) {
                    slGuiTools.ShowNotificationAtBottom("Could not change meta-information.", project, MessageType.ERROR);
                    dispose();
                    return;
                }
                if (!myAnnotationArea.getText().equals(metadata.getAnnotation())) {
                    changed = true;
                    metadata.setAnnotation(myAnnotationArea.getText());
                }
            }
        }
        if (changed) {
            @NotNull ChangeArtifactMetaDataRequestBuilder builder = new ChangeArtifactMetaDataRequestBuilder(knownAnchor);
            builder.queue(knownAnchor.getProject(), null, true);
        }
        dispose();
    }

    private void onCancel() {
// add your code here if necessary
        dispose();
    }

    @NotNull
    private DocumentListener createDocumentListener(@NotNull final JTextComponent comp) {
        return new DocumentListener() {
            @Nullable
            private Color original = null;

            @NotNull
            public DocumentListener runOnce() {
                changed();
                return this;
            }

            @Override
            public void insertUpdate(DocumentEvent e) {
                changed();
            }

            @Override
            public void removeUpdate(DocumentEvent e) {
                changed();
            }

            @Override
            public void changedUpdate(DocumentEvent e) {
                changed();
            }

            private void changed() {
                if (comp.getText().trim().isEmpty()) {
//                    if (original == null)
//                        original = comp.getBackground();
//                    comp.setBackground(JBColor.RED);
                    buttonOK.setEnabled(false);
                } else {
//                    if (original != null) {
//                        comp.setBackground(original);
//                        original = null;
//                    }
                    if (isValidData()) {
                        buttonOK.setEnabled(true);
                    }
                }
            }
        }.runOnce();
    }

}
