/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview;

import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.popup.*;
import com.intellij.ui.awt.RelativePoint;
import com.intellij.ui.components.JBList;
import com.intellij.util.Alarm;
import com.intellij.util.Processor;
import com.intellij.util.containers.ConcurrentHashSet;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.LinkAnchor;
import plg.sketchlinks.PreviewableAnchor;
import plg.sketchlinks.RemoteLinkAnchorS;
import plg.sketchlinks.SketchMetadataTypeS;
import plg.slApplicationState;
import plg.slCurrentProjectSettings;
import plg.websocket.commands.builder.LoggingRequestBuilder;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ExecutionException;

/**
 * Created by Peter on 05.10.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class PreviewHandlerPopup extends PreviewHandlerBase {
    @NotNull
    private final Project project;
    @NotNull
    private final JComponent myPreviewContainer;
    @NotNull
    private final MouseMotionListener myMouseMotionListener;
    @NotNull
    private final MouseListener myMouseListener;
    @NotNull
    private List<LinkAnchor> content = new ArrayList<>();
    @NotNull
    private Set<String> linkedIDs = new ConcurrentHashSet<>();
    @NotNull
    private String currentTitle = "";
    @NotNull
    private RelativePoint relativeLocation = new RelativePoint(new Point(0, 0));
    private JBPopup myListPopup;
    @Nullable
    private JBPopup myPreviewPopup;
    @Nullable
    private Alarm myAlarmTimeTillPreview;
    private Dimension myPreviewLastDimension = new Dimension(500, 400);

    public PreviewHandlerPopup(@NotNull final Project project) {
        this.project = project;
        myPreviewContainer = new JPanel(new BorderLayout(0, 0));
        myPreviewContainer.setSize(myPreviewLastDimension);
        myMouseMotionListener = new MouseMotionAdapter() {
            @Override
            public void mouseMoved(@NotNull MouseEvent e) {
                @NotNull Point pointMouse = e.getPoint();
                if (!(e.getComponent() instanceof JBList)) {
                    return;
                }
                @NotNull JBList list = (JBList) e.getComponent();
                int newIndex = list.locationToIndex(pointMouse);

                if (newIndex == list.getSelectedIndex()) {
                    return;
                }
                list.clearSelection();
                list.setSelectedIndex(newIndex);
                if (list.getSelectedValue() instanceof PreviewListItem) {
                    showPreviewAnchor(((PreviewListItem) list.getSelectedValue()).getSketchAnchor());
                } else {
                    showPreviewAnchor(null);
                }

            }
        };
        myMouseListener = new MouseAdapter() {
//            @Override
//            public void mouseClicked(@NotNull MouseEvent e) {
//                if (!(e.getComponent() instanceof JBList)) {
//                    return;
//                }
//                @NotNull JBList list = (JBList) e.getComponent();
//                Object object = list.getSelectedValue();
//                if (object instanceof ListItem) {
//                    if (myAlarmTimeTillPreview != null) {
//                        myAlarmTimeTillPreview.cancelAllRequests();
//                        myAlarmTimeTillPreview = null;
//                    }
//                    if (myPreviewPopup != null) {
//                        myPreviewPopup.cancel();
//                    }
//                    ((ListItem) object).navigateTo();
//                }
//            }

            @Override
            public void mouseEntered(@NotNull MouseEvent e) {
                if (!(e.getComponent() instanceof JBList)) {
                    return;
                }
                @NotNull JBList list = (JBList) e.getComponent();
                @NotNull Point pointMouse = e.getPoint();
                int newIndex = list.locationToIndex(pointMouse);
                if (newIndex == list.getSelectedIndex()) {
                    return;
                }
                list.setSelectedIndex(newIndex);
                if (list.getSelectedValue() instanceof PreviewListItem) {
                    showPreviewAnchor(((PreviewListItem) list.getSelectedValue()).getSketchAnchor());
                } else {
                    showPreviewAnchor(null);
                }
            }
        };
    }

    private void showPreviewAnchor(@Nullable final PreviewableAnchor sketchAnchor) {
        if (sketchAnchor == null) {
            new LoggingRequestBuilder("closed sketch;").send(project);
            if (myPreviewPopup != null) {
                myPreviewPopup.cancel();
                myPreviewPopup = null;
            }
            if (myAlarmTimeTillPreview != null) {
                myAlarmTimeTillPreview.cancelAllRequests();
                myAlarmTimeTillPreview = null;
            }
            return;
        }
        sketchAnchor.getPreviewItem().setHighlightedIDs(getHighlighting());
        myPreviewLastDimension = (Dimension) myPreviewContainer.getSize().clone();
        myPreviewContainer.removeAll();
        sketchAnchor.getPreviewItem().setToolWindowMode(false);
        //horizontal scale, keeping aspect ratio intact
        sketchAnchor.getPreviewItem().setWidth(myPreviewLastDimension.width, 0);
        //ready scaled, get preview panel
        @NotNull JPanel preview = sketchAnchor.getPreview();
        myPreviewLastDimension = preview.getSize();
        myPreviewContainer.add(preview, BorderLayout.CENTER);
        myPreviewLastDimension.height += 2;
        myPreviewContainer.setSize(myPreviewLastDimension);
        myPreviewContainer.revalidate();
        if (myPreviewPopup != null && myPreviewPopup.isVisible()) {
            @NotNull Dimension popupDim = new Dimension(myPreviewLastDimension);
            popupDim.height += 14;
            myPreviewPopup.setSize(popupDim);
            new LoggingRequestBuilder("preview popup sketch change; " + sketchAnchor.getAnchorId()).send(project);
        } else {
            createNewPreviewPopup(sketchAnchor.getPreviewItem());
        }

    }

    private void createNewPreviewPopup(@NotNull final PreviewItem previewItem) {
        @NotNull Runnable loadPopup = new Runnable() {
            @Override
            public void run() {
                @NotNull JPanel picture = previewItem.getPanel();
                picture.revalidate();
                myPreviewPopup = JBPopupFactory.getInstance()
                        .createComponentPopupBuilder(myPreviewContainer, null)
                        .setCancelOnWindowDeactivation(true)
                        .setCancelOnClickOutside(true)
//                        .setAdText(remoteSketchAnchor.getDescription(), SwingConstants.LEADING)
                        .setResizable(true)
                        .setMinSize(new Dimension(100, 100))
                        .setRequestFocus(false)
                        .setCancelOnOtherWindowOpen(true)
                        .setMovable(true)
                        .setLocateWithinScreenBounds(true)
                        .setCouldPin(new Processor<JBPopup>() {
                            @Override
                            public boolean process(JBPopup jbPopup) {
                                PreviewManager.getInstance(project).setToolwindowMode();
                                myPreviewPopup.cancel();
                                return true;
                            }
                        })
                        .addListener(new JBPopupAdapter() {
                            @Override
                            public void onClosed(LightweightWindowEvent event) {
                                new LoggingRequestBuilder("preview popup closed ").send(project);
                            }
                        })
                        .createPopup();

                if (myAlarmTimeTillPreview != null) {
                    myAlarmTimeTillPreview.cancelAllRequests();
                    myAlarmTimeTillPreview = null;
                }
                if (!myListPopup.isVisible() || myListPopup.getContent() == null) {
                    return;
                }
                myPreviewPopup.show(RelativePoint.getNorthEastOf(myListPopup.getContent()));
//                myPreviewPopup.getContent().repaint();
//                myPreviewContainer.revalidate();
//                myPreviewContainer.repaint();
//                picture.revalidate();
//                picture.repaint();
                new LoggingRequestBuilder("preview popup shown").send(project);
            }
        };

        if (this.myAlarmTimeTillPreview == null) {
            this.myAlarmTimeTillPreview = new Alarm();
        } else {
            myAlarmTimeTillPreview.cancelAllRequests();
        }
        if (this.myAlarmTimeTillPreview.getActiveRequestCount() == 0) {
            this.myAlarmTimeTillPreview.addRequest(loadPopup, slCurrentProjectSettings.getInstance(project).getPreviewPopupTime());
        }
    }


    @Override
    public void addContent(@NotNull LinkAnchor anchor) {
        content.add(anchor);
        if (anchor instanceof PreviewableAnchor) {
            PreviewItem previewItem = ((PreviewableAnchor) anchor).getPreviewItem();
            previewItem.setEditMode(false);
            updateContentSettings(previewItem);
        }
        if (isVisible()) {
            setVisible(false);
            setVisible(true);
        }

    }

    @NotNull
    @Override
    public List<LinkAnchor> getContent() {
        return content;
    }

    @Override
    public void setContent(@NotNull List<LinkAnchor> anchors) {
        content.clear();
        content.addAll(anchors);
        updateContentSettings();
    }

    private void updateContentSettings(@NotNull PreviewItem previewItem) {
        previewItem.setToolWindowMode(false);
        previewItem.setEditMode(false);
        previewItem.setHighlightedIDs(getHighlighting());
    }

    private void updateContentSettings() {
        for (LinkAnchor linkAnchor : content) {
            if (linkAnchor instanceof PreviewableAnchor) {
                updateContentSettings(((PreviewableAnchor) linkAnchor).getPreviewItem());
            }
        }

    }

    @Override
    public void addHighlighting(@NotNull String id) {
        linkedIDs.add(id);
        updateContentSettings();
    }

    @NotNull
    @Override
    public Set<String> getHighlighting() {
        return linkedIDs;
    }

    /**
     * Set the ids to "linked" highlighting.
     *
     * @param linkedIDs
     */
    @Override
    public void setHighlighting(@NotNull Set<String> linkedIDs) {
        this.linkedIDs.clear();
        this.linkedIDs.addAll(linkedIDs);
        updateContentSettings();
    }

    @NotNull
    @Override
    public String getTitle() {
        return currentTitle;
    }

    @Override
    public void setTitle(@NotNull String title) {
        currentTitle = title;
    }

    @Override
    @NotNull
    public RelativePoint getRelativeLocation() {
        return relativeLocation;
    }

    @Override
    public void setRelativeLocation(@NotNull RelativePoint point) {
        this.relativeLocation = point;
        if (isVisible()) {
            myListPopup.setLocation(point.getScreenPoint());
            if (myPreviewPopup != null)
                myPreviewPopup.cancel();
        }
    }

    @Override
    public boolean isVisible() {
        return myListPopup != null && myListPopup.isVisible();
    }

    @Override
    public void setVisible(boolean show) {
        PreviewManager.getInstance(project).updateAvailableViews();
        if (show && (myListPopup == null || myListPopup.isDisposed())) {
            //create list content
            @NotNull List<ListItem> listToUse = new ArrayList<>(2 + content.size());
//            if (getContentAnchor() != null) {
//                listToUse.add(new UploadListItem(getContentAnchor()));
//                listToUse.add(new LinkListItem(getContentAnchor()));
//            }
            for (LinkAnchor linkAnchor : content) {
                if (linkAnchor instanceof PreviewableAnchor) {
                    listToUse.add(new PreviewListItem((PreviewableAnchor) linkAnchor));
                } else {
                    listToUse.add(new ListItem(linkAnchor));
                }
            }
            //create GUI list and add listeners
            @NotNull JBList list = new JBList(listToUse);
            list.addMouseMotionListener(myMouseMotionListener);
            list.addMouseListener(myMouseListener);
            list.setAutoscrolls(true);
            list.setVisibleRowCount(10);

            //create the actual popup
            @NotNull ComponentPopupBuilder popupBuilder = JBPopupFactory.getInstance().createComponentPopupBuilder(list, list);
            popupBuilder = popupBuilder.setCancelOnClickOutside(true)
                    .setCancelOnOtherWindowOpen(true)
                    .setCancelOnWindowDeactivation(true)
                    .setCancelOnClickOutside(true)
                    .setRequestFocus(false)
                    .setCancelOnOtherWindowOpen(true)
                    .setLocateWithinScreenBounds(true)
                    .setTitle(slApplicationState.RESOURCE_BUNDLE.getString("overlay.popup.title.multiple.ids.found"));
            myListPopup = popupBuilder.createPopup();
            myListPopup.show(getRelativeLocation());
            return;
        }
        //popup is visible already, nothing to do
        if (show) {
            return;
        }
        if (myListPopup != null)
            myListPopup.cancel();
        if (myPreviewPopup != null)
            myPreviewPopup.cancel();
    }

    @Override
    public void unregister() {
        if (myListPopup != null) {
            myListPopup.cancel();

            if (myListPopup.getContent() instanceof JBList) {
                @NotNull JBList list = (JBList) myListPopup.getContent();
                list.clearSelection();
                list.removeMouseListener(this.myMouseListener);
                list.removeMouseMotionListener(this.myMouseMotionListener);
                list.removeAll();
            }
        }
        if (myPreviewPopup != null) {
            myPreviewPopup.cancel();
        }
        myPreviewContainer.removeAll();
        if (myAlarmTimeTillPreview != null) {
            myAlarmTimeTillPreview.cancelAllRequests();
            myAlarmTimeTillPreview = null;
        }

    }

    private class PreviewListItem extends ListItem {

        public PreviewListItem(@NotNull PreviewableAnchor anchor) {
            super(anchor);
        }

        @NotNull
        public PreviewableAnchor getSketchAnchor() {
            return (PreviewableAnchor) getAnchor();
        }

        @NotNull
        @Override
        public String toString() {
            if (getSketchAnchor() instanceof RemoteLinkAnchorS) {
                @Nullable SketchMetadataTypeS metadata = ((RemoteLinkAnchorS) getSketchAnchor()).getMetadata();
                if (metadata != null) {
                    @NotNull String time = metadata.getTimestamp().contains(" ") ? metadata.getTimestamp().split(" ")[0] : metadata.getTimestamp();
                    return time + " " + metadata.getTitle();
                }
            }
            return getSketchAnchor().getDescription();
        }

    }

    private class ListItem {
        @NotNull
        private final LinkAnchor linkAnchor;

        public ListItem(@NotNull LinkAnchor anchor) {
            linkAnchor = anchor;
        }

        @NotNull
        @Override
        public String toString() {
            return linkAnchor.getDescription();
        }

        public void navigateTo() {
            linkAnchor.navigateTo();
        }

        @NotNull
        public String getAnchorID() {
            return linkAnchor.getAnchorId();
        }

        @NotNull
        public LinkAnchor getAnchor() {
            return linkAnchor;
        }
    }

    private class LinkListItem extends ListItem {

        public LinkListItem(@NotNull LinkAnchor anchor) {
            super(anchor);
        }

        @NotNull
        @Override
        public String toString() {
            return slApplicationState.RESOURCE_BUNDLE.getString("create.new.link");
        }

        @Override
        public void navigateTo() {
            PreviewManager.getInstance(project).showAnchorInWebView(getAnchor().getAnchorId(), false);
        }
    }

    private class UploadListItem extends ListItem {
        public UploadListItem(@NotNull LinkAnchor anchor) {
            super(anchor);
        }

        @NotNull
        @Override
        public String toString() {
            return "Import new sketch";
        }

        @Override
        public void navigateTo() {
            setVisible(false);
            RemoteLinkAnchorS remoteLinkAnchorS = null;
            try {
                remoteLinkAnchorS = PreviewManager.getInstance(project).showUploadSketchWindow(null, getAnchor());
                if (remoteLinkAnchorS != null) {
                    remoteLinkAnchorS.preparePreview();
                }
            } catch (ExecutionException | InterruptedException e) {
                e.printStackTrace();
            }
            if (remoteLinkAnchorS == null) {
                setVisible(true);
            } else {
                PreviewManager.getInstance(project).showUploadedSketches();
            }
        }

    }


}
