/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview;

import com.intellij.openapi.project.Project;
import com.intellij.openapi.wm.ToolWindow;
import com.intellij.openapi.wm.ToolWindowAnchor;
import com.intellij.openapi.wm.ToolWindowManager;
import com.intellij.ui.awt.RelativePoint;
import com.intellij.util.containers.ConcurrentHashSet;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.LinkAnchor;
import plg.sketchlinks.PreviewableAnchor;
import plg.websocket.commands.builder.LoggingRequestBuilder;

import java.awt.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.Vector;

/**
 * Created by Peter on 05.10.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class PreviewHandlerToolwindow extends PreviewHandlerBase {
    public static final String toolwindowName = "Sketchlink";
    @NotNull
    private final Project project;
    @NotNull
    private List<LinkAnchor> content = new Vector<>();
    @NotNull
    private Set<String> linkedIDs = new ConcurrentHashSet<>();
    @NotNull
    private String currentTitle = "";
    @NotNull
    private RelativePoint relativeLocation = new RelativePoint(new Point(0, 0));
    ;
    @Nullable
    private ToolWindow toolwindow;
    @Nullable
    private PreviewToolWindow myPreviewToolWindow;
    private EmptyPreviewOptions previewOptions;

    public PreviewHandlerToolwindow(@NotNull Project project) {
        this.project = project;
    }

    @Override
    public void addContent(@NotNull LinkAnchor anchor) {
        if (content.contains(anchor)) {
            return;
        }
        content.add(anchor);
        if (anchor instanceof PreviewableAnchor) {
            @NotNull PreviewItem previewItem = ((PreviewableAnchor) anchor).getPreviewItem();
            createToolwindowIfNeccessary();
            updateContentSettings(previewItem);
            if (myPreviewToolWindow != null) {
                myPreviewToolWindow.addSketches(previewItem);
            }
        }

    }

    private void updateContentSettings(@NotNull PreviewItem previewItem) {
        previewItem.setToolWindowMode(true);
        previewItem.setHighlightedIDs(getHighlighting());
    }

    private void updateContentSettings() {
        PreviewItem pinnedContent = getPinnedContent();
        if (pinnedContent != null) {
            updateContentSettings(pinnedContent);
        }
        for (LinkAnchor linkAnchor : content) {
            if (linkAnchor instanceof PreviewableAnchor) {
                updateContentSettings(((PreviewableAnchor) linkAnchor).getPreviewItem());
            }
        }
    }

    @NotNull
    @Override
    public List<LinkAnchor> getContent() {
        return content;
    }

    @Override
    public void setContent(@NotNull List<LinkAnchor> anchors) {
        this.content.clear();
        this.content.addAll(anchors);
        updateContentSettings();
        setVisible(true);
    }

    @Override
    public void addHighlighting(@NotNull String id) {
        linkedIDs.add(id);
        updateContentSettings();
    }

    @NotNull
    @Override
    public Set<String> getHighlighting() {
        return linkedIDs;
    }

    /**
     * Set the ids to "linked" highlighting.
     *
     * @param linkedIDs
     */
    @Override
    public void setHighlighting(@NotNull Set<String> linkedIDs) {
        this.linkedIDs.clear();
        this.linkedIDs.addAll(linkedIDs);
        updateContentSettings();
    }

    @NotNull
    @Override
    public String getTitle() {
        return currentTitle;
    }

    @Override
    public void setTitle(@NotNull String title) {

        currentTitle = title.startsWith(" - ") ? title : " - " + title;
        if (myPreviewToolWindow != null) {
            myPreviewToolWindow.setTitle(title);
        }
    }

    @NotNull
    @Override
    public RelativePoint getRelativeLocation() {
        return relativeLocation;
    }

    @Override
    public void setRelativeLocation(@NotNull RelativePoint point) {
        this.relativeLocation = point;
    }

    @Override
    public boolean isVisible() {
        return myPreviewToolWindow != null && myPreviewToolWindow.isShowing();
    }

    @Override
    public void setVisible(boolean showing) {
        PreviewManager.getInstance(project).updateAvailableViews();
        if (createToolwindowIfNeccessary()) {
            setContentToToolwindow();
            if (myPreviewToolWindow != null)
                myPreviewToolWindow.setOpen(showing);
        }
    }

    @Override
    public void unregister() {
        if (toolwindow == null)
            return;
        toolwindow.setAvailable(false, null);
        ToolWindowManager.getInstance(project).unregisterToolWindow(toolwindowName);
        toolwindow = null;
        myPreviewToolWindow = null;
    }

    private void setContentToToolwindow() {
        @NotNull List<PreviewItem> previewItems = new ArrayList<>(content.size());
        for (LinkAnchor linkAnchor : content) {
            if (linkAnchor instanceof PreviewableAnchor) {
                previewItems.add(((PreviewableAnchor) linkAnchor).getPreviewItem());
            }
        }
        PreviewItem pinnedContent = getPinnedContent();
        if (pinnedContent != null) {
            if (previewItems.contains(pinnedContent)) {
                previewItems.remove(pinnedContent);
            }
            previewItems.add(0, pinnedContent);
        }
        if (createToolwindowIfNeccessary() && myPreviewToolWindow != null) {
            if (content.size() == 0) {
                previewOptions = new EmptyPreviewOptions(PreviewManager.getInstance(project));
                myPreviewToolWindow.setContent(previewOptions.getRootPanel());
            } else {
                if (previewOptions != null) {
                    previewOptions = null;
                }
                myPreviewToolWindow.setSketches(previewItems);
            }
            myPreviewToolWindow.setTitle(getTitle());
            myPreviewToolWindow.scrollToTop();
        }
    }

    private boolean createToolwindowIfNeccessary() {
        if (myPreviewToolWindow != null) {
            return true;
        }
        toolwindow = ToolWindowManager.getInstance(project).registerToolWindow(toolwindowName, true, ToolWindowAnchor.RIGHT);
        if (toolwindow == null) {
            return false;
        }
        myPreviewToolWindow = new PreviewToolWindow(currentTitle, toolwindow, project);
        myPreviewToolWindow.setTitle(getTitle());
        myPreviewToolWindow.setActive(true);
        myPreviewToolWindow.setOpen(true);
        new LoggingRequestBuilder("enter toolwindow mode;").send(project);
        return true;
    }
}
