/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview;

import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import com.intellij.ui.awt.RelativePoint;
import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.SLConfigForm;
import plg.sketchlinks.LinkAnchor;
import plg.sketchlinks.RemoteLinkAnchorS;
import plg.sketchlinks.SketchlinkManager;
import plg.slApplicationState;
import plg.slProjectComp;
import plg.tools.slToolBox;
import plg.ui.sketchupload.SketchUpload;
import plg.websocket.RunCmd;
import plg.websocket.commands.builder.ShowRequestBuilder;
import plg.websocket.commands.getAvailableClientsCMD;
import plg.websocket.commands.parameter.EmptyPayload;
import plg.websocket.slWebsocketManager;

import javax.swing.*;
import java.awt.*;
import java.util.*;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;

/**
 * Created by Peter on 05.04.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class PreviewManager {
    private final static Logger LOG = Logger.getInstance(PreviewManager.class);
    @NotNull
    public final PreviewHandler toolwindowPreview;
    @NotNull
    public final PreviewHandler popupPreview;
    @NotNull
    private final Project project;
    private final AvailableViewImpl openLocalTab = new AvailableViewImpl(slApplicationState.RESOURCE_BUNDLE.getString("new.local.browser.tab"), "webview", "localhost"); //NON-NLS NON-NLS
    @NotNull
    private Map<String, AvailableViews> availableViewsSet = new ConcurrentHashMap<>();
    @Nullable
    private List<AvailableViews> defaultViews;
    private PreviewHandler previewHandler;
    private List<LinkAnchor> uploadedSketches = new ArrayList<>();

    public PreviewManager(@NotNull Project project) {
        this.project = project;
        toolwindowPreview = new PreviewHandlerToolwindow(project);
        popupPreview = new PreviewHandlerPopup(project);
        previewHandler = popupPreview;
    }

    @NotNull
    public static PreviewManager getInstance(@NotNull Project project) {
        return slProjectComp.getInstance(project).getPreviewManager();
    }

    /**
     * Opens a new Browser tab and loads the webview with the supplied anchorID.
     * If the anchorID is a c-anchor, it opens the webview using "create new link" parameters, "open" mode otherwise
     * TODO: Rethink when Webview starts be able to preview source code
     *
     * @param anchorID the anchorID to view or link in the webview
     */
    public static void openAnchorInLocalTab(@NotNull String anchorID, @NotNull Project project) {
        //JOptionPane.showMessageDialog(null,"openAnchorInLocalTab in PreviewManager is not implemented yet");
        @NotNull StringBuilder sb = new StringBuilder(slProjectComp.getInstance(project).getServerURL().replaceFirst("ws://", "http://")); //NON-NLS NON-NLS
        //"http://server/link/{codeAnchorID}"
        if (sb.charAt(sb.length() - 1) != '/') {
            sb.append('/');
        }
        if (!sb.substring(0, 7).toLowerCase().equals("http://")) { //NON-NLS
            sb.insert(0, "http://"); //NON-NLS
        }
        if (anchorID.toLowerCase().startsWith("c")) { //NON-NLS
            sb.append("link/"); //NON-NLS
        } else {
            sb.append("open/"); //NON-NLS
        }

        sb.append(anchorID);
        slToolBox.OpenBrowserTab(sb.toString(), project);
    }

    @Nullable
    public RemoteLinkAnchorS showUploadSketchWindow(@Nullable JComponent parentComponent, @Nullable LinkAnchor linkSketchTo) throws ExecutionException, InterruptedException {
        if (!slProjectComp.getInstance(project).isConnectedToServer()) {
            slProjectComp.getInstance(project).setConnectedToServer(true);
        }
//        if (!slWebsocketManager.getInstance(project).isConnected()) {
//            slWebsocketManager.getInstance(project).toggleWebsocketConnection(true);
//        }
        String serverURL = slProjectComp.getInstance(project).getServerURL();
        if (serverURL.trim().isEmpty()) {
            SLConfigForm.showInvalidServerMessage("Server URL not set", project);
            return null;
        }
        @NotNull SketchUpload sketchUpload = new SketchUpload(project, serverURL, linkSketchTo != null ? linkSketchTo.getAnchorId() : null);
        sketchUpload.revalidate();
//        sketchUpload.pack();
        sketchUpload.setSize(500, 400);
        if (parentComponent != null) {
            sketchUpload.setLocationRelativeTo(parentComponent);
        } else {
            @NotNull Window activeWindow = UIUtil.getActiveWindow();
            Dimension size = activeWindow.getSize();
            Point onScreen = activeWindow.getLocationOnScreen();
            onScreen.x += size.width / 2;
            onScreen.y += size.height / 2;
            onScreen.x -= onScreen.x < 250 ? 0 : 250;
            onScreen.y -= onScreen.y < 200 ? 0 : 200;
            sketchUpload.setLocation(onScreen);
        }
        sketchUpload.run();
        @Nullable String result = null;
        try {
            result = sketchUpload.get();
        } finally {
            sketchUpload.dispose();
        }
        if (result == null || result.isEmpty()) {
            return null;
        }
        @NotNull LinkAnchor linkAnchor = SketchlinkManager.getInstance(project).addRemote(result, null, project);
        uploadedSketches.add(0, linkAnchor);
        return (RemoteLinkAnchorS) linkAnchor;
    }

    public void showUploadedSketches() {
        setToolwindowMode();
        getPreviewHandler().setPinnedContent(null);
        getPreviewHandler().setHighlighting(new HashSet<String>());
        getPreviewHandler().setTitle("Uploaded sketches in this session");
        getPreviewHandler().setContent(uploadedSketches);
        getPreviewHandler().setVisible(true);
    }

    public void showAllKnownSketches() {
        Collection<RemoteLinkAnchorS> anchors = SketchlinkManager.getInstance(project).getAllAnchors(RemoteLinkAnchorS.class);
        setToolwindowMode();
        getPreviewHandler().setPinnedContent(null);
        getPreviewHandler().setHighlighting(new HashSet<String>());
        getPreviewHandler().setTitle("Cached sketches");
        getPreviewHandler().setContent(new ArrayList<LinkAnchor>(anchors));
        getPreviewHandler().setVisible(true);
    }

    @NotNull
    public Project getProject() {
        return project;
    }

    @Nullable
    public PreviewItem createSVGPreview(@NotNull RemoteLinkAnchorS anchor, @Nullable LinkAnchor linkProvidingAnchor) {
        return new PreviewSketch(this, anchor, linkProvidingAnchor);
        //return new PreviewSvg(anchor, previewLocation, this);
    }


    /**
     * If true, use the toolwindow for sketch preview, if false, use the popup.
     */
    public void setToolwindowMode() {
        if (getPreviewHandler() != toolwindowPreview) {
            setPreviewHandler(toolwindowPreview);
        }
    }

    public void setPopupPreviewMode() {
        if (getPreviewHandler() != popupPreview) {
            setPreviewHandler(popupPreview);
        }
    }

    public void showPreview(@NotNull LinkAnchor source, @NotNull JComponent parentLocation) {
        @NotNull PreviewHandler handler = getPreviewHandler();
        handler.setRelativeLocation(RelativePoint.getNorthEastOf(parentLocation));
        handler.SetContentFromAnchor(source, null);
        handler.setVisible(true);
    }

    @NotNull
    public AvailableViews addViewer(@NotNull AvailableViews viewer) {
        if (defaultViews != null) {
            defaultViews.clear();
        }
        defaultViews = null;
        AvailableViews foundView = availableViewsSet.get(viewer.getIdentifier());
        if (foundView == null) {
            availableViewsSet.put(viewer.getIdentifier(), viewer);
            foundView = viewer;
        }
        return foundView;
    }

    @NotNull
    public AvailableViews addViewer(@NotNull String displayName, @NotNull String type, @NotNull String identifier, @NotNull String... displayableArtifactID) {
        @NotNull AvailableViews view = addViewer(new AvailableViewImpl(displayName, type, identifier));
        return view;
    }

    public void updateAvailableViews() {
        try {
            @NotNull RunCmd command = slWebsocketManager.getInstance(project).getCommandsManager().getCommandOrInstantiate(getAvailableClientsCMD.class);
            command.sendRequest(new EmptyPayload(), null, null, false);
        } catch (@NotNull IllegalAccessException | InstantiationException e) {
            LOG.error(e);
        }
    }

    public void setAvailableViewsSet(@NotNull Collection<AvailableViews> availableViewsCollection) {
        availableViewsSet.clear();
        for (@NotNull AvailableViews views : availableViewsCollection) {
            addViewer(views);
        }
    }

    @NotNull
    public List<AvailableViews> getViewsFor(@Nullable LinkAnchor anchor) {
        if (anchor == null) {
            return new ArrayList<>(availableViewsSet.values());
        }
        @NotNull List<AvailableViews> result = new ArrayList<>();
        for (@NotNull AvailableViews availableView : availableViewsSet.values()) {
            if (availableView.canShow(anchor)) {
                result.add(availableView);
            }
        }
        return result;
    }

    public void showAnchorInWebView(@NotNull String anchorID, boolean allowMultipleSelections) {
        if (defaultViews != null && defaultViews.size() > 0 && (allowMultipleSelections || defaultViews.size() != 1)) {
            showAnchorInDefaultViews(defaultViews, anchorID);
            return;
        }

        @NotNull List<AvailableViews> views = new ArrayList<>();
        for (@NotNull AvailableViews availableViews : availableViewsSet.values()) {
            if (availableViews.isWebview()) {
                views.add(availableViews);
            }
        }
        //add test data
//        views.add(new AvailableViewImpl("Here comes Johnny","webview","0.0.0.1"));
//        views.add(new AvailableViewImpl("Here comes Johnny Walker","webview","0.0.0.1"));
//        views.add(new AvailableViewImpl("Never gonna give you up","webview","0.0.6.1"));
//        views.add(new AvailableViewImpl("Here comes Johnny","ide","100.5.0.1"));

        Collections.sort(views, new Comparator<AvailableViews>() {
            @Override
            public int compare(@NotNull AvailableViews o1, @NotNull AvailableViews o2) {
                if (o1.getIpAddress().equals(o2.getIpAddress())) {
                    return o1.getDisplayName().compareTo(o2.getDisplayName());
                }
                return o1.getIdentifier().compareTo(o2.getIdentifier());
            }
        });
        views.add(0, openLocalTab);
        @NotNull List<AvailableViews> results = new ArrayList<>();
        boolean remember = ViewsDialog.showDialog(views, results, allowMultipleSelections);
        if (remember) {
            defaultViews = results;
        }
        showAnchorInDefaultViews(results, anchorID);
//        System.out.println("Previews selected: "+results.size()+" Remember? "+remember);
//        for (AvailableViews result : results) {
//            System.out.println("View: "+result);
//        }
    }

    private void showAnchorInDefaultViews(@NotNull Collection<AvailableViews> views, @NotNull String anchorID) {
        boolean hasExternal = false;
        @NotNull ShowRequestBuilder showRequestBuilder = new ShowRequestBuilder(anchorID);
        for (@NotNull AvailableViews defaultView : views) {
            if (defaultView.equals(openLocalTab)) {
                openAnchorInLocalTab(anchorID, project);
                continue;
            }
            hasExternal = true;
            showRequestBuilder.addView(defaultView);
        }
        if (!hasExternal) {
            return;
        }
        showRequestBuilder.send(project, null);
    }

    public void clearAvailablePreviews() {
        availableViewsSet.clear();
    }

    @NotNull
    public PreviewHandler getPreviewHandler() {
        return previewHandler;
    }

    public void setPreviewHandler(@NotNull PreviewHandler previewHandler) {
        if (this.previewHandler == previewHandler) {
            return;
        }
        this.previewHandler.setVisible(false);
        previewHandler.copyFrom(this.previewHandler);
        this.previewHandler = previewHandler;
    }
}
