/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview;

import com.intellij.openapi.actionSystem.DataProvider;
import com.intellij.openapi.ui.MessageType;
import com.intellij.openapi.util.Pair;
import com.intellij.util.containers.ConcurrentHashSet;
import com.kitfox.svg.SVGCache;
import com.kitfox.svg.SVGUniverse;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.*;
import plg.slApplicationState;
import plg.tools.slGuiTools;
import plg.ui.overlay.CodeHighlighter;
import plg.ui.preview.svgwrapper.SVGDiagramEx;
import plg.ui.preview.svgwrapper.SVGPanelExt;
import plg.ui.preview.svgwrapper.SVGPreviewContainer;

import javax.swing.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.net.URI;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;
import java.util.UUID;

/**
 * Created by noxin on 15.07.14.
 */
public class PreviewSketch implements PreviewItem, DataProvider {

    /**
     * The Anchor that provides this with content.
     */
    @NotNull
    private final RemoteLinkAnchorS myTopSketchAnchor;
    /**
     * The PreviewManager that manages this Preview
     */
    @NotNull
    private final PreviewManager myManager;
    /**
     * Delegate the actual content and handling of the preview to a SVGPreviewContainer instance
     */
    @NotNull
    private final SVGPreviewContainer svgPreviewContainer;
    private final Set<String> highlightedIDs = new ConcurrentHashSet<>();
    @Nullable
    private LinkAnchor myLinkedAnchor;
    @Nullable
    private SVGUniverse myUniverse;
    @Nullable
    private URI mySvgUri;
    @NotNull
    private PropertyChangeSupport propertyChangeSupport =
            new PropertyChangeSupport(this);
    private long timeOfLastAccess;
    /**
     * Save the remoteURL in case the connection is disabled after loading a sketch.
     */
    private URL currentRemoteUrl;


    /**
     * Only for usage of RemoteLinkAnchorC, as long as there is no other way to preview remote code.
     */
    public PreviewSketch(@NotNull PreviewManager manager, @NotNull RemoteLinkAnchorC sourceAnchor) {
        myManager = manager;
        myTopSketchAnchor = new RemoteLinkAnchorS("s-" + UUID.randomUUID(), manager.getProject());
        myUniverse = SVGCache.getSVGUniverse();
        mySvgUri = myUniverse.loadSVG(PreviewSketch.class.getResource("/plg/ui/icons/tux.svg"));

        svgPreviewContainer = new SVGPreviewContainer(new SVGDiagramEx(myUniverse.getDiagram(mySvgUri), manager.getProject()), sourceAnchor.getProject());
        svgPreviewContainer.setAnnotation(sourceAnchor.getDescription());
    }


    public PreviewSketch(@NotNull PreviewManager manager, @NotNull RemoteLinkAnchorS sketchAnchor, @Nullable LinkAnchor referencedAnchor) {
        this.myManager = manager;
        this.myTopSketchAnchor = sketchAnchor;
        this.myLinkedAnchor = referencedAnchor;
        updateTimeOfLastAccess();

        myUniverse = new SVGUniverse();
        @Nullable URL previewLocation = myTopSketchAnchor.getPreviewLocation();
        if (previewLocation == null) {
            previewLocation = PreviewSketch.class.getResource("/plg/ui/icons/tux.svg");
        }
        mySvgUri = myUniverse.loadSVG(previewLocation);
        @NotNull SVGDiagramEx svgDiagramEx = new SVGDiagramEx(myUniverse.getDiagram(mySvgUri), myManager.getProject());
        svgPreviewContainer = new SVGPreviewContainer(svgDiagramEx, sketchAnchor.getProject());
        svgPreviewContainer.setAnnotation(getDefaultTitle());
        //lookup the metadata of all marker contained in this sketch
        for (@NotNull String shapeID : svgPreviewContainer.getDiagram().getContainedIDs()) {
            SketchlinkManager.getInstance(sketchAnchor.getProject()).addRemote(shapeID, myTopSketchAnchor.getAnchorId(), myManager.getProject());
        }
        updateSVGElementColors();
        PropertyChangeListener listener = new PropertyChangeListener() {
            @Override
            public void propertyChange(PropertyChangeEvent evt) {
                Object source = evt.getSource();
                if (svgPreviewContainer.getCurrentID() == null) {
                    updateDescription((String) null);
                    return;
                }
                if ((!(source instanceof RemoteLinkAnchor)) || !((RemoteLinkAnchor) source).getAnchorId().equals(svgPreviewContainer.getCurrentID())) {
                    return;
                }
                updateDescription((RemoteLinkAnchor) source);
            }
        };
        sketchAnchor.addPropertyChangeListener(listener);
        svgPreviewContainer.addPropertyChangeListener(new PropertyChangeListener() {
            @Override
            public void propertyChange(@NotNull PropertyChangeEvent evt) {
                if (SVGPreviewContainer.CURRENT_ID_Property.equals(evt.getPropertyName())) {
                    updateDescription((String) evt.getNewValue());
                    CodeHighlighter.getInstance(myTopSketchAnchor.getProject()).unmarkAllRanges();
                    if (evt.getNewValue() == null) {
                        return;
                    }
                    @NotNull String currentID = (String) evt.getNewValue();
                    @Nullable LinkAnchor knownAnchor = SketchlinkManager.getInstance(myManager.getProject()).getKnownAnchor(currentID);
                    if (knownAnchor == null) {
                        return;
                    }
                    @NotNull Set<LocalLinkAnchor> codeHighlight = new HashSet<LocalLinkAnchor>();
                    for (@Nullable Pair<LinkAnchor, Boolean> booleanPair : SketchlinkManager.getInstance(myManager.getProject()).getLinksFor(knownAnchor)) {
                        if (booleanPair == null) {
                            continue;
                        }
                        if (booleanPair.second) {
                            continue; //skip deprecated
                        }
                        if (booleanPair.first instanceof LocalLinkAnchor) {
                            codeHighlight.add((LocalLinkAnchor) booleanPair.first);
                        }
                    }
                    CodeHighlighter.getInstance(myTopSketchAnchor.getProject()).markCode(codeHighlight);
                }
            }
        });
    }

    private void updateDescription(@NotNull RemoteLinkAnchor anchor) {
        svgPreviewContainer.setAnnotation(anchor.getDescription());
    }

    private void updateDescription(@Nullable String currentID) {
        if (currentID == null) {
            svgPreviewContainer.setAnnotation(getDefaultTitle(), getDefaultShortTitle());
            return;
        }
        @Nullable LinkAnchor knownAnchor = SketchlinkManager.getInstance(myManager.getProject()).getKnownAnchor(currentID);
        if (!(knownAnchor instanceof RemoteLinkAnchor)) {
            svgPreviewContainer.setAnnotation(knownAnchor != null ? "Invalid Anchor" : slApplicationState.RESOURCE_BUNDLE.getString("no.description.available"));
            return;
        }
        updateDescription((RemoteLinkAnchor) knownAnchor);
    }

    @NotNull
    private String getDefaultShortTitle() {
        @Nullable SketchMetadataTypeS metadata = myTopSketchAnchor.getMetadata();
        if (metadata == null) {
            return "No information available.";
        }
        @NotNull StringBuilder sb = new StringBuilder();
        if (!metadata.getTimestamp().trim().isEmpty()) {
            sb.append(metadata.getTimestamp());
            sb.append(" ");
        }
        if (!metadata.getTitle().trim().isEmpty()) {
            sb.append("\"");
            sb.append(metadata.getTitle());
            sb.append("\" ");
        } else {
            sb.append("[Title was not set] ");
        }
        if (!metadata.getAuthors().trim().isEmpty()) {
            sb.append("(");
            sb.append(metadata.getAuthors());
            sb.append(")");
        }
        return sb.toString();
    }

    @NotNull
    private String getDefaultTitle() {
        @Nullable SketchMetadataTypeS metadata = myTopSketchAnchor.getMetadata();
        if (metadata == null) {
            return "No information available.";
        }
        @NotNull StringBuilder sb = new StringBuilder();
        if (!metadata.getTimestamp().trim().isEmpty()) {
            sb.append("Date: ");
            sb.append(metadata.getTimestamp());
            sb.append("\n");
        }
        if (!metadata.getTitle().trim().isEmpty()) {
            sb.append("Title: \"");
            sb.append(metadata.getTitle());
            sb.append("\"\n");
        } else {
            sb.append("[Title was not set]\n");
        }
        if (!metadata.getAuthors().trim().isEmpty()) {
            sb.append("Authors: ");
            sb.append(metadata.getAuthors());
        }
        return sb.toString();
    }

    @Override
    public void setRemoteLocation(@NotNull URL url) {
        if (url.equals(currentRemoteUrl)) {
            return;
        }
        @Nullable URI oldUri = mySvgUri;
        currentRemoteUrl = url;
        @NotNull SVGPanelExt svgPanelExt = svgPreviewContainer.getSVGPanel();
        if (myUniverse != null && myUniverse != SVGCache.getSVGUniverse()) {
            myUniverse.clear();
        } else {
            myUniverse = new SVGUniverse();
            svgPanelExt.setSvgUniverse(myUniverse);
        }
        mySvgUri = myUniverse.loadSVG(url);
        if (mySvgUri == null) {
            slGuiTools.ShowNotificationAtBottom("Received an invalid Sketch-URL or the Server is currently not available.", getAnchorOfSketch().getProject(), MessageType.ERROR);
            return;
        }
//        SVGDiagram diagram = myUniverse.getDiagram(mySvgUri);

        //lookup the metadata of all marker contained in this sketch
        for (@NotNull String shapeID : svgPreviewContainer.getDiagram().getContainedIDs()) {
            SketchlinkManager.getInstance(myManager.getProject()).addRemote(shapeID, myTopSketchAnchor.getAnchorId(), myManager.getProject());
        }
        @NotNull SVGDiagramEx svgDiagramEx = new SVGDiagramEx(myUniverse.getDiagram(mySvgUri), myManager.getProject());
        svgPreviewContainer.setDiagram(svgDiagramEx);
        updateSVGElementColors();
        svgPanelExt.repaint();
        propertyChangeSupport.firePropertyChange("uri", oldUri, mySvgUri);
    }

    private void updateSVGElementColors() {
        @NotNull SVGDiagramEx svgDiagramEx = svgPreviewContainer.getDiagram();
        svgDiagramEx.setLinkedIDs(highlightedIDs);
    }

    @NotNull
    @Override
    public JPanel getPanel() {
        updateTimeOfLastAccess();
        return svgPreviewContainer.getRootPanel();
    }

    @NotNull
    public SVGPreviewContainer getSVGPreviewContainer() {
        return svgPreviewContainer;
    }

    @NotNull
    public SVGPanelExt getSVGPanelExt() {
        updateTimeOfLastAccess();
        return svgPreviewContainer.getSVGPanel();
    }

    /**
     * Gets the highlighted ids that were set previously.
     */
    @NotNull
    @Override
    public Set<String> getHighlightedIDs() {
        return highlightedIDs;
    }

    /**
     * Tells the PreviewItem to highlight the following ids.
     * Using an empty set disables highlighting.
     *
     * @param ids
     */
    @Override
    public void setHighlightedIDs(@NotNull Set<String> ids) {
        this.highlightedIDs.clear();
        this.highlightedIDs.addAll(ids);
        updateSVGElementColors();
    }

    /**
     * Gets the IDs that are highlighted in this item, returns a subset of <code>getHighlightedIDs()</code>
     */
    @NotNull
    @Override
    public Set<String> getHighlightedIDsInThisItem() {
        SVGDiagramEx diagramEx = getSVGPanelExt().getDiagramEx();
        @NotNull Set<String> containedIDs = diagramEx.getContainedIDs();
        @NotNull Set<String> result = new HashSet<>(containedIDs.size());
        for (String containedID : containedIDs) {
            if (highlightedIDs.contains(containedID))
                result.add(containedID);
        }
        return result;
    }

    @NotNull
    @Override
    public RemoteLinkAnchorS getAnchorOfSketch() {
        return myTopSketchAnchor;
    }


    @Override
    public long getTimeOfLastAccess() {
        return timeOfLastAccess;
    }

    private void updateTimeOfLastAccess() {
        long oldval = timeOfLastAccess;
        timeOfLastAccess = System.currentTimeMillis();
        propertyChangeSupport.firePropertyChange(PreviewItem.TIME_OF_LAST_ACCESS_PROPERTY, oldval, timeOfLastAccess);
    }

    @Override
    public void clear() {
        if (myUniverse != null && myUniverse != SVGCache.getSVGUniverse()) {
            myUniverse.clear();
            myUniverse = null;
        }
        mySvgUri = null;
    }


    @Override
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener(listener);
    }

    @Override
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener(listener);
    }

    /**
     * Sets the width of the preview.
     * It should calculate the height automatically based on the aspect ratio.
     * If the scrollBarWidth is >0, then the scrollbar is visible and will hide the rightmost pixel of the panel.
     * the complete_width includes the scrollbarwidth, so the actual width for content might be smaller.
     *
     * @param complete_width
     * @param scrollBarWidth
     */
    @Override
    public void setWidth(int complete_width, int scrollBarWidth) {
        svgPreviewContainer.setWidth(complete_width, scrollBarWidth);
    }


    @Override
    public void setToolWindowMode(boolean active) {
        svgPreviewContainer.setToolWindowMode(active);

    }

    @Override
    public void setEditMode(boolean active) {
        svgPreviewContainer.setEditMode(active);
    }

    @Nullable
    @Override
    public Object getData(@NonNls String dataId) {
        if (DATA_KEY.is(dataId)) {
            return this;
        }
        return null;
    }
}
