/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview;

import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.IconLoader;
import com.intellij.openapi.wm.ToolWindow;
import com.intellij.ui.AncestorListenerAdapter;
import com.intellij.ui.content.Content;
import com.intellij.ui.content.ContentFactory;
import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import javax.swing.event.AncestorEvent;
import java.awt.*;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * Created by Peter on 26.03.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */

public class PreviewToolWindow {
    private JPanel rootPanel;

    private JPanel myPreviewContent;
    private JPanel toolBar;

    private JScrollPane scrollPanel;
    private JLabel myEditToggle;
    private JLabel myAccordionButton;
    private JLabel myUnpinButton;
    private JButton myEditMode;
    @NotNull
    private List<PreviewItem> myPreviewItems = new ArrayList<>();

    @NotNull
    private ToolWindow myToolWindow;


    private String myTitle;

    private boolean isRegistered;
    private boolean activeEditMode = false;
    private boolean sketchesExpanded = true;
    private Icon editModeOffIcon = IconLoader.getIcon("/plg/ui/icons/EditModeOff.png");
    private Icon editModeOnIcon = IconLoader.getIcon("/plg/ui/icons/EditModeOn.png");
    private Icon accordionUpIcon = IconLoader.getIcon("/plg/ui/icons/arrows_up.png");
    private Icon accordionDownIcon = IconLoader.getIcon("/plg/ui/icons/arrows_down.png");


    public PreviewToolWindow(@NotNull String title, @NotNull final ToolWindow toolWindow, @NotNull final Project project) {
        super();
        myTitle = title;
        isRegistered = true;

        this.myToolWindow = toolWindow;
        this.myToolWindow.setTitle(myTitle);
        //System.out.println("PreviewToolwindow Create");
        @NotNull VerticalLayout verticalLayout = new VerticalLayout(5, VerticalLayout.CENTER, VerticalLayout.TOP);
        myPreviewContent.setLayout(verticalLayout);

        myPreviewContent.revalidate();
        myPreviewContent.repaint();
        rootPanel.addAncestorListener(new AncestorListenerAdapter() {
            @Override
            public void ancestorAdded(AncestorEvent event) {
                myPreviewContent.repaint();
            }

        });
        rootPanel.addComponentListener(new ComponentAdapter() {
            @Override
            public void componentResized(@NotNull ComponentEvent e) {
//                System.out.println("RootPanel Toolwindow Resize: "+e.getComponent().getSize()+ " ----------------------"+scrollPanel.getVerticalScrollBar().isVisible());
                updateContentSize(e.getComponent().getSize());
            }
        });
        scrollPanel.getHorizontalScrollBar().setEnabled(false);
        scrollPanel.getVerticalScrollBar().setUnitIncrement(16);
        myEditToggle.setIcon(editModeOffIcon);
        myEditToggle.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                activeEditMode = !activeEditMode;
//                myEditToggle.setText(activeEditMode ? "Edit mode on" : "Edit mode off");
                myEditToggle.setIcon(activeEditMode ? editModeOnIcon : editModeOffIcon);
                for (@NotNull PreviewItem myPreviewItem : myPreviewItems) {
                    myPreviewItem.setEditMode(activeEditMode);
                }
            }
        });
        myAccordionButton.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                setSketchesVisible(!sketchesExpanded);
                if (sketchesExpanded) {
                    myAccordionButton.setIcon(accordionUpIcon);
                } else {
                    myAccordionButton.setIcon(accordionDownIcon);
                }

            }
        });
//        myEditMode.addActionListener(new ActionListener() {
//            @Override
//            public void actionPerformed(ActionEvent e) {
//                activeEditMode = !activeEditMode;
//                for (@NotNull PreviewItem myPreviewItem : myPreviewItems) {
//                    myPreviewItem.setEditMode(activeEditMode);
//                }
//            }
//        });
        ContentFactory contentFactory = ContentFactory.SERVICE.getInstance();
        @NotNull Content content = contentFactory.createContent(this.rootPanel, "", false);
        toolWindow.getContentManager().addContent(content);
        myUnpinButton.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                PreviewManager.getInstance(project).getPreviewHandler().setVisible(false);
                PreviewManager.getInstance(project).setPopupPreviewMode();
                isRegistered = false;
                clearSketches();
            }
        });
    }

    public void setSketchesVisible(boolean visible) {
        if (sketchesExpanded == visible)
            return;
        sketchesExpanded = visible;
        for (PreviewItem myPreviewItem : myPreviewItems) {
            if (myPreviewItem instanceof PreviewSketch) {
                ((PreviewSketch) myPreviewItem).getSVGPreviewContainer().setSketchPreviewVisible(sketchesExpanded);
            }
        }
        updateContentSize(rootPanel.getParent().getSize());
    }


    public void scrollToTop() {
        UIUtil.invokeLaterIfNeeded(new Runnable() {
            public void run() {
                scrollPanel.getVerticalScrollBar().setValue(0);
            }
        });
    }

    public void clearSketches() {
        myPreviewContent.removeAll();
        myPreviewItems.clear();
    }

    public void addSketches(@NotNull PreviewItem preview) {
        if (myPreviewItems.contains(preview)) {
            return;
        }
        preview.setEditMode(activeEditMode);
        preview.setToolWindowMode(true);
        myPreviewItems.add(preview);

        myPreviewContent.add(preview.getPanel());
        if (preview instanceof PreviewSketch) {
            ((PreviewSketch) preview).getSVGPanelExt().setVisible(true);
        }
        preview.setWidth(myPreviewContent.getWidth() - myPreviewContent.getInsets().left - myPreviewContent.getInsets().right, getScrollBarWidth());
    }

    public void setSketches(@NotNull Collection<PreviewItem> previews) {
        clearSketches();
        for (@NotNull PreviewItem preview : previews) {
            addSketches(preview);
        }
        updateContentSize(rootPanel.getParent().getSize());
    }


    private void updateContentSize(@NotNull Dimension size) {
        int width = rootPanel.getInsets().left + rootPanel.getInsets().right;
        width += scrollPanel.getInsets().left + scrollPanel.getInsets().right;
        width = size.width - width;
        int initialScrollbarWidth = getScrollBarWidth();
        for (@NotNull PreviewItem myPreviewItem : myPreviewItems) {
            myPreviewItem.setWidth(width, getScrollBarWidth());
        }
        if (initialScrollbarWidth != getScrollBarWidth()) {
            for (@NotNull PreviewItem myPreviewItem : myPreviewItems) {
                myPreviewItem.setWidth(width, getScrollBarWidth());
            }
        }
        myPreviewContent.revalidate();
        myPreviewContent.repaint();
    }

    private int getScrollBarWidth() {
        if (!scrollPanel.getVerticalScrollBar().isVisible()) {
            return 0;
        }
        int result = scrollPanel.getVerticalScrollBar().getWidth();
        result += scrollPanel.getVerticalScrollBar().getInsets().left + scrollPanel.getVerticalScrollBar().getInsets().right;
        return result;
    }

    public void setTitle(@NotNull String title) {
        this.myTitle = title;
        myToolWindow.setTitle(title);
    }


    public boolean isViewable() {
        return myToolWindow.isAvailable() && isRegistered;
    }


    /**
     * Sets the Sketchlink toolwindow to custom a custom content, e.g. information.
     */
    public void setContent(@NotNull JComponent content) {
        myPreviewContent.removeAll();
        myPreviewContent.add(content);
        myPreviewContent.revalidate();
        myPreviewContent.repaint();
    }

    @Nullable
    public ToolWindow getToolWindow() {
        return this.myToolWindow;
    }

    public boolean isShowing() {
        return myToolWindow.isVisible();
    }

    public boolean isActive() {
        return myToolWindow.isActive();
    }

    public void setActive(boolean active) {
        myToolWindow.setAvailable(active, null);
    }

    public void setOpen(boolean isOpen) {
        if (isOpen) {
            myToolWindow.show(null);
            updateContentSize(rootPanel.getParent().getSize());
        } else
            myToolWindow.hide(null);

    }

    public JPanel getRootPanel() {
        return this.rootPanel;
    }

}
