/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview;

import com.intellij.openapi.actionSystem.ActionGroup;
import com.intellij.openapi.actionSystem.ActionManager;
import com.intellij.openapi.actionSystem.ActionPlaces;
import com.intellij.openapi.actionSystem.ActionPopupMenu;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import com.intellij.ui.awt.RelativePoint;
import com.intellij.util.containers.ConcurrentHashSet;
import com.kitfox.svg.ShapeElement;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.LinkAnchor;
import plg.sketchlinks.LocalLinkAnchor;
import plg.sketchlinks.RemoteLinkAnchorS;
import plg.sketchlinks.SketchlinkManager;
import plg.ui.preview.svgwrapper.SVGPanelExt;
import plg.websocket.commands.builder.LoggingRequestBuilder;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * Created by Peter on 08.09.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class SketchHighlightMouseListener implements MouseMotionListener, MouseListener {


    @NotNull
    private final Project project;

    public SketchHighlightMouseListener(@NotNull Project project) {
        this.project = project;
    }

    @Override
    public void mouseClicked(@NotNull MouseEvent e) {
        Component component = e.getComponent();
        if (!(component instanceof SVGPanelExt)) {
            return;
        }
        @NotNull SVGPanelExt svgPanelExt = (SVGPanelExt) component;
        //should not be null
        @Nullable Set<String> highlightedIDs = svgPanelExt.getHighlightedIDs();
        //local navigation
        String sketchID = svgPanelExt.getDiagramEx().getSketchID();
        @Nullable LinkAnchor sketchAnchor = SketchlinkManager.getInstance(project).getKnownAnchor(sketchID);
        if (!(sketchAnchor instanceof RemoteLinkAnchorS)) {
            return;
        }
        if (e.getButton() == MouseEvent.BUTTON1) {
            if (highlightedIDs == null || highlightedIDs.size() == 0) {
                return;
            }
            //lookup the anchors of the markers
            @NotNull Set<LinkAnchor> selectedMarkers = new HashSet<>(highlightedIDs.size());
            for (@NotNull String highlightedID : highlightedIDs) {
                @Nullable LinkAnchor knownAnchor = SketchlinkManager.getInstance(project).getKnownAnchor(highlightedID);
                if (knownAnchor != null) {
                    selectedMarkers.add(knownAnchor);
                }
            }
            //if not empty, navigate to their linked artifacts
            if (selectedMarkers.size() > 0) {
                PreviewHandler previewHandler = PreviewManager.getInstance(project).getPreviewHandler();
                Set<LocalLinkAnchor> localLinkAnchors = SketchlinkManager.getInstance(project).browseToLocalAnchorsLinkedTo(selectedMarkers, new RelativePoint(e), (RemoteLinkAnchorS) sketchAnchor);
                if (localLinkAnchors.size() == 1) {
                    LocalLinkAnchor next = localLinkAnchors.iterator().next();
                    if (next.isValid())
                        previewHandler.SetContentFromAnchor(next, ((RemoteLinkAnchorS) sketchAnchor).getPreviewItem());
                }
                new LoggingRequestBuilder("local sketch browsing;").send(project);
            }
        }
        //open sketch in Webview
        if (e.getButton() == MouseEvent.BUTTON3) {
            ActionGroup group;
            if (ActionManager.getInstance().isGroup("Sketchlink.SketchHighlightContext")) { //NON-NLS
                group = (ActionGroup) ActionManager.getInstance().getAction("Sketchlink.SketchHighlightContext");
            } else {
                Logger.getInstance(SketchEditMouseListener.class)
                        .warn("Expected Actiongroup for Sketchlink.SVGDeleteMarker, but received only normal action");
                return;
            }
            ActionPopupMenu actionPopupMenu = ActionManager.getInstance().createActionPopupMenu(ActionPlaces.UNKNOWN, group);
            actionPopupMenu.getComponent().show(svgPanelExt, e.getPoint().x, e.getPoint().y);
//            sketchAnchor.navigateTo();
//            new LoggingRequestBuilder("open webview by sketch;").send(sketchAnchor.getProject());
        }
    }

    @Override
    public void mousePressed(MouseEvent e) {

    }

    @Override
    public void mouseReleased(MouseEvent e) {

    }

    @Override
    public void mouseEntered(MouseEvent e) {

    }

    @Override
    public void mouseExited(@NotNull MouseEvent e) {
        Component component = e.getComponent();
        if (!(component instanceof SVGPanelExt)) {
            return;
        }
        //disable highlighting
        @NotNull SVGPanelExt svgPanelExt = (SVGPanelExt) component;
        svgPanelExt.highlightMarkerAtPoint(null);
        svgPanelExt.setHighlightedIDs(null);
//        svgPanelExt.repaint();

    }

    @Override
    public void mouseDragged(MouseEvent e) {
//        System.out.println("MouseDragged: "+e.getPoint()+
//                "\n\t"+e.getButton()+
//                        "\n\t"+e.getWhen()+
//                        "\n\t"+e
//        );

    }

    @Override
    public void mouseMoved(@NotNull MouseEvent e) {
        if (!(e.getComponent() instanceof SVGPanelExt)) {
            return;
        }
        //disable highlighting
        @NotNull SVGPanelExt svgPanelExt = (SVGPanelExt) e.getComponent();
        //highlight sketch
        @NotNull List<ShapeElement> shapeElements = svgPanelExt.highlightMarkerAtPoint(e.getPoint());
        @NotNull Set<String> ids = new ConcurrentHashSet<>(shapeElements.size());
        for (@NotNull ShapeElement element : shapeElements) {
            String id = element.getId();
            if (id != null) {
                ids.add(id);
            }
        }
        //set list of hightliged ids in panel
        svgPanelExt.setHighlightedIDs(ids);

    }
}
