/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview;

import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.NotNull;
import plg.slApplicationState;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.RunnableFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

public class ViewsDialog<T> extends JDialog implements RunnableFuture<List<T>> {

    private JPanel contentPane;
    private JButton buttonOK;
    private JButton buttonCancel;
    private JList<T> list1;
    private JLabel labelHeader;
    private JCheckBox rememberChoiceForThisCheckBox;
    private List<T> result;


    public ViewsDialog() {
        setContentPane(contentPane);
        setModal(true);
        getRootPane().setDefaultButton(buttonOK);
        this.setTitle(slApplicationState.RESOURCE_BUNDLE.getString("open.webview"));
        rememberChoiceForThisCheckBox.setSelected(false);

        buttonOK.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onOK();
            }
        });

        buttonCancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onCancel();
            }
        });

// call onCancel() when cross is clicked
        setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                onCancel();
            }
        });

// call onCancel() on ESCAPE
        contentPane.registerKeyboardAction(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onCancel();
            }
        }, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
    }

    /**
     * Opens a Dialog and asks the user to select entries.
     *
     * @param availableEntries All Entries selectable by the user
     * @param resultEntries    Entries selected by the user, is empty if user canceled the dialog
     * @param allowMultiSelect allow selection of multiple entries
     * @return if true, the result should be saved for the session and reused if possible, false otherwise
     */
    public static <T> boolean showDialog(@NotNull List<T> availableEntries, @NotNull List<T> resultEntries, boolean allowMultiSelect) {
        @NotNull ViewsDialog<T> dialog = new ViewsDialog<>();
        dialog.setList(availableEntries);
        dialog.setMultiSelect(allowMultiSelect);
        List<T> results;
        UIUtil.invokeLaterIfNeeded(dialog);
        try {
            results = dialog.get();
        } catch (@NotNull InterruptedException | ExecutionException e) {
            JOptionPane.showMessageDialog(null, MessageFormat.format(slApplicationState.RESOURCE_BUNDLE.getString("an.error.occurred.while.waiting.on.data.returned.from.user.interaction.n.0"), e.getMessage()));
            resultEntries.clear();
            return false;
        }
        resultEntries.clear();
        resultEntries.addAll(results);
        return dialog.isRemember() && resultEntries.size() != 0;
    }

    private void onOK() {
        result = list1.getSelectedValuesList();
        dispose();
    }

    private void onCancel() {
        result = new ArrayList<>(0);
        dispose();
    }

    private void setList(@NotNull List<T> newList) {
        list1.removeAll();
        @NotNull DefaultListModel<T> tDefaultListModel = new DefaultListModel<>();
        for (T t : newList) {
            tDefaultListModel.addElement(t);
        }
        list1.setModel(tDefaultListModel);
        list1.setSelectedIndex(0);
    }

    private void setMultiSelect(boolean allow) {
        list1.setSelectionMode(allow ? ListSelectionModel.MULTIPLE_INTERVAL_SELECTION : ListSelectionModel.SINGLE_SELECTION);
    }

    private boolean isRemember() {
        return rememberChoiceForThisCheckBox.isSelected();
    }

    @Override
    public void run() {
        this.pack();
        Point p = UIUtil.getCenterPoint(UIUtil.getActiveWindow().getSize(), this.getSize());
        Point locationOnScreen = UIUtil.getActiveWindow().getLocationOnScreen();
        p.x += locationOnScreen.x;
        p.y += locationOnScreen.y;
        this.setLocation(p);
        this.setVisible(true);
    }

    @Override
    public boolean cancel(boolean mayInterruptIfRunning) {
        this.setVisible(false);
        result = new ArrayList<>(0);
        return true;
    }

    @Override
    public boolean isCancelled() {
        return result != null;
    }

    @Override
    public boolean isDone() {
        return result != null;
    }

    @Override
    public List<T> get() throws InterruptedException, ExecutionException {
        return result;
    }

    @Override
    public List<T> get(long timeout, TimeUnit unit) throws InterruptedException, ExecutionException, TimeoutException {
        return result;
    }

//    public static void main(String[] args) {
//        ViewsDialog dialog = new ViewsDialog();
//        dialog.pack();
//        dialog.setVisible(true);
//        System.exit(0);
//    }
}
