/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview.svgwrapper;

import com.kitfox.svg.Group;
import com.kitfox.svg.Rect;
import com.kitfox.svg.SVGRoot;
import com.kitfox.svg.ShapeElement;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.tools.SVGTools;

import java.awt.*;
import java.util.ArrayList;
import java.util.Collection;

/**
 * Created by Peter on 23.09.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class SVGDiagram2XML {
    private static final String xmlHeader = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
    private static final String svgTag = "<svg version=\"1.1\" id=\"insertID\" xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\">\n";
    private static final String svgTagClose = "</svg>\n";
    private static final String groupTag = "<g>\n";
    private static final String groupTagClose = "</g>\n";
    private static final String rectTag = "\t<rect id=\"insertID\" x=\"insertX\" y=\"insertY\" width=\"insertWidth\" height=\"insertHeight\" insertStyle />\n";
    private static final String imageTag = "<image x=\"0\" y=\"0\" height=\"insertHeightpx\" width=\"insertWidthpx\" xlink:href=\"/sketches/bitmap/insertID\" preserveAspectRatio=\"xMinYMin slice\"/>\n";


    @NotNull
    private static StringBuilder getElementString(@NotNull SVGRoot root, @NotNull String sketchID, @NotNull Dimension size) {
        @NotNull StringBuilder sb = new StringBuilder(svgTag.replaceAll("insertID", sketchID));
        sb.append(imageTag.replaceAll("insertID", sketchID).replaceAll("insertHeight", String.valueOf(size.height)).replaceAll("insertWidth", String.valueOf(size.width)));
        return sb;

    }

    @NotNull
    private static StringBuilder getStyle(@NotNull ShapeElement element) {
        @NotNull String styleStroke = SVGTools.getDataOf(element, SVGTools.SVGDrawing.STROKE);
        @NotNull String styleStrWidth = SVGTools.getDataOf(element, SVGTools.SVGDrawing.STROKE_WIDTH);
        @NotNull String fill = SVGTools.getDataOf(element, SVGTools.SVGDrawing.FILL);
        @NotNull String fillOpacity = SVGTools.getDataOf(element, SVGTools.SVGDrawing.FILL_OPACITY);

        @NotNull StringBuilder sb = new StringBuilder("style=\"");
        if (!styleStroke.isEmpty()) {
            sb.append("stroke: ");
            sb.append(styleStroke);
            sb.append("; ");
        }
        if (!styleStrWidth.isEmpty()) {
            sb.append("stroke-width: ");
            sb.append(styleStrWidth);
            sb.append("; ");
        }
        if (!fill.isEmpty()) {
            sb.append("fill: ");
            sb.append(fill);
            sb.append("; ");
        }
        if (!fillOpacity.isEmpty()) {
            sb.append("fill-opacity: ");
            sb.append(fillOpacity);
            sb.append("; ");
        }
        sb.append("\"");
        int i = sb.lastIndexOf(";");
        if (i < 0) {
            sb.setLength(0);
        } else {
            sb.deleteCharAt(i);
        }
        return sb;
    }

    @NotNull
    private static StringBuilder getElementString(@NotNull Rect rect) {
        String id = rect.getId();
        @NotNull Rectangle markerBounds = SVGDiagramEx.getMarkerBounds(rect);
        if (id == null) {
            return new StringBuilder();
        }
        @NotNull StringBuilder sb = new StringBuilder(rectTag
                .replaceAll("insertID", id)
                .replaceAll("insertHeight", String.valueOf(markerBounds.height))
                .replaceAll("insertWidth", String.valueOf(markerBounds.width))
                .replaceAll("insertX", String.valueOf(markerBounds.x))
                .replaceAll("insertY", String.valueOf(markerBounds.y))
                .replaceAll("insertStyle", getStyle(rect).toString())
        );
        return sb;
    }

    @NotNull
    private static StringBuilder getElementString(@NotNull Group group) {
        @NotNull StringBuilder result = new StringBuilder(groupTag);
        for (Object o : group.getChildren(null)) {
            if (o instanceof Rect) {
                result.append(getElementString((Rect) o));
            }
        }
        result.append(groupTagClose);
        return result;
    }

    @Nullable
    public static String export(@NotNull SVGDiagramEx diagramEx) {
        @NotNull Dimension size = diagramEx.getSize();
        @NotNull String sketchID = diagramEx.getSketchID();
        @NotNull Collection<String> linkedIDs = diagramEx.getLinkedIDs();
        //clear highlghting colors
        diagramEx.setLinkedIDs(new ArrayList<String>());
        diagramEx.highlightMarkerAtPoint(null, null);

        @NotNull StringBuilder result = new StringBuilder(xmlHeader);
        SVGRoot root = diagramEx.getRoot();
        if (root == null) {
            return null;
        }
        //svg start tag and image tag
        result.append(getElementString(root, sketchID, size));
        //marker group
        for (Object o : root.getChildren(null)) {
            //only add the marker group, which has no ID. skip any group that starts with temp in its id
            if (o instanceof Group && (((Group) o).getId() == null || !((Group) o).getId().startsWith("temp"))) {
                result.append(getElementString((Group) o));
            }
        }
        result.append(svgTagClose);
        diagramEx.setLinkedIDs(linkedIDs);
        return result.toString();
    }
}
