/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview.svgwrapper;

import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.project.Project;
import com.kitfox.svg.*;
import com.kitfox.svg.animation.AnimationElement;
import com.kitfox.svg.xml.StyleAttribute;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.tools.SVGTools;
import plg.tools.SketchlinkServerRESTUtil;

import javax.swing.*;
import java.awt.*;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.net.URI;
import java.util.*;
import java.util.List;

/**
 * Created by Peter on 08.09.2014.
 * Wrapper for SVGDiagram
 */
public class SVGDiagramEx {

    @NonNls
    public static final String CONTENT_CHANGED_Property = "content_changed";
    @NotNull
    private final SVGDiagram myDiagram;
    @NotNull
    private final Project project;
    @NotNull
    private Collection<String> myLinkedIDs = new HashSet<>();
    @NotNull
    private Collection<ShapeElement> myHighlightedElements = new HashSet<>();
    @NotNull
    private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);
    private boolean myIsEdited;
    private boolean myAutoUploadOnChange;
//    @NonNls
//    public static final String HIGHLIGHTED_ELEMENTS_Property = "highlighted_elements";
//    @NonNls
//    public static final String LINKED_ELEMENTS_Property = "currently_linked";


    /**
     * Creates a new instance of SVGDiagram
     */
    public SVGDiagramEx(@NotNull SVGDiagram diagram, @NotNull Project project) {
        myDiagram = diagram;
        this.project = project;
        checkDiagram();
        myAutoUploadOnChange = true;
        myIsEdited = false;
    }

    @NotNull
    private static float[] getSVGSize(@NotNull SVGDiagram diagram) throws SVGElementException {
        SVGRoot root = diagram.getRoot();
        @NotNull float[] result = new float[]{0.f, 0.f, 600.f, 300.f};
        if (root == null) {
            return result;
        }
        if (root.hasAttribute("viewBox", AnimationElement.AT_AUTO)) { //NON-NLS
            StyleAttribute attribute = root.getPresAbsolute("viewBox"); //NON-NLS
            result[2] = attribute.getFloatList()[2];
            result[3] = attribute.getFloatList()[3];
            return result;
        }
        if (root.hasAttribute("width", AnimationElement.AT_CSS)) { //NON-NLS

            result[2] = root.getStyleAbsolute("width").getFloatValue(); //NON-NLS
            result[3] = root.getStyleAbsolute("height").getFloatValue(); //NON-NLS
//            if (root.getStyleAbsolute("x")!=null) {
//                result[0] = root.getStyleAbsolute("x").getFloatValue(); //NON-NLS
//            }
//            if (root.getStyleAbsolute("y")!=null) {
//                result[1] = root.getStyleAbsolute("y").getFloatValue(); //NON-NLS
//            }
            return result;
        }
        for (Object o : root.getChildren(null)) {
            if (o instanceof ImageSVG) {
                @NotNull ImageSVG image = (ImageSVG) o;
                if (((ImageSVG) o).hasAttribute("width", AnimationElement.AT_AUTO)) { //NON-NLS
                    result[2] = image.getPresAbsolute("width").getFloatValue(); //NON-NLS
                    result[3] = image.getPresAbsolute("height").getFloatValue(); //NON-NLS
//                    result[0] = image.getPresAbsolute("x").getFloatValue(); //NON-NLS
//                    result[1] = image.getPresAbsolute("y").getFloatValue(); //NON-NLS
                    return result;
                }
            }
        }
        return result;
    }

    @NotNull
    public static Rectangle getMarkerBounds(@NotNull ShapeElement shape) {
        Shape shapeOfShape = shape.getShape();
        Rectangle bounds;
        if (shapeOfShape == null) {
            bounds = new Rectangle(shape.getPresAbsolute("x").getIntValue(), shape.getPresAbsolute("y").getIntValue(), shape.getPresAbsolute("width").getIntValue(), shape.getPresAbsolute("height").getIntValue());
        } else {
            bounds = shapeOfShape.getBounds();
        }
        if (bounds == null) {
            System.err.println("SVGDiagramEx: could not detect correct bounds for rectangle with id: " + shape.getId());
            return new Rectangle(0, 0, 10, 10);
        }
        return bounds;
    }

    /**
     * Removes a shapeelement from the current sketch.
     *
     * @param localChange true, if the method should not call setEdited() afterwards, the server is not updated
     */
    public void removeShapeElement(@NotNull ShapeElement element, boolean localChange) throws SVGElementException {
        SVGElement parent = element.getParent();
        if (parent == null) {
            return;
        }
        if (!getContainedMarkerElements().contains(element)) {
            throw new SVGElementException(element, "ShapeElement not contained in this diagram");
        }
        synchronized (myDiagram) {
            parent.removeChild(element);
        }
        updateTime();
        if (!localChange) {
            setEdited();
        }
    }

    @NotNull
    public Project getProject() {
        return project;
    }

    /**
     * Highlights the markers at the given point, returns their ShapeElements.
     * Can return a ShapeElement with the ID of the sketch if no other marker is highlighted
     *
     * @param point  Point of the Marker to highlight, null to disable highlight
     * @param result the list where to store highlighted Shapeelements to, null to disable highlight
     * @return true if the content of the svg has been changed
     */
    public boolean highlightMarkerAtPoint(@Nullable Point point, @Nullable List<ShapeElement> result) {
        if (point == null || result == null) {
            if (myHighlightedElements.size() == 0) {
                return false;
            }
            for (@NotNull ShapeElement element : myHighlightedElements) {
                resetShapeElement(element);
            }
            myHighlightedElements.clear();
            try {
                myDiagram.updateTime(0.0d);
            } catch (SVGException ignored) {
            }
            return true;
        }
        @NotNull List<ShapeElement> shapeElementsAt = getShapeElementsAt(point);
        boolean needsRefresh = false;
        //highlight elements and fill result
        for (@NotNull ShapeElement shapeElement : shapeElementsAt) {
            //skip sketch rectangle if other shapeelements are present
            if (shapeElement.getId() != null && shapeElement.getId().equals(getSketchID()) && shapeElementsAt.size() > 1) {
                continue;
            }
            result.add(shapeElement);
            //already highlighted
            if (myHighlightedElements.contains(shapeElement)) {
                continue;
            }
            needsRefresh = true;
            myHighlightedElements.add(shapeElement);
            synchronized (myDiagram) {
                SVGTools.ShapeElementHighlightedColors(shapeElement, project);
            }
        }
        //remove no longer highlighted markers
        @NotNull Iterator<ShapeElement> iterator = myHighlightedElements.iterator();
        while (iterator.hasNext()) {
            ShapeElement next = iterator.next();
            if (result.contains(next)) {
                continue;
            }
            needsRefresh = true;
            resetShapeElement(next);
            iterator.remove();
        }
        if (needsRefresh) {
            updateTime();
        }
        return needsRefresh;
    }

    private void resetShapeElement(@NotNull ShapeElement element) {
        synchronized (myDiagram) {
            if (myLinkedIDs.contains(element.getId())) {
                SVGTools.ShapeElementLinkedColors(element, project);
            } else {
                SVGTools.ShapeElementNormalColors(element, project);
            }
        }
    }

    /**
     * Gets the IDs in the Sketch that have "linked" coloring, ignores possible highlighting through mouse over.
     */
    @NotNull
    public Collection<String> getLinkedIDs() {
        return myLinkedIDs;
    }

    /**
     * Sets the SVGs marker colors to the state as if the mouse were outside
     *
     * @param ids the ids of the marker to color as linked, can include the sketch anchor id
     */
    public void setLinkedIDs(@NotNull Collection<String> ids) {
        myLinkedIDs.clear();
        myLinkedIDs.addAll(ids);
        @NotNull Set<String> containedIDs = getContainedIDs();
        myLinkedIDs.retainAll(containedIDs);
        SVGRoot root = myDiagram.getRoot();
        @NotNull Stack<SVGElement> stack = new Stack<>();
        stack.push(root);
        while (!stack.isEmpty()) {
            SVGElement element = stack.pop();
            if (element instanceof Group) {
                for (Object o : element.getChildren(null)) {
                    if (o instanceof SVGElement) {
                        stack.push((SVGElement) o);
                    }
                }
                continue;
            }
            //if it's a circle, rectangle, polygon, etc.
            if (element instanceof ShapeElement) {
                resetShapeElement((ShapeElement) element);
            } //if
        }
        updateTime();
    } //changecolorofIDElement

    @NotNull
    public String getSketchID() {
        SVGRoot root = myDiagram.getRoot();
        String result = root.getId();
        if (result == null) {
            return "Invalid SVG Format, no ID in svg-tag found: " + myDiagram.getXMLBase();
        }
        return result;
    }

    @NotNull
    private List<ShapeElement> getContainedMarkerElements() {
        @NotNull List<ShapeElement> result = new ArrayList<>();
        @Nullable Group markerGroup = getMarkerGroup();
        if (markerGroup == null) {
            return result;
        }
        for (Object o : markerGroup.getChildren(null)) {
            if (o instanceof ShapeElement) {
                result.add(((ShapeElement) o));
            }
        }
        return result;
    }

    @NotNull
    public Set<String> getContainedIDs() {
        @NotNull Set<String> result = new HashSet<>();
        String id = myDiagram.getRoot().getId();
        if (id != null) {
            result.add(id);
        }
        for (@NotNull ShapeElement shapeElement : getContainedMarkerElements()) {
            if (shapeElement.getId() != null)
                result.add(shapeElement.getId());
        }
        return result;
    }

    /**
     * Checks the Diagram for known incompatible settings from the Sketchlink-Server
     * Adds a fake border in a new group if not already present
     */
    private void checkDiagram() {
//        System.out.println("Before: \n------------------------------");
//        printSvgToStdOut(diagram);
        SVGRoot root = myDiagram.getRoot();
        if (root == null) {
            return;
        }
        try {
            //set the xy part of the viewbox to 0 0, can contain coordinates from the last viewer
            @NotNull float[] svgSize = getSVGSize(myDiagram);
            float width = svgSize[2];
            float height = svgSize[3];
            if (root.hasAttribute("viewBox", AnimationElement.AT_XML)) { //NON-NLS
                @NotNull StringBuilder sb = new StringBuilder();
                StyleAttribute attribute = root.getPresAbsolute("viewBox"); //NON-NLS
                sb.append("0 0 ");
                if (attribute.getStringList().length >= 4) {
                    sb.append(attribute.getStringList()[2]);
                    sb.append(" ");
                    sb.append(attribute.getStringList()[3]);
                    width = attribute.getFloatList()[2];
                    height = attribute.getFloatList()[3];
                }
                root.setAttribute("viewBox", AnimationElement.AT_AUTO, sb.toString()); //NON-NLS
                root.updateTime(0);
            }
            for (@NotNull Group group : getTopGroups()) {
                if (group.getId() != null && group.getId().startsWith("temp")) {
                    return;
                }
            }

            addFakeBorder(root, width, height);
            setLinkedIDs(new ArrayList<String>(0));
            myDiagram.updateTime(0.0d);
        } catch (SVGException ignored) {

        }
//        SVGTools.printSvgToStdOut(myDiagram);
    }

    private void addFakeBorder(@NotNull SVGRoot root, float width, float height) throws SVGException {
        @Nullable Group topGroup = getMarkerGroup();
        if (null == topGroup) {
            return;
        }
        @NotNull String topID = getSketchID();
        @NotNull Group tempGroup = new Group();
        SVGTools.setDataOf(tempGroup, SVGTools.SVGData.ID, "temp");
        //insert fake rectangle for highlighting
        @NotNull Rect rectangle = new Rect();
        SVGTools.setDataOf(rectangle, SVGTools.SVGData.ID, topID);
        SVGTools.setDataOf(rectangle, SVGTools.SVGDrawing.FILL_OPACITY, "0");
        SVGTools.setDataOf(rectangle, SVGTools.SVGDrawing.FILL, "#333333");
        SVGTools.setDataOf(rectangle, SVGTools.SVGDrawing.STROKE, "#FFFFFF"); //NON-NLS
        SVGTools.setDataOf(rectangle, SVGTools.SVGDrawing.STROKE_WIDTH, String.valueOf(width * 0.010));
        SVGTools.setDataOf(rectangle, SVGTools.SVGData.WIDTH, String.valueOf(width));
        SVGTools.setDataOf(rectangle, SVGTools.SVGData.HEIGHT, String.valueOf(height));
        SVGTools.setDataOf(rectangle, SVGTools.SVGData.X, "0");
        SVGTools.setDataOf(rectangle, SVGTools.SVGData.Y, "0");
        synchronized (myDiagram) {
            //add to diagram
            root.loaderAddChild(null, tempGroup);
            tempGroup.loaderAddChild(null, rectangle);
        }
        updateTime();
    }

    /**
     * Adds the provided ShapeElement to the Marker group.
     *
     * @param element  the element to add
     * @param anchorID the id for this element.
     * @param x        the x position
     * @param y        the y position
     * @param width    the initial width
     * @param height   the initial height
     * @return the created and set-up element or null, if the element couldn't be added
     */
    @Nullable
    public ShapeElement addMarker(@NotNull ShapeElement element, @NotNull String anchorID, int x, int y, int width, int height) {
        SVGTools.setDataOf(element, SVGTools.SVGData.ID, anchorID);
        SVGTools.setDataOf(element, SVGTools.SVGData.WIDTH, String.valueOf(width));
        SVGTools.setDataOf(element, SVGTools.SVGData.HEIGHT, String.valueOf(height));
        SVGTools.setDataOf(element, SVGTools.SVGData.X, String.valueOf(x));
        SVGTools.setDataOf(element, SVGTools.SVGData.Y, String.valueOf(y));

        SVGTools.setDataOf(element, SVGTools.SVGDrawing.FILL_OPACITY, "0");
        SVGTools.setDataOf(element, SVGTools.SVGDrawing.FILL, "#333333");
        SVGTools.setDataOf(element, SVGTools.SVGDrawing.STROKE, "#FFFFFF"); //NON-NLS
        SVGTools.setDataOf(element, SVGTools.SVGDrawing.STROKE_WIDTH, String.valueOf(width * 0.010));
        @Nullable Group markerGroup = getMarkerGroup();
        if (markerGroup == null) {
            return null;
        }

        try {
            synchronized (myDiagram) {
                markerGroup.loaderAddChild(null, element);
            }
//            markerGroup.getRoot().updateTime(0.d);
        } catch (SVGException e) {
            return null;
        }

        resetShapeElement(element);
        return element;
    }

    @NotNull
    public Dimension getSize() {
        @NotNull float[] svgSize = new float[0];
        try {
            svgSize = getSVGSize(myDiagram);
        } catch (SVGElementException e) {
            e.printStackTrace();
        }
        return new Dimension(Math.round(svgSize[2]), Math.round(svgSize[3]));
    }

    /**
     * Gets the Diagram for this instance
     */
    @NotNull
    public SVGDiagram getDiagram() {
        return myDiagram;
    }

    /**
     * Gets a list of Shape elements that are at the point-coordinates of the Sketch
     */
    @NotNull
    public List<ShapeElement> getShapeElementsAt(@NotNull Point point) {
        @NotNull List<SVGElement> elementList = getElementsAt(point);
        @NotNull List<ShapeElement> result = new ArrayList<>(elementList.size());
        for (SVGElement svgElement : elementList) {
            if (svgElement instanceof ShapeElement) {
                result.add((ShapeElement) svgElement);
            }
        }
        return result;
    }

    @NotNull
    public List<String> getShapeElementsIDsAt(@NotNull Point point) {
        @NotNull List<String> result = new ArrayList<>();
        for (@NotNull ShapeElement shapeElement : getShapeElementsAt(point)) {
            if (shapeElement.getId() != null) {
                result.add(shapeElement.getId());
            }
        }
        return result;
    }

    /**
     * Get the defined SVGElements inside the svg-File at the supplied Point.
     *
     * @param p the point to look for elements
     * @return A List of SVGElements, including "background"
     */
    @NotNull
    public List<SVGElement> getElementsAt(@NotNull Point p) {
        List foundTargetsList;
        @NotNull List<SVGElement> result = new ArrayList<>(4);
        try {
            //returns a List of List, each list contains a "path"
            foundTargetsList = myDiagram.pick(p, true, null);
        } catch (SVGException e) {
            return result;
        }
        for (Object o : foundTargetsList) {
            if (!(o instanceof List)) {
                continue;
            }
            @NotNull List innerList = (List) o;
            Object lastElement = innerList.get(innerList.size() - 1);
            if (lastElement instanceof SVGElement) {

                result.add((SVGElement) lastElement);
                //some debug and comparison output for the current lastElement
//                if (lastElement instanceof ShapeElement) {
//                    ShapeElement shapeElement = (ShapeElement) lastElement;
//                    Rectangle2D r2d = null;
//                    try {
//                        r2d = shapeElement.getBoundingBox();
//                    } catch (SVGException e) {
//                        e.printStackTrace();
//                        continue;
//                    }
//                    int x = new BigDecimal(Math.round(r2d.getX() * this.getWidth())).intValue();
//                    int y = new BigDecimal(Math.round(r2d.getY() * this.getHeight())).intValue();
//                    int width = new BigDecimal(Math.round(r2d.getWidth() * this.getWidth())).intValue() + x;
//                    int height = new BigDecimal(Math.round(r2d.getHeight() * this.getHeight())).intValue() + y;
//                    Rectangle rectangle = new Rectangle(x, y, width, height);
//
//                    System.out.println("Highlighting " + shapeElement.getId() + " P:" + p + " R:" + rectangle + " "+r2d);
//                    outputSizeOf(shapeElement);
//                }
            }
        }
        return result;
    }

    /**
     * Gets the group that contains markers, detected by missing an ID
     * <p/>
     * returns null, if no Diagram loaded or the root element has no matching child
     */
    @Nullable
    public Group getMarkerGroup() {
        for (@NotNull Group group : getTopGroups()) {
            if (group.getId() == null) {
                return group;
            }
        }
        return null;
    }

    /**
     * Gets a list of all group declarations at the top level
     */
    @NotNull
    public List<Group> getTopGroups() {
        SVGRoot root = getDiagram().getRoot();
        @NotNull List<Group> result = new ArrayList<>(5);
        for (Object element : root.getChildren(null)) {
            if (element instanceof Group) {
                result.add((Group) element);
            }
        }
        return result;
    }

    public URI getXMLBase() {
        return myDiagram.getXMLBase();
    }

    public SVGUniverse getUniverse() {
        return myDiagram.getUniverse();
    }

    public SVGRoot getRoot() {
        return myDiagram.getRoot();
    }

    /**
     * Calculates the Size and position of the shapeElement in full SVG resolution and prints the result to stdout
     * for debugging purposes.
     */
    private void outputSizeOf(@NotNull ShapeElement shapeElement) {
        float x = shapeElement.getPresAbsolute("x").getFloatValue() * myDiagram.getWidth();
        float y = shapeElement.getPresAbsolute("y").getFloatValue() * myDiagram.getHeight();
        float width = shapeElement.getPresAbsolute("width").getFloatValue() * myDiagram.getWidth();
        float height = shapeElement.getPresAbsolute("height").getFloatValue() * myDiagram.getHeight();
        System.out.println(x + " " + y + " " + (width + x) + " " + (height + y));

    }

    /**
     * Looks up the Marker with the provided id in this svg.
     *
     * @param id the id of the shapeelement to look up
     * @return the shapeelement of the marker or null, if it does not exist
     */
    @Nullable
    public ShapeElement getShapeElement(@NotNull String id) {
        @Nullable Group markerGroup = getMarkerGroup();
        if (markerGroup == null) {
            return null;
        }
        for (Object o : markerGroup.getChildren(null)) {
            if (o instanceof ShapeElement && id.equals(((ShapeElement) o).getId())) {
                return (ShapeElement) o;
            }
        }
        return null;
    }

    public void setEdited() {
        myIsEdited = true;
        if (myAutoUploadOnChange) {
            ApplicationManager.getApplication().executeOnPooledThread(new Runnable() {
                @Override
                public void run() {
                    try {
                        uploadCurrentSketch();
                    } catch (IOException e) {
                        JOptionPane.showMessageDialog(null, "Could not update sketch on the server.\nReason:\n\n" + e.getMessage());
                    }
                }
            });

        }
    }

    /**
     * true, if the sketch's svg file has been significantly changed from the server,
     * false, if the sketch is the same as the server's.
     * <p/>
     * Ignores changes of style information.
     */
    public boolean isEdited() {
        return myIsEdited;
    }

    /**
     * If true, the sketch is uploaded to the server after each change,
     * if false, no automatic uploading
     */
    public void setAutoUploadOnChange(boolean upload) {
        myAutoUploadOnChange = upload;
    }


    /**
     * Tries to upload the current SVG to the Sketchlink server currently defined in slProjectComp.
     * Should not be called from the UI thread.
     *
     * @return true if upload returned http code 200 or SVG hasn't been edited, false otherwise
     */
    public boolean uploadCurrentSketch() throws IOException {
        if (!isEdited())
            return true;
        String[] strings = SketchlinkServerRESTUtil.updateSVG(this, null);
        if (strings == null || !"200".equals(strings[2]))
            return false;
        myIsEdited = false;
        return true;
    }

    public boolean uploadCurrentSketch(@NotNull String serverUrl) throws IOException {
        if (!isEdited())
            return true;
        String[] strings = SketchlinkServerRESTUtil.updateSVG(serverUrl, this, null);
        if (strings == null || !"200".equals(strings[2]))
            return false;
        myIsEdited = false;
        return true;
    }

    public void updateTime() {
        synchronized (myDiagram) {
            try {
                getDiagram().updateTime(0.d);
            } catch (SVGException e) {
                e.printStackTrace();
            }
        }
    }
}
