/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview.svgwrapper;

import com.intellij.openapi.util.Pair;
import com.kitfox.svg.Rect;
import com.kitfox.svg.SVGElementException;
import com.kitfox.svg.ShapeElement;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.tools.SVGTools;

import java.awt.*;
import java.util.UUID;

/**
 * Created by Peter on 19.09.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class SVGEditSupport {
    @NotNull
    private final SVGPanelExt myPanel;
    @NotNull
    private SVGAction myLockedAction = SVGAction.NONE;
    @Nullable
    private ShapeElement myLockedShapeElement = null;
    /**
     * The size and location of myLockedShapeElement when it is locked
     */
    @Nullable
    private Rectangle myLockedStartBounds;
    /**
     * Translate the point when an action is locke to the "nearest" edge and remember the distance.
     * Using deltas between the "current" and the "last" point results in rounding errors.
     */
    private Point myOffsetPoint;
    /**
     * Is always the corner on the opposite site, whose position should not changed during resizing.
     * Calculate width/height of the rectangle as the difference between this and the updated point.
     */
    private Point myResizeOppositeCorner;


    public SVGEditSupport(@NotNull SVGPanelExt svgPanel) {
        myPanel = svgPanel;
    }

    @NotNull
    public SVGDiagramEx getDiagram() {
        return myPanel.getDiagramEx();
    }


    @Nullable
    private ShapeElement getSupportedShapeElementAt(@NotNull Point point) {
        for (@NotNull ShapeElement shapeElement : getDiagram().getShapeElementsAt(point)) {
            if (shapeElement.getId() != null && shapeElement.getId().startsWith("m-")) {
                return shapeElement;
            }
        }
        return null;
    }

    @Nullable
    public String unlockAction() {
        return unlockAction(null);
    }

    /**
     * Unlocks the currently active action.
     * If no action is currently locked, returns null.
     *
     * @param point the current point
     * @return the ID of the current marker if the action was CREATE and a marker was created, null otherwise
     */
    @Nullable
    public String unlockAction(@Nullable Point point) {
        @Nullable String result = null;
        if (myLockedShapeElement == null) {
            return null;
        }
        if (myLockedAction == SVGAction.CREATE) {
            @NotNull Rectangle bounds = SVGDiagramEx.getMarkerBounds(myLockedShapeElement);
            //delete new rectangle if smaller than 10x10 pixel in current drawing size
            @Nullable Point p;
            p = myPanel.convertVisualPointToSVGPoint(new Point(15, 15));
            if (bounds.width < p.x || bounds.height < p.y) {
                //remove temporary marker
                try {
                    getDiagram().removeShapeElement(myLockedShapeElement, true);
                } catch (SVGElementException ignored) {
                }
            } else {
                //tell the diagram it has been edited
                getDiagram().setEdited();
                result = myLockedShapeElement.getId();
//                JOptionPane.showMessageDialog(null,"New Marker created with ID:\n\t"+myLockedShapeElement.getId()+"\n"+(knownAnchor instanceof RemoteLinkAnchorS ? "Successfully.":""));
            }
        }
        myLockedAction = SVGAction.NONE;
        myLockedShapeElement = null;
        myLockedStartBounds = null;
        myResizeOppositeCorner = null;
        return result;
    }

    /**
     * Locks the currently available action for a ShapeElement at the point.
     * All subsequent calls to action(point) will update the ShapeElement.
     */
    @Nullable
    public Pair<SVGAction, ShapeElement> lockAction(@NotNull Point point) {
        unlockAction(point);
        @NotNull SVGAction action = getActionForPoint(point);
        @Nullable ShapeElement shapeElement = getSupportedShapeElementAt(point);

        if (action == SVGAction.NONE || (shapeElement == null && action != SVGAction.CREATE)) {
            return null;
        }
        if (action == SVGAction.CREATE && shapeElement == null) {
            @NotNull String id = "m-" + UUID.randomUUID();
            myLockedShapeElement = getDiagram().addMarker(new Rect(), id, point.x, point.y, 2, 2);
            myLockedStartBounds = new Rectangle(point.x, point.y, 2, 2);
        } else {
            myLockedStartBounds = SVGDiagramEx.getMarkerBounds(shapeElement);
            myLockedShapeElement = shapeElement;
        }
        myLockedAction = action;
        switch (myLockedAction) {
            //use topleft point for reference
            case MOVE:
            case TOPLEFT_MOVE:
            case TOP_MOVE:
            case ROTATE:
            case LEFT_MOVE:
                myOffsetPoint = new Point(point.x - myLockedStartBounds.x, point.y - myLockedStartBounds.y);
                myResizeOppositeCorner = new Point(myLockedStartBounds.x + myLockedStartBounds.width, myLockedStartBounds.y + myLockedStartBounds.height);
                break;
            //use topright point for reference
            case TOPRIGHT_MOVE:
            case RIGHT_MOVE:
                myOffsetPoint = new Point(point.x - myLockedStartBounds.x - myLockedStartBounds.width, point.y - myLockedStartBounds.y);
                myResizeOppositeCorner = new Point(myLockedStartBounds.x, myLockedStartBounds.y + myLockedStartBounds.height);
                break;
            //use bottom left point for reference
            case BOTTOMLEFT_MOVE:
                myOffsetPoint = new Point(point.x - myLockedStartBounds.x, point.y - myLockedStartBounds.y - myLockedStartBounds.height);
                myResizeOppositeCorner = new Point(myLockedStartBounds.x + myLockedStartBounds.width, myLockedStartBounds.y);
                break;
            //use bottom right point for reference
            case CREATE:
            case BOTTOM_MOVE:
            case BOTTOMRIGHT_MOVE:
                myOffsetPoint = new Point(point.x - myLockedStartBounds.x - myLockedStartBounds.width, point.y - myLockedStartBounds.y - myLockedStartBounds.height);
                myResizeOppositeCorner = new Point(myLockedStartBounds.x, myLockedStartBounds.y);
                break;
        }
//        System.out.println("Action locked with "+myLockedStartBounds+" "+myOffsetPoint+" "+point);
        return new Pair<>(myLockedAction, myLockedShapeElement);
    }

    public boolean isActionLocked() {
        return myOffsetPoint != null && myLockedShapeElement != null && myLockedAction != SVGAction.NONE;
    }

    public void action(@NotNull Point currentPoint) {
        if (!isActionLocked() || myLockedShapeElement == null || myLockedStartBounds == null) {
            return;
        }
        @NotNull Point usePoint = new Point(currentPoint.x - myOffsetPoint.x, currentPoint.y - myOffsetPoint.y);
        @NotNull Rectangle oldBounds = myLockedStartBounds;
        int x = Math.min(myResizeOppositeCorner.x, usePoint.x);
        int y = Math.min(myResizeOppositeCorner.y, usePoint.y);
        int width = Math.abs(myResizeOppositeCorner.x - usePoint.x);
        int height = Math.abs(myResizeOppositeCorner.y - usePoint.y);
        switch (myLockedAction) {
            case LEFT_MOVE:
            case RIGHT_MOVE:
                y = oldBounds.y;
                height = oldBounds.height;
                break;
            case TOP_MOVE:
            case BOTTOM_MOVE:
                x = oldBounds.x;
                width = oldBounds.width;
                break;
            case MOVE:
                x = usePoint.x;
                y = usePoint.y;
                height = oldBounds.height;
                width = oldBounds.width;
        }
        @NotNull Dimension diagramSize = getDiagram().getSize();
        x = x < 0 ? 0 : (x + width < diagramSize.width ? x : diagramSize.width - width);
        y = y < 0 ? 0 : (y + height < diagramSize.height ? y : diagramSize.height - height);
        SVGTools.setDataOf(myLockedShapeElement, SVGTools.SVGData.X, String.valueOf(x));
        SVGTools.setDataOf(myLockedShapeElement, SVGTools.SVGData.Y, String.valueOf(y));
        SVGTools.setDataOf(myLockedShapeElement, SVGTools.SVGData.WIDTH, String.valueOf(width));
        SVGTools.setDataOf(myLockedShapeElement, SVGTools.SVGData.HEIGHT, String.valueOf(height));
    }


    /**
     * @param point Point in SVG Coordinates
     */
    @NotNull
    public SVGAction getActionForPoint(@NotNull Point point) {
        @Nullable ShapeElement shapeElement = getSupportedShapeElementAt(point);
        if (shapeElement == null) {
            return SVGAction.CREATE;
        }
        @Nullable Point defaultDistances = myPanel.convertVisualPointToSVGPoint(new Point(7, 7));
        @NotNull Rectangle markerBounds = SVGDiagramEx.getMarkerBounds(shapeElement);
        //calculate the size of the possible distances to the borders when resizing should occur
        int x_distance = defaultDistances.x;
        x_distance = (x_distance * 2 > markerBounds.width) ? markerBounds.width / 2 : x_distance;
        int y_distance = defaultDistances.y;
        y_distance = (y_distance * 2 > markerBounds.height) ? markerBounds.height / 2 : y_distance;

        //test horizontal,
        int choice;
        if (point.x < markerBounds.x + x_distance) {
            choice = 1;
        } else {
            if (point.x > markerBounds.x + markerBounds.width - x_distance) {
                choice = 3;
            } else {
                choice = 2;
            }
        }
        //test vertical
        if (point.y < markerBounds.y + y_distance) {
            choice += 10;
        } else {
            if (point.y > markerBounds.y + markerBounds.height - y_distance) {
                choice += 30;
            } else {
                choice += 20;
            }
        }
//        System.out.println("SVGEdit: "+markerBounds+" "+ x_distance +" "+ y_distance +" "+choice);
        SVGAction result;
        switch (choice) {
            //top row
            case 11:
                result = SVGAction.TOPLEFT_MOVE;
                break;
            case 12:
                result = SVGAction.TOP_MOVE;
                break;
            case 13:
                result = SVGAction.TOPRIGHT_MOVE;
                break;
            //middle row
            case 21:
                result = SVGAction.LEFT_MOVE;
                break;
            case 22:
                result = SVGAction.MOVE;
                break; //center
            case 23:
                result = SVGAction.RIGHT_MOVE;
                break;
            //bottom row
            case 31:
                result = SVGAction.BOTTOMLEFT_MOVE;
                break;
            case 32:
                result = SVGAction.BOTTOM_MOVE;
                break;
            case 33:
                result = SVGAction.BOTTOMRIGHT_MOVE;
                break;
            //nothing, rotation is not possible for now
            default:
                result = SVGAction.NONE;
        }
        return result;
    }


    public static enum SVGAction {
        //move
        MOVE,
        //resize
        TOP_MOVE,
        LEFT_MOVE,
        RIGHT_MOVE,
        BOTTOM_MOVE,
        //corner resize over two borders
        TOPLEFT_MOVE,
        TOPRIGHT_MOVE,
        BOTTOMLEFT_MOVE,
        BOTTOMRIGHT_MOVE,
        //rotate--not yet supported, current implementation requires alignment to coordinate axis
        ROTATE,
        //create new
        CREATE,
        //nothing supported
        NONE
    }

}
