/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview.svgwrapper;

import com.intellij.openapi.actionSystem.DataKey;
import com.intellij.openapi.actionSystem.DataProvider;
import com.intellij.openapi.util.Pair;
import com.kitfox.svg.SVGElement;
import com.kitfox.svg.ShapeElement;
import com.kitfox.svg.app.beans.SVGPanel;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.awt.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;


/**
 * Created by Peter on 17.03.14.
 * Part of Sketchlink Plugin for IntelliJ
 * <p/>
 * Wrapper Class for SVGSalamanders SVGPanel, contains several helper Methods to find the SVGElement under the Mouse cursor
 * and extract Data from SVG Files, which were needed for the original SVG File version.
 * TODO: Put SVGPanelEx and SVGDiagramEx in the same interface, since panelex only translates the coordinate system for the diagram.
 */
public class SVGPanelExt extends SVGPanel implements DataProvider {
    public static final DataKey<SVGPanelExt> DATA_KEY = DataKey.create("sketchlink.svgPanel");
    @NonNls
    public static final String HIGHLIGHTED_IDS_Property = "highlighted_ids";
    @NonNls
    public static final String SVGDIAGRAM_EX_Property = "svgdiagram_ex";
    @NotNull
    private final SVGEditSupport myEditSupport;
    private SVGDiagramEx myDiagram;
    @Nullable
    private Set<String> myHighlightedIDs;
    private Point lastPoint;
    private SVGPreviewContainer SVGPreviewContainer;


    public SVGPanelExt() {
        super();
        this.setAntiAlias(true);
        this.setAutosize(SVGPanel.AUTOSIZE_BESTFIT);
        myEditSupport = new SVGEditSupport(this);
        lastPoint = new Point(0, 0);
        //super.setAutosize(SVGPanel.AUTOSIZE_HORIZ);
    }

    @NotNull
    public Point getLastPoint() {
        return lastPoint;
    }

    public void setLastPoint(@NotNull Point point) {
        if (this.contains(point))
            lastPoint = point;
    }

    @NotNull
    public SVGEditSupport.SVGAction getActionForPoint(@NotNull Point point) {
        setLastPoint(point);
        return myEditSupport.getActionForPoint(convertVisualPointToSVGPoint(point));
    }

    @Nullable
    public String unlockAction() {
        String s = myEditSupport.unlockAction();
        updateTime();
        return s;
    }

    @Nullable
    public Pair<SVGEditSupport.SVGAction, ShapeElement> lockAction(@NotNull Point point) {
        setLastPoint(point);
        return myEditSupport.lockAction(convertVisualPointToSVGPoint(point));
    }

    public void action(@NotNull Point point) {
        setLastPoint(point);
        if (!this.contains(point)) {
            unlockAction();
            return;
        }
        myEditSupport.action(convertVisualPointToSVGPoint(point));
        updateTime();
    }

    @Nullable
    public Set<String> getHighlightedIDs() {
        return myHighlightedIDs;
    }

    /**
     * Mouse highlighting
     */
    public void setHighlightedIDs(@Nullable Set<String> newValue) {
        if (myHighlightedIDs == newValue || (myHighlightedIDs != null && myHighlightedIDs.equals(newValue))) {
            return;
        }
        @Nullable Set<String> oldVal = myHighlightedIDs;
        myHighlightedIDs = newValue;
        this.firePropertyChange(HIGHLIGHTED_IDS_Property, oldVal, newValue);
    }

    public SVGDiagramEx getDiagramEx() {
        return myDiagram;
    }

    public void setDiagramEx(@NotNull SVGDiagramEx diagramEx) {
        if (diagramEx == myDiagram) {
            return;
        }
        SVGDiagramEx oldVal = myDiagram;
        myDiagram = diagramEx;
        this.setSvgUniverse(diagramEx.getUniverse());
        this.setSvgURI(diagramEx.getXMLBase());
        this.firePropertyChange(SVGDIAGRAM_EX_Property, oldVal, myDiagram);
    }

    @Nullable
    @Contract("null -> null")
    public Point convertVisualPointToSVGPoint(@Nullable Point point) {
        if (point == null) {
            return null;
        }
        @NotNull Pair<Float, Float> xyScalingFactors = getXYScalingFactors();
        float dx = xyScalingFactors.first;
        float dy = xyScalingFactors.second;
        @NotNull Point p = new Point(point);
        //the svg lookup function only works with "unscaled" coordinates
        //recalculate the point to svg coordinates
        int autosize = this.getAutosize();
        if (SVGPanelExt.AUTOSIZE_BESTFIT == autosize
                || SVGPanelExt.AUTOSIZE_HORIZ == autosize
                || SVGPanelExt.AUTOSIZE_VERT == autosize) {
            //most common case
            if (dx < dy) {
                p.x = Math.round(p.x * dy);
                p.y = Math.round(p.y * dy);
            } else {
                p.x = Math.round(p.x * dx);
                p.y = Math.round(p.y * dx);
            }
        } else {
            if (autosize == SVGPanelExt.AUTOSIZE_STRETCH) {
                p.x = Math.round(p.x * dx);
                p.y = Math.round(p.y * dy);
            }
            //AUTOSIZE_NONE doesn't need a recalculation of the point
        }
//        System.out.println("Point from "+point+" to "+p+" with scaling factors: dx="+dx+" dy="+dy+" Size: "+diagram.getWidth()+" "+diagram.getHeight());
        return p;
    }

//    /**
//     * Gets the first element after the SVGRoot element.
//     * <p/>
//     * returns null, if no Diagram loaded or the root element has no child of type svgelement
//     */
//    @Nullable
//    public Group getTopGroup() {
//        SVGDiagram svgDiagram = getDiagram();
//        if (svgDiagram == null)
//            return null;
//        SVGRoot root = svgDiagram.getRoot();
//        for (Object o : root.getChildren(null)) {
//            if (o instanceof Group) {
//                return (Group) o;
//            }
//        }
//        return null;
//    }

    @NotNull
    @Contract("null -> null")
    public List<ShapeElement> highlightMarkerAtPoint(@Nullable Point point) {
        if (point != null)
            setLastPoint(point);
        @Nullable Point p = convertVisualPointToSVGPoint(point);
        @NotNull List<ShapeElement> result = new ArrayList<>(3);
        if (getDiagramEx().highlightMarkerAtPoint(p, result)) {
            this.repaint();
        }
//        System.out.println("SVGPanelEx: "+point+" to "+p+" found: "+ result.size() );
        return result;
    }

    @NotNull
    public List<ShapeElement> getShapeElementsAt(@NotNull Point point) {
        setLastPoint(point);
        return getDiagramEx().getShapeElementsAt(convertVisualPointToSVGPoint(point));
//        List<SVGElement> elementList = getElementsAt(convertVisualPointToSVGPoint(point));
//        List<ShapeElement> result = new ArrayList<>(elementList.size());
//        for (SVGElement svgElement : elementList) {
//            if (svgElement instanceof ShapeElement) {
//                result.add((ShapeElement) svgElement);
//            }
//        }
//        return result;
    }

    @NotNull
    public List<String> getShapeElementIDsAt(@NotNull Point point) {
        setLastPoint(point);
        return getDiagramEx().getShapeElementsIDsAt(convertVisualPointToSVGPoint(point));
    }


    /**
     * Get the defined SVGElements inside the svg-File at the supplied Point of the rendered SVGPanel.
     *
     * @param point the point to look for elements
     * @return A List of SVGElements, including "background"
     */
    @NotNull
    public List<SVGElement> getElementsAt(@NotNull Point point) {
        setLastPoint(point);
        @Nullable Point p = convertVisualPointToSVGPoint(point);
        return getDiagramEx().getElementsAt(p);
    }


    /**
     * Calculates the Size and position of the shapeElement in full SVG resolution and prints the result to stdout
     * for debugging purposes
     */
    private void outputSizeOf(@NotNull ShapeElement shapeElement) {
        float x = shapeElement.getPresAbsolute("x").getFloatValue() * getSVGWidth();
        float y = shapeElement.getPresAbsolute("y").getFloatValue() * getSVGHeight();
        float width = shapeElement.getPresAbsolute("width").getFloatValue() * getSVGWidth();
        float height = shapeElement.getPresAbsolute("height").getFloatValue() * getSVGHeight();
        System.out.println(x + " " + y + " " + (width + x) + " " + (height + y));

    }

    /**
     * Gets the Anchor-IDs in this Diagram at the supplied Point
     *
     * @param point the point to look up
     * @return all IDs of Elements at the point, empty if no Element is at this point or it has no ID
     */
    @NotNull
    public List<String> getShapeTargetIDsAtPoint(@NotNull Point point) {
        setLastPoint(point);
        @NotNull List<String> result = new ArrayList<>(3);
        for (SVGElement element : getElementsAt(point)) {
            if (element instanceof ShapeElement) {
                String id = element.getId();
                if (id != null && !id.trim().isEmpty()) {
                    result.add(id);
                }
            }
        }
        return result;
    }


    public int setHorizontalStretch(int width) {
        float ratio = 1.f * myDiagram.getSize().height / myDiagram.getSize().width;
        int height = Math.round(ratio * width) + 1;
//        float hscale = 1.f * width / myDiagram.getSize().width;
//        int height = Math.round(hscale * myDiagram.getSize().height)+1;
        @NotNull Dimension dimension = new Dimension(width, height);

        this.setPreferredSize(dimension);
        this.setSize(dimension);
//        this.setMaximumSize(dimension);
        return height;
    }

    public void updateTime() {
        getDiagramEx().updateTime();
        repaint();
    }

    /**
     * @return a Pair of the Scaling in the X as first, and scaling of the Y axis as second.
     */
    @NotNull
    public Pair<Float, Float> getXYScalingFactors() {
//        SVGDiagram diagram = getDiagramEx().getDiagram();
        @NotNull Dimension size = getDiagramEx().getSize();
        float dx = 1.f * size.width / this.getWidth();
        float dy = 1.f * size.height / this.getHeight();
        return new Pair<>(dx, dy);
    }

    public boolean isActionLocked() {
        return myEditSupport.isActionLocked();
    }

    @Nullable
    @Override
    public Object getData(@NonNls String dataId) {
        if (DATA_KEY.is(dataId)) {
            return this;
        }
        return null;
    }

    public SVGPreviewContainer getSVGPreviewContainer() {
        return SVGPreviewContainer;
    }

    public void setSVGPreviewContainer(SVGPreviewContainer SVGPreviewContainer) {
        this.SVGPreviewContainer = SVGPreviewContainer;
    }
}