/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.preview.svgwrapper;

import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.IconLoader;
import com.intellij.openapi.util.Pair;
import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.LinkAnchor;
import plg.sketchlinks.SketchlinkManager;
import plg.ui.overlay.SketchAnchorTransferable;
import plg.ui.preview.SketchEditMouseListener;
import plg.ui.preview.SketchHighlightMouseListener;

import javax.swing.*;
import java.awt.*;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.*;
import java.awt.event.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

/**
 * Created by Peter on 08.09.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class SVGPreviewContainer {
    //    @NonNls
//    public static final String STROKE_Property = "stroke";
//    @NonNls
//    public static final String STROKE_WIDTH_Property = "stroke-width";
//    @NonNls
//    public static final String FILL_Property = "fill";
//    @NonNls
//    public static final String FILL_OPACITY_Property = "fill-opacity";
    @NonNls
    public static final String SVGDIAGRAM_Property = "diagram";
    @NonNls
    public static final String CURRENT_ID_Property = "ID";
    @NonNls
    public static final String ANNOTATION_Property = "annotation";
    @NonNls
    public static final String ANNOTATION_SHORTER_Property = "annotation.shorter";
    //accordion icons
    private static final Icon myAccordeonExpanded = IconLoader.getIcon("/plg/ui/icons/arrows_up_single.png");
    private static final Icon myAccordeonCollapsed = IconLoader.getIcon("/plg/ui/icons/arrows_down_single.png");
    //Drag n drop
    @NotNull
    final private DragSourceListener myDragSourceListener;
    @NotNull
    final private DragSource dragSource;
    @Nullable
    final private TransferHandler myTransferHandler;
    @Nullable
    final private DragGestureListener myDragGestureListener;
    //mouselisteners
    @NotNull
    private final SketchHighlightMouseListener myHighlightingMouse;
    @NotNull
    private final SketchEditMouseListener myEditMouse;
    @NotNull
    private final Project project;
    //swing stuff
    private SVGDiagramEx myDiagramEx;
    private JPanel rootPanel;
    private JPanel contentPanel;
    private SVGPanelExt mySVGPanel;
    private JPanel bottomScale;
    private JTextArea myLongText;
    private JPanel myShortAnnotation;
    private JPanel myLongAnnotation;
    private JLabel myShortText;
    private JButton myLongHideButton;
    private JButton myChangeButton;
    private JButton myShortHideButton;
    private JPanel scrollBarOffsetPanel;
    private JPanel myLongButtonContainer;
    private JPanel myAnnotationPanel;
    //property support
    @NotNull
    private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);
    //    @NonNls
//    public static final String STROKE_OPACITY_Property = "stroke-opacity";
    @Nullable
    private String svgId;
    //disable some features
    private boolean isToolWindowMode = false;
    private boolean isEditMode = true;

    //Constructors
    public SVGPreviewContainer(@NotNull SVGDiagramEx diagramEx, @NotNull Project project) {
        //svgpanel setup
        this.project = project;
        myDiagramEx = diagramEx;
        mySVGPanel.setSVGPreviewContainer(this);
        mySVGPanel.setDiagramEx(diagramEx);
        mySVGPanel.addPropertyChangeListener(SVGPanelExt.HIGHLIGHTED_IDS_Property, new PropertyChangeListener() {
            @Override
            public void propertyChange(@NotNull PropertyChangeEvent evt) {
                Object newValue = evt.getNewValue();
                if (!(newValue instanceof Collection)) {
                    SVGPreviewContainer.this.setCurrentID(null);
                    return;
                }
                @NotNull Collection collection = (Collection) newValue;
                @NotNull Iterator iterator = collection.iterator();
                @Nullable String id = null;
                if (iterator.hasNext()) {
                    Object next = iterator.next();
                    if (next instanceof String) {
                        id = (String) next;
                    }
                }
                //only set the first id, since there is currently no plans to display several annotations
                SVGPreviewContainer.this.setCurrentID(id);
            }
        });

        //dnd setup
        myHighlightingMouse = new SketchHighlightMouseListener(project);
        myEditMouse = new SketchEditMouseListener(project);
        myTransferHandler = createTransferHandler(myHighlightingMouse);
        myDragSourceListener = createDragSourceAdapter();
        myDragGestureListener = createDragGestureListener();
        dragSource = new DragSource();
        setEditMode(false); //register highlighting and dnd listeners

        //annotation bar setup
        myLongAnnotation.setVisible(false);
        myLongText.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                myLongAnnotation.setVisible(false);
                myShortAnnotation.setVisible(true);

                updateSize();

            }
        });


        myShortAnnotation.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                myShortAnnotation.setVisible(false);
                myLongAnnotation.setVisible(true);
                updateSize();

            }
        });
        @NotNull ActionListener actionListener = new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                setSketchPreviewVisible(!mySVGPanel.isVisible());
            }
        };
        myLongHideButton.addActionListener(actionListener);
        myShortHideButton.addActionListener(actionListener);
        myShortHideButton.setLayout(new BorderLayout(0, 0));


    }

    public void setSketchPreviewVisible(boolean visible) {
        if (mySVGPanel.isVisible() == visible)
            return;
        mySVGPanel.setVisible(visible);
//        String buttonText = "▼"; //triangle down
//        if (visible) {
//            buttonText = "▲"; //triangle up
//        }
//        myShortHideButton.setText(buttonText);
//        myLongHideButton.setText(buttonText);
        Icon icon = visible ? myAccordeonExpanded : myAccordeonCollapsed;
        myShortHideButton.setIcon(icon);
        myLongHideButton.setIcon(icon);
        updateSize();
    }

    public boolean isEditMode() {
        return isEditMode;
    }

    public void setEditMode(boolean active) {
        isEditMode = active;
        for (MouseListener mouseListener : mySVGPanel.getMouseListeners()) {
            mySVGPanel.removeMouseListener(mouseListener);
        }
        for (MouseMotionListener listener : mySVGPanel.getMouseMotionListeners()) {
            mySVGPanel.removeMouseMotionListener(listener);
        }
        for (DragSourceListener listener : dragSource.getDragSourceListeners()) {
            dragSource.removeDragSourceListener(listener);
        }
        for (DragSourceMotionListener listener : dragSource.getDragSourceMotionListeners()) {
            dragSource.removeDragSourceMotionListener(listener);
        }

        if (active) {
            mySVGPanel.removeMouseListener(myHighlightingMouse);
            mySVGPanel.removeMouseMotionListener(myHighlightingMouse);
            mySVGPanel.setTransferHandler(null);
            dragSource.removeDragSourceListener(myDragSourceListener);
            mySVGPanel.setDropTarget(null);

            mySVGPanel.addMouseListener(myEditMouse);
            mySVGPanel.addMouseMotionListener(myEditMouse);
        } else {
            mySVGPanel.removeMouseListener(myEditMouse);
            mySVGPanel.removeMouseMotionListener(myEditMouse);
            mySVGPanel.addMouseMotionListener(myHighlightingMouse);
            mySVGPanel.addMouseListener(myHighlightingMouse);
            mySVGPanel.setTransferHandler(myTransferHandler);
            dragSource.createDefaultDragGestureRecognizer(mySVGPanel, DnDConstants.ACTION_LINK, myDragGestureListener);
        }
    }

    public boolean isToolWindowMode() {
        return isToolWindowMode;
    }

    public void setToolWindowMode(boolean isToolWindowMode) {
        if (isToolWindowMode) {
            scrollBarOffsetPanel.setVisible(true);
            myLongHideButton.setVisible(true);
            myShortHideButton.setVisible(true);
        } else {
            scrollBarOffsetPanel.setVisible(false);
            myShortHideButton.setVisible(false);
            myLongHideButton.setVisible(false);
            mySVGPanel.setVisible(true);
            setWidth(rootPanel.getWidth(), 0);
        }
        this.isToolWindowMode = isToolWindowMode;
    }

    @NotNull
    public SVGPanelExt getSVGPanel() {
        return mySVGPanel;
    }

    @NotNull
    public JPanel getRootPanel() {
        return rootPanel;
    }

    @NotNull
    public SVGDiagramEx getDiagram() {
        return myDiagramEx;
    }

    public void setDiagram(@NotNull SVGDiagramEx diagram) {
        @NotNull SVGDiagramEx oldVal = getDiagram();
        if (diagram.equals(oldVal)) {
            return;
        }
        myDiagramEx = diagram;
        mySVGPanel.setDiagramEx(diagram);
        mySVGPanel.repaint();
        propertyChangeSupport.firePropertyChange(SVGDIAGRAM_Property, oldVal, diagram);
    }

    @Nullable
    @NonNls
    public String getCurrentID() {
        return this.svgId;
    }

    private void setCurrentID(@Nullable String svgId) {
        if ((svgId != null && svgId.equals(this.svgId)) || (this.svgId != null && this.svgId.equals(svgId))) {
            return;
        }
        @Nullable String oldVal = this.svgId;
        this.svgId = svgId;
        propertyChangeSupport.firePropertyChange(CURRENT_ID_Property, oldVal, svgId);

    }

    @NotNull
    public String getAnnotation() {
        String text = myLongText.getText();
        if (text == null) {
            return "";
        }
        return text;
    }

    public void setAnnotation(@NotNull String text) {
        setAnnotation(text, text.replaceAll("\\n", " "));
    }

    public void setAnnotation(@NotNull final String text, @NotNull final String shortText) {

//        if (text.equals(oldVal)) {
//            return;
//        }
        UIUtil.invokeLaterIfNeeded(new Runnable() {
            @Override
            public void run() {
                final String oldVal = myLongText.getText();
                final String oldValShort = myShortText.getText();
                myLongText.setText(text);
                myShortText.setText(shortText);
                propertyChangeSupport.firePropertyChange(ANNOTATION_Property, oldVal, text);
                propertyChangeSupport.firePropertyChange(ANNOTATION_SHORTER_Property, oldValShort, shortText);

            }
        });
    }


    public void updateSize() {
        myAnnotationPanel.revalidate();
        setWidth(rootPanel.getWidth(), scrollBarOffsetPanel.isVisible() ? scrollBarOffsetPanel.getWidth() : 0);
        Container parent = rootPanel.getParent();
        if (parent != null)
            parent.revalidate();
    }

    public void setWidth(int newWidth, int vScrollWidth) {
        rootPanel.revalidate();
        rootPanel.repaint();
        Insets insets = contentPanel.getInsets();
        int reduce = insets.left + insets.right;
        Insets rootInsets = rootPanel.getInsets();
        reduce += rootInsets.left + rootInsets.right;
        reduce += vScrollWidth;
        int svgWidth = newWidth - reduce;
        int height = mySVGPanel.setHorizontalStretch(svgWidth);

        height = mySVGPanel.isVisible() ? height : 0;
        if (scrollBarOffsetPanel.isVisible() != (vScrollWidth > 0)) {
            if (vScrollWidth > 0 && scrollBarOffsetPanel.getWidth() != vScrollWidth) {
                @NotNull Dimension dim = new Dimension(vScrollWidth, 0);
                scrollBarOffsetPanel.setPreferredSize(dim);
                scrollBarOffsetPanel.setSize(dim);
            }
            scrollBarOffsetPanel.setVisible(vScrollWidth > 0);
        }
        height += getHeight(myShortAnnotation);
        height += getHeight(myLongAnnotation);
        height += myAnnotationPanel.getInsets().bottom + myAnnotationPanel.getInsets().top;
        contentPanel.setSize(new Dimension(svgWidth, height));
        contentPanel.setPreferredSize(new Dimension(svgWidth, height));

        contentPanel.setMaximumSize(new Dimension(svgWidth, height));
        contentPanel.revalidate();
        height += contentPanel.getInsets().bottom + contentPanel.getInsets().top;
        @NotNull Dimension dimension = new Dimension(newWidth, height);


        contentPanel.setPreferredSize(dimension);
        @NotNull Dimension rootDim = new Dimension(dimension);
        rootDim.height += rootPanel.getInsets().top + rootPanel.getInsets().bottom;
        rootPanel.setPreferredSize(rootDim);
        rootPanel.setMinimumSize(rootDim);
        rootPanel.setMaximumSize(rootDim);
        rootPanel.setSize(rootDim);
        rootPanel.revalidate();
        rootPanel.repaint();
//        System.out.println("New RootPanel: "+rootPanel.getSize()+" "+contentPanel.getSize()+" "+mySVGPanel.getSize()+" Width: "+svgWidth+" Height: "+height+" Scroll: "+vScrollWidth);
    }

    private int getHeight(@NotNull JComponent comp) {
        if (!comp.isVisible()) {
            return 0;
        }
        int result = comp.getHeight();
        if (result == 0) {
            result = comp.getPreferredSize().height;
        }
        result += comp.getInsets().top + comp.getInsets().bottom;
        return result;
    }

    @Nullable
    private DragGestureListener createDragGestureListener() {
        return new DragGestureListener() {
            @Override
            public void dragGestureRecognized(@NotNull DragGestureEvent dge) {
                //      dragSource.startDrag(dge,DragSource.DefaultLinkDrop,new StringSelection("dragTEST"),this);
                System.out.println("dragGestureRecognize: " + svgId);
                if (svgId == null) {
                    return;
                }
                @Nullable LinkAnchor knownAnchor = SketchlinkManager.getInstance(project).getKnownAnchor(svgId);
                if (knownAnchor == null) {
                    return;
                }
                dge.getDragSource().startDrag(dge, DragSource.DefaultLinkDrop, new SketchAnchorTransferable(knownAnchor), myDragSourceListener);
            }
        };
    }

    @NotNull
    private DragSourceAdapter createDragSourceAdapter() {
        return new DragSourceAdapter() {
            @Override
            public void dragEnter(DragSourceDragEvent dsde) {
                System.out.println("DragEnter to Icon? " + svgId);
            }

            @Override
            public void dragDropEnd(DragSourceDropEvent dsde) {
                System.out.println("dragDropEnd " + svgId);

            }
        };
    }

    @Nullable
    private TransferHandler createTransferHandler(@NotNull final SketchHighlightMouseListener listener) {
        return new TransferHandler() {
            @Nullable
            private LinkAnchor getDropData(@NotNull TransferSupport support) {
                if (!support.isDataFlavorSupported(SketchAnchorTransferable.SketchFlavor)) {
//                    System.out.println("Sketchflavor on importData not supported");
                    return null;
                }
                @Nullable LinkAnchor data = null;
                try {
                    data = (LinkAnchor) support.getTransferable().getTransferData(SketchAnchorTransferable.SketchFlavor);
                } catch (@NotNull UnsupportedFlavorException | IOException e) {
                    e.printStackTrace();
                }
                return data;
            }

            @Nullable
            private LinkAnchor getTarget(@NotNull TransferSupport support) {
                Point dropPoint = support.getDropLocation().getDropPoint();
                @NotNull MouseEvent mouseEvent = new MouseEvent(mySVGPanel, 0, System.currentTimeMillis(), 0, dropPoint.x, dropPoint.y, 0, false);
                listener.mouseMoved(mouseEvent);
                return svgId == null ? null : SketchlinkManager.getInstance(project).getKnownAnchor(svgId);
            }

            @Override
            public boolean canImport(@NotNull TransferSupport support) {
                @Nullable LinkAnchor dropData = getDropData(support);
                @Nullable LinkAnchor target = getTarget(support);
                if (dropData == null || target == null || dropData.getAnchorId().equals(target.getAnchorId())) {
                    return false;
                }
                @NotNull List<Pair<LinkAnchor, Boolean>> linksFor = SketchlinkManager.getInstance(project).getLinksFor(target);
                for (@NotNull Pair<LinkAnchor, Boolean> booleanPair : linksFor) {
                    if (booleanPair.first.getAnchorId().equals(dropData.getAnchorId())) {
                        if (!booleanPair.second) {
                            return true;
                        } else {
                            return false;
                        }
                    }
                }
                return true;
            }

            @Override
            public boolean importData(@NotNull TransferSupport support) {
                @Nullable LinkAnchor dropData = getDropData(support);
                @Nullable LinkAnchor target = getTarget(support);
                if (dropData == null || target == null || dropData.getAnchorId().equals(target.getAnchorId())) {
                    return false;
                }
                @NotNull List<Pair<LinkAnchor, Boolean>> linksFor = SketchlinkManager.getInstance(project).getLinksFor(target);
                for (@NotNull Pair<LinkAnchor, Boolean> booleanPair : linksFor) {
                    if (booleanPair.first.getAnchorId().equals(dropData.getAnchorId())) {
                        if (!booleanPair.second) {
                            break;
                        }
                    }
                }
                SketchlinkManager.getInstance(project).addLink(dropData, target, false, true);
                return true;
            }
        };
    }

    public void addPropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(@NotNull PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener(listener);
    }
}
