/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.ui.sketchupload;


import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.project.Project;
import org.apache.http.conn.HttpHostConnectException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.tools.SketchlinkServerRESTUtil;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.util.concurrent.*;

public class SketchUpload extends JDialog implements RunnableFuture<String> {
    @NotNull
    private final Project project;
    @NotNull
    private final String serverURL;
    @Nullable
    private final String linkWithID;
    private JPanel contentPane;
    private JButton buttonOK;
    private JButton buttonCancel;
    private JPanel sketchDataContent;
    private SingleSketchUpload singleSketchUpload;
    @Nullable
    private String result = null;
    private boolean cancelled = false;


    public SketchUpload(@NotNull Project project, @NotNull String serverUrl, @Nullable String linkWithID) {
        this.serverURL = serverUrl;
        this.linkWithID = linkWithID;

        this.project = project;
        setContentPane(contentPane);
        setModal(true);
        getRootPane().setDefaultButton(buttonOK);

        buttonOK.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onOK();
            }
        });

        buttonCancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onCancel();
            }
        });

// call onCancel() when cross is clicked
        setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                onCancel();
            }
        });

// call onCancel() on ESCAPE
        contentPane.registerKeyboardAction(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onCancel();
            }
        }, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);

        //add sketch entry
        singleSketchUpload = new SingleSketchUpload();
        sketchDataContent.add(singleSketchUpload.getRootPanel(), BorderLayout.CENTER);
        pack();
        Insets insets = getInsets();
        @NotNull Dimension dim = new Dimension(getRootPane().getMinimumSize());
        dim.width += insets.left + insets.right;
        dim.height += insets.bottom + insets.top;
        setMinimumSize(dim);
        buttonOK.setEnabled(false);
        singleSketchUpload.addPropertyChangeListener(new PropertyChangeListener() {
            @Override
            public void propertyChange(@NotNull PropertyChangeEvent evt) {
                if ("valid".equals(evt.getPropertyName())) {
                    buttonOK.setEnabled((Boolean) evt.getNewValue());
                }
            }
        });
        this.setTitle(linkWithID != null ? "Upload New Sketch" : "Upload and Link New Sketch");
    }

    private void onOK() {
        @NotNull final String serverUrl = serverURL;
        final File imageFile = singleSketchUpload.getFile();
        final String title = singleSketchUpload.getTitle();
        final String authors = singleSketchUpload.getAuthors();
        final String annotation = singleSketchUpload.getAnnotation();
        @NotNull final String artifact_id = linkWithID != null ? linkWithID : "";
        if (imageFile == null || title == null || authors == null || title.trim().isEmpty() || authors.trim().isEmpty()) {
            return;
        }
        Future<String> stringFuture = ApplicationManager.getApplication().executeOnPooledThread(new Callable<String>() {

            @Override
            public String call() throws Exception {
                try {
                    //upload new image
                    return SketchlinkServerRESTUtil.uploadSketchAndGetResponseBody(serverUrl, imageFile, title, authors, annotation, artifact_id);
                } catch (HttpHostConnectException e1) {
                    e1.printStackTrace();
                    System.out.println("IP not available " + e1.getMessage());
                } catch (IOException e) {
                    e.printStackTrace();
                    System.out.println("File not found " + e.getMessage());
                }
                return null;
            }
        });

        try {
            result = stringFuture.get();
        } catch (InterruptedException | ExecutionException e) {
            e.printStackTrace();
        }
        if (result == null) {
            result = "";
            return;
        }
        dispose();
    }

    private void onCancel() {
// add your code here if necessary
        dispose();
    }

    public void run() {
        setVisible(true);
    }


    @Override
    public boolean cancel(boolean mayInterruptIfRunning) {
        if (cancelled)
            return true;
        if (mayInterruptIfRunning) {
            dispose();
            cancelled = true;
            return true;
        }
        return false;
    }


    @Override
    public boolean isCancelled() {
        return cancelled;
    }


    @Override
    public boolean isDone() {
        return result != null || cancelled;
    }

    /**
     * Gets the new ID for the uploaded sketch, or an empty string if the process was cancelled.
     */
    @Nullable
    @Override
    public String get() throws InterruptedException, ExecutionException {
        return result;
    }


    @Nullable
    @Override
    public String get(long timeout, TimeUnit unit) throws InterruptedException, ExecutionException, TimeoutException {
        return result;
    }

//    public static void main(String[] args) {
//        SketchUpload dialog = new SketchUpload();
//        dialog.pack();
//        dialog.setSize(500, 400);
//        Insets insets = dialog.getInsets();
//        Dimension dim = new Dimension(dialog.getRootPane().getMinimumSize());
//        dim.width += insets.left + insets.right;
//        dim.height += insets.bottom + insets.top;
//        dialog.setMinimumSize(dim);
//        dialog.setVisible(true);
//        System.exit(0);
//    }
}
