/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.websocket;

import com.intellij.openapi.project.Project;
import io.netty.bootstrap.Bootstrap;
import io.netty.channel.Channel;
import io.netty.channel.group.ChannelGroup;
import io.netty.channel.group.DefaultChannelGroup;
import io.netty.handler.codec.http.websocketx.CloseWebSocketFrame;
import io.netty.handler.codec.http.websocketx.TextWebSocketFrame;
import io.netty.util.concurrent.GlobalEventExecutor;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.SLConfigForm;
import plg.slApplicationState;
import plg.slProjectComp;

import java.net.URI;
import java.net.URISyntaxException;
import java.security.InvalidParameterException;
import java.util.*;

/**
 * Created by Peter on 22.02.14.
 */
public class ChannelManagerImpl extends ChannelManager {
    @NotNull
    private final slWebsocketManager myManager;
    private final Set<SketchlinkChannel> myChannels = new HashSet<>();
    private final Map<String, ChannelGroup> myGroups = new HashMap<>();

    public ChannelManagerImpl(@NotNull slWebsocketManager manager) {
        this.myManager = manager;
    }

    @NotNull
    public slWebsocketManager getManager() {
        return myManager;
    }


    @Override
    public void broadcastText(@NotNull String textData, @Nullable String... groups) {
        if (myChannels.isEmpty()) {
            //JOptionPane.showMessageDialog(null, "Tried sending Data while not connected to any channels!");
            return;
        }
        Set<Channel> uniqueChannels;
        if (groups != null && groups.length > 0) {
            uniqueChannels = new HashSet<>(myChannels.size());
            //A SketchlinkChannel might be Part of several ChannelGroups

            for (@NotNull ChannelGroup group : myGroups.values()) {
                for (@Nullable String name : groups) {
                    if (name == null || !name.equals(group.name())) {
                        continue;
                    }
                    Collections.addAll(uniqueChannels, group.toArray(new Channel[group.size()]));
                }
            }
            @NotNull Iterator<Channel> iterator = uniqueChannels.iterator();
            @NotNull TextWebSocketFrame frame = new TextWebSocketFrame(textData);
            while (iterator.hasNext()) {
                iterator.next().writeAndFlush(frame);
            }
            return;
        }
        //no single groups set, broadcast to all

        for (@NotNull SketchlinkChannel channel : myChannels) {

            channel.send(textData);
        }

    }

    @NotNull
    @Override
    public ChannelGroup getChannelGroup(@NotNull String name) {
        ChannelGroup group = myGroups.get(name);
        if (group == null) {
            group = new DefaultChannelGroup(name, GlobalEventExecutor.INSTANCE);
            myGroups.put(name, group);
        }
        return group;
    }

    @Override
    public void removeChannelGroup(@NotNull String name) {
        if (myGroups.get(name) != null) {
            myGroups.get(name).clear();
        }
        myGroups.remove(name);
    }

    @NotNull
    @Override
    public ChannelGroup getDefaultGroup() {
        return getChannelGroup(DEFAULT_GROUP);
    }

    @NotNull
    @Override
    public Collection<ChannelGroup> getChannelGroups() {
        return myGroups.values();
    }

    @Override
    public void addChannelToGroup(@NotNull SketchlinkChannel channel, @NotNull String groupname) {
        getChannelGroup(groupname).add(channel.getChannel());
        if (!myChannels.contains(channel)) {
            myChannels.add(channel);
        }
    }

    @Override
    public void addChannelToDefaultGroup(@NotNull SketchlinkChannel channel) {
        if (getDefaultGroup().contains(channel.getChannel())) {
            return;
        }
        getDefaultGroup().add(channel.getChannel());
    }


    @Override
    public void connectToDefault() throws URISyntaxException {
        String serverLocation = slProjectComp.getInstance(myManager.getProject()).getServerURL();
        URI uri;
//        if (serverLocation.trim().isEmpty()) {
//            throw new URISyntaxException("", "Serverlocation is not set.");
//        }
        //make sure the url stars with "ws://" and not http:// or similiar
        serverLocation = serverLocation.replaceFirst("[\\w\\d]*://", "ws://"); //NON-NLS NON-NLS
        if (!serverLocation.startsWith("ws://")) //NON-NLS
            serverLocation = "ws://" + serverLocation; //NON-NLS
        if (!serverLocation.endsWith("/")) {
            serverLocation = serverLocation + "/";
        }


        uri = new URI(serverLocation + "websocket"); //NON-NLS
        if (uri.getPort() <= 0) {
            uri = new URI(uri.getScheme(), uri.getUserInfo(), uri.getHost(), ChannelManager.DEFAULT_PORT, uri.getPath(), uri.getQuery(), uri.getFragment());
        }

        for (@NotNull SketchlinkChannel channel : getChannelByDescription("Default")) { //NON-NLS
            close(channel, 1000, "Reconnecting to (possible) new Server");
        }
        createChannel(uri, "Default");
    }

    @Override
    public void removeChannelFromGroup(@NotNull SketchlinkChannel channel, @NotNull String groupname) {
        getChannelGroup(groupname).remove(channel.getChannel());
    }

    private void removeChannel(@NotNull SketchlinkChannel sketchlinkChannel) {
        myChannels.remove(sketchlinkChannel);
        for (@NotNull ChannelGroup channelGroup : myGroups.values()) {
            @Nullable Channel channel = sketchlinkChannel.getChannel();
            if (channel == null) {
                continue;
            }
            channelGroup.remove(channel);
        }
    }


    @Override
    public void broadcastDisconnectRequest(int statuscode, @Nullable String reason) {
        for (@NotNull SketchlinkChannel channel : getAllChannels()) {
            channel.requestClosing(statuscode, reason);
        }
    }

    @Override
    public void closeAll(int statuscode, @Nullable String reason) {
        for (@NotNull SketchlinkChannel channel : myChannels) {
            channel.forceClosing(statuscode, reason);
        }
    }

    @Override
    public void close(@NotNull Channel channel, int statuscode, @NonNls @Nullable String reason) {
        @Nullable SketchlinkChannel sketchlinkChannel = getSketchlinkChannel(channel);
        if (sketchlinkChannel != null) {
            close(sketchlinkChannel, statuscode, reason);
        } else {
            //System.out.println("Closed unregistered Channel");
            channel.writeAndFlush(new CloseWebSocketFrame(statuscode, reason));
            channel.close();
        }
    }


    @Override
    public void close(@NotNull SketchlinkChannel channel, int statuscode, @Nullable String reason) {
        removeChannel(channel);
        channel.requestClosing(statuscode, reason);
    }

    @Override
    public void close(@NotNull Channel channel) {
        close(channel, 1000, "No reason given");
    }

    @NotNull
    @Override
    public SketchlinkChannel createChannel(@NotNull URI uri, @NotNull String description) {
        final Project project = this.myManager.getProject();
        if (project.isDisposed()) {
            throw new InvalidParameterException("Tried to create a channel for a disposed Project");

        }
        @NotNull final SketchlinkChannel result = new SketchlinkChannelImpl(uri, description, project);
        @NotNull Thread clientThread = new Thread(new Runnable() {
            @Override
            public void run() {
                try {
                    result.open();
                } catch (Exception e) {
                    SLConfigForm.showInvalidServerMessage(e.getMessage() != null ? e.getMessage() : slApplicationState.RESOURCE_BUNDLE.getString("channelmanager.invalidserver.no.further.description"), getManager().getProject());
                }
            }
        }, "Sketchlink_WebsocketClient " + uri.toString());
        this.myChannels.add(result);
        clientThread.setDaemon(true);
        clientThread.start();
//        if (myChannels.size() == 1) {
//            ApplicationManager.getApplication().executeOnPooledThread(new Runnable() {
//                @Override
//                public void run() {
//                    SketchlinkManager.getInstance().lookupAnchors(false, project);
//                }
//            });
//        }
        return result;
    }

    @NotNull
    @Override
    public List<SketchlinkChannel> getChannelByDescription(@NotNull String description) {
        @NotNull List<SketchlinkChannel> result_list = new ArrayList<>(myChannels);
        for (@NotNull SketchlinkChannel channel : getAllChannels()) {
            if (channel.getName().equals(description)) {
                result_list.add(channel);
            }
        }
        return result_list;
    }

    @NotNull
    @Override
    public List<SketchlinkChannel> getSketchlinkChannelByTarget(@NotNull URI uri) {
        @NotNull List<SketchlinkChannel> result_list = new ArrayList<>(myChannels);
        for (@NotNull SketchlinkChannel channel : getAllChannels()) {
            if (channel.getEndpointUri().equals(uri)) {
                result_list.add(channel);
            }
        }
        return result_list;
    }


    @NotNull
    @Override
    public List<SketchlinkChannel> getAllChannels() {
        @NotNull Iterator<SketchlinkChannel> iterator = myChannels.iterator();
        @NotNull List<SketchlinkChannel> result = new ArrayList<>(myChannels.size());
        while (iterator.hasNext()) {
            SketchlinkChannel channel = iterator.next();
//            if (channel == null) {
//                iterator.remove(); //cleanup null items if found
//                continue;
//            }
            result.add(channel);
        }
        return result;
    }


    @Nullable
    @Override
    public SketchlinkChannel getSketchlinkChannel(@NotNull Channel channel) {
        for (@NotNull SketchlinkChannel slchan : getAllChannels()) {
            if (channel.equals(slchan.getChannel())) {
                return slchan;
            }
        }
        return null;
    }

    @NotNull
    @Override
    protected Bootstrap getBootstrab() {
        return new Bootstrap();
    }


}
