/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.websocket;


import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.DumbService;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.MessageType;
import io.netty.channel.*;
import io.netty.handler.codec.http.FullHttpResponse;
import io.netty.handler.codec.http.HttpResponseStatus;
import io.netty.handler.codec.http.websocketx.*;
import io.netty.util.CharsetUtil;
import org.jetbrains.annotations.NotNull;
import plg.SLConfigForm;
import plg.slApplicationState;
import plg.slCurrentProjectSettings;
import plg.tools.slGuiTools;
import plg.tools.slSearch;
import plg.websocket.commands.builder.RegisterRequestBuilder;

import java.text.MessageFormat;

/**
 * Created by Peter on 19.02.14.
 */
public class ClientHandler extends SimpleChannelInboundHandler<Object> {
    private final Logger LOG = Logger.getInstance(ClientHandler.class);
    private final WebSocketClientHandshaker handshaker;
    private final Project project;
    //private final slWebsocketManager websocketManager;

    private ChannelPromise handshakeFuture;


    public ClientHandler(WebSocketClientHandshaker handshaker, Project project) {
        this.handshaker = handshaker;
        this.project = project;
        // this.websocketManager = ServiceManager.getService(project, slWebsocketManager.class);

    }

    public ChannelFuture handshakeFuture() {
        return handshakeFuture;
    }

    @Override
    public void handlerAdded(@NotNull ChannelHandlerContext ctx) throws Exception {
        handshakeFuture = ctx.newPromise();
    }

    /**
     * @see io.netty.channel.SimpleChannelInboundHandler#channelActive(io.netty.channel.ChannelHandlerContext)
     */
    @Override
    public void channelActive(@NotNull ChannelHandlerContext ctx) throws Exception {
        handshaker.handshake(ctx.channel());
    }

    @Override
    public void channelInactive(ChannelHandlerContext ctx) throws Exception {
        LOG.info("WebSocket Client disconnected!");
    }


    /**
     * Netty 5 method
     */
    protected void messageReceived(@NotNull ChannelHandlerContext ctx, @NotNull Object msg) throws Exception {
        Channel inboundChannel = ctx.channel();
        if (!handshaker.isHandshakeComplete()) {

            if (((FullHttpResponse) msg).getStatus().equals(HttpResponseStatus.NOT_FOUND)) {

                SLConfigForm.showInvalidServerMessage(ctx.channel().remoteAddress().toString() + " Type: " + ((FullHttpResponse) msg).getStatus(), project);
                LOG.info("404: Server not found");
                ctx.close();
                return;
            }
            handshaker.finishHandshake(inboundChannel, (FullHttpResponse) msg);
            LOG.info("WebSocket Client connected!");
            handshakeFuture.setSuccess();
            new RegisterRequestBuilder(slCurrentProjectSettings.getInstance(project).getClientName()).send(project, ctx.channel());
            //search the project, to remove chance to accidentally identify a received local id as remote
            DumbService.getInstance(project).smartInvokeLater(new Runnable() {
                @Override
                public void run() {
                    new slSearch(project).runSearchForIDs(null, true, false);
                }
            });

            return;
        }
        if (msg instanceof FullHttpResponse) {
            @NotNull FullHttpResponse response = (FullHttpResponse) msg;
            throw new Exception("Unexpected FullHttpResponse (getStatus=" + response.getStatus() + ", content="
                    + response.content().toString(CharsetUtil.UTF_8) + ')');
        }
        @NotNull WebSocketFrame frame = (WebSocketFrame) msg;

        if (frame instanceof TextWebSocketFrame) {
            final String text = ((TextWebSocketFrame) frame).text();
            //System.out.println("ClientHandler received Text: " + text);
            if (text == null)
                return;
            //System.out.println("Received Data in ClientHandler: " + text);
            try {
                slWebsocketManager.getInstance(project).getCommandsManager().executeCommand(text, inboundChannel);
            } catch (IllegalArgumentException e) {
                LOG.error(e);
                slWebsocketManager.getInstance(project).getChannelManager().close(inboundChannel, 1002, e.getMessage());
            }


//                ApplicationManager.getApplication().executeOnPooledThread(new Runnable() {
//                    public void run() {
//                        ApplicationManager.getApplication().runReadAction(new Runnable() {
//                            public void run() {
//                                if (text.equals("0")) {
//                                    //if id== 0, then query for all svglinks
//                                    queryProjectFor(PRE_REGEX + "([0-9])[0-9]*" + POST_REGEX, true);
//                                } else {
//                                    queryProjectFor(PRE_REGEX + text + POST_REGEX, true);
//                                }
//                            }
//                        });
//                    }
//                });
                /*ProgressManager.getInstance().run(new Task.Backgroundable(project, "Server request"){
                    @Override
                    public void run(@NotNull ProgressIndicator progressIndicator) {

                        // start your process
                        queryProjectFor(PRE_REGEX + textFrame.text() + POST_REGEX, true);

                        // Set the progress bar percentage and text
                        progressIndicator.setFraction(0.10);
                        progressIndicator.setText("90% to finish");


                        // 50% done
                        progressIndicator.setFraction(0.50);
                        progressIndicator.setText("50% to finish");


                        // Finished
                        progressIndicator.setFraction(1.0);
                        progressIndicator.setText("finished");

                    }});*/
        } else if (frame instanceof PongWebSocketFrame) {
            LOG.info("WebSocket Client received pong");
        } else if (frame instanceof CloseWebSocketFrame) {
            String text = ((CloseWebSocketFrame) frame).reasonText();
//            slProjectComp.getInstance(project).setConnectedToServer(false);
            LOG.info("WebSocket Client received closing. Reason: " + text);
            inboundChannel.close();
        }
    }

    @Override
    public void exceptionCaught(@NotNull ChannelHandlerContext ctx, @NotNull Throwable cause) throws Exception {
        slGuiTools.ShowNotificationAtBottom(
                MessageFormat.format(slApplicationState.RESOURCE_BUNDLE.getString("unexpected.loss.of.connection.to.the.sketchlink.synchronisation.server.n.0"), cause.getMessage()), project, MessageType.ERROR);
        cause.printStackTrace();
        if (!handshakeFuture.isDone()) {
            handshakeFuture.setFailure(cause);
        }
        ctx.close();
    }


    /**
     * Netty 4
     */
    protected void channelRead0(@NotNull ChannelHandlerContext ctx, @NotNull Object msg) throws Exception {
        messageReceived(ctx, msg);
    }
}
