/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.websocket;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.extensions.Extensions;
import io.netty.channel.Channel;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.websocket.extensionPoints.WebsocketCommandBean;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Created by Peter on 18.02.14.
 */
public class CommandsManagerImpl extends CommandsManager {
    private final Logger LOG = Logger.getInstance(CommandsManagerImpl.class);

    private final Map<String, RunCmd> myCommands = new HashMap<>();

    public CommandsManagerImpl(@NotNull slWebsocketManagerImpl slWebsocketManager) {
        super(slWebsocketManager);

//        registerCmd(new GetArtifactMetaDataCMD());
//        registerCmd(new getAvailableClientsCMD());
//        registerCmd(new GetContainedAnchorsCMD());
//        registerCmd(new GetLinkedArtifactsCMD());
//        registerCmd(new ShowCMD());
//        registerCmd(new UpdateArtifactMetaDataCMD());
//        registerCmd(new CreateArtifactMetaDataCMD());
//        registerCmd(new CreateLinkCMD());
//        registerCmd(new GetSketchesCMD());
        //initialize commands via extensionpoint declared in the plugin.xml file
        @NotNull WebsocketCommandBean[] beans = Extensions.getExtensions(WebsocketCommandBean.EP_NAME);
        for (@NotNull WebsocketCommandBean extensionBean : beans) {
            @Nullable RunCmd cmd = extensionBean.getCommandHandler();
            registerCmd(cmd, extensionBean.getCommand());
//            System.out.println("Registering: "+cmd.getCommandIdentifier() +" "+extensionBean.getCommand());
        }

    }


    /**
     * Registers a new Command in this Manager.
     *
     * @param runCmd the Command to register
     * @return true if registration was successful, false if this command and it's version is already known.
     */
//    @Override
    private boolean registerCmd(@NotNull RunCmd runCmd, @NotNull String commandIdentifier) {
        @NotNull String key = commandIdentifier;
        if (key.trim().isEmpty()) {
            key = runCmd.getCommandIdentifier();
            if (key.trim().isEmpty()) {
                return false;
            }
        }
        if (myCommands.containsKey(key))
            return false;
        runCmd.init(this);
        myCommands.put(key, runCmd);
        return true;

    }

    @Nullable
    @Override
    public RunCmd getCommand(@NotNull String commandIdentifier) {
        return myCommands.get(commandIdentifier);
    }

    @NotNull
    @Override
    public List<String> getAllKnownCommandNames() {
        return new ArrayList<>(myCommands.keySet());
    }

    @NotNull
    @Override
    public List<RunCmd> getAllKnownCommands() {
        @NotNull List<RunCmd> result = new ArrayList<>(myCommands.size());
        result.addAll(myCommands.values());
        return result;
    }

    @Override
    public void executeCommand(@NotNull CommandTopParameterReceive toplevel, @NotNull Channel inboundChannel, @NotNull String receivedCommand) {
        RunCmd cmd = myCommands.get(toplevel.message);
        if (cmd == null) {

            LOG.info("Received unknown command: " + toplevel.message + " and payload ");
            return;
        }

        cmd.executeCmd(toplevel, inboundChannel, receivedCommand);
    }

    @Override
    public void executeCommand(@NotNull String receivedCommand, @NotNull Channel inboundChannel) {

        @Nullable CommandTopParameterReceive command = null;
        //Please note the {} after instantiation, they're important
        Gson json = new GsonBuilder().serializeNulls().create();

        try {

            command = json.fromJson(receivedCommand, CommandTopParameterReceive.class);
            if (command == null) {
                throw new IllegalArgumentException("The Textframe could be parsed, but no interpretable data was found inside.");
            }
        } catch (Exception e) {
            throw new IllegalArgumentException("execute Command received an invalid Textframe (Sketchlink API error),\nSource error: " + e.getMessage() + "\nReceived Data: " + receivedCommand);
        }
        executeCommand(command, inboundChannel, receivedCommand);
    }

    @Override
    @Nullable
    public RunCmd unregisterCmd(@NotNull String commandIdentifier) {
        return myCommands.remove(commandIdentifier);
    }

    @NotNull
    @Override
    public RunCmd getCommandOrInstantiate(@NotNull Class<? extends RunCmd> cmdClass) throws IllegalAccessException, InstantiationException {
        for (@NotNull RunCmd runCmd : getAllKnownCommands()) {
            if (runCmd.getClass() == cmdClass) {
                return (RunCmd) runCmd;
            }
        }
        RunCmd stub;
        stub = cmdClass.newInstance();
        if (myCommands.containsKey(stub.getCommandIdentifier())) {
            return myCommands.get(stub.getCommandIdentifier());
        }
        registerCmd(stub, stub.getCommandIdentifier());
        return stub;
    }


}
