/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.websocket;


import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import io.netty.channel.Channel;
import io.netty.handler.codec.http.websocketx.TextWebSocketFrame;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.LinkAnchor;
import plg.sketchlinks.LocalLinkAnchor;
import plg.sketchlinks.SketchlinkManager;
import plg.tools.slGuiTools;
import plg.websocket.commands.builder.ChangeArtifactMetaDataRequestBuilder;
import plg.websocket.commands.parameter.ErrorParameter;

import javax.swing.*;
import java.lang.reflect.Type;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;


/**
 * Created by Peter on 19.02.14.
 * <p/>
 * T Class of request parameter
 * V Class of response parameter
 */
public abstract class RunCmdStub<T extends Payload, V extends Payload> implements RunCmd {
    private final Logger LOG = Logger.getInstance(RunCmdStub.class);
    private final Gson myGson = new GsonBuilder().serializeNulls().create();
    private final Map<UUID, Payload> myRequestHistory = new ConcurrentHashMap<>();
    private CommandsManager manager;

    @Override
    public void init(@NotNull CommandsManager manager) {
        this.manager = manager;
    }

    @NotNull
    public CommandsManager getManager() {
        return manager;
    }


    //    private class requestWrap {
//        T request;
//        long time;
//        requestWrap(T param) {
//            this.request = param;
//            time = System.currentTimeMillis();
//        }
//    }

    /**
     * Incoming Commandpackages get compared to the return value of this Identifier. If equal, the
     * response-Method will be run with it's data.
     *
     * @return the CommandIdentifier of this Method
     */
    @NotNull
    @NonNls
    @Override
    public abstract String getCommandIdentifier();

    /**
     * Override this Method to implement Handling of Request-Pakets.
     *
     * @param params         the parameters for execution, supplied from the server
     * @param receivedUUID   the UUID of the response Frame
     * @param inboundChannel the channel this Command got called from, might be null for direct calls
     */
    public void request(@NotNull final T params, @NotNull UUID receivedUUID, @NotNull Channel inboundChannel) {

        @NotNull ErrorParameter parameter = new ErrorParameter();
        parameter.code = 4;
        parameter.message = "Command not available";
        this.sendError(parameter, inboundChannel, receivedUUID);
    }

    /**
     * Override this Method to implement Handling of Response-Pakets.
     *
     * @param params         the parameters for execution, supplied from the server
     * @param receivedUUID   the UUID of the response Frame
     * @param inboundChannel the channel this Command got called from, might be null for direct calls
     */
    public void response(@NotNull final V params, @NotNull UUID receivedUUID, @NotNull Channel inboundChannel) {

        @NotNull ErrorParameter parameter = new ErrorParameter();
        parameter.code = 4;
        parameter.message = "Command not available";
        this.sendError(parameter, inboundChannel, receivedUUID);
    }


    /**
     * Override this Method to implement Handling of Error-Pakets.
     *
     * @param params         the parameters for execution, supplied from the server
     * @param receivedUUID   the UUID of the response Frame
     * @param inboundChannel the channel this Command got called from, might be null for direct calls
     */
    public void error(@NotNull final ErrorParameter params, @NotNull UUID receivedUUID, @Nullable Channel inboundChannel) {
        @Nullable String artifactID = params.getArtifactID();
        @Nullable LinkAnchor knownAnchor = null;
        @Nullable Project project = CommandsManager.getProjectForCommand(this);

        if (artifactID != null && project != null) {
            knownAnchor = SketchlinkManager.getInstance(project).getKnownAnchor(artifactID);
        }

        if (artifactID == null || !(knownAnchor instanceof LocalLinkAnchor)) {
            JOptionPane.showMessageDialog(null,
                    "Received an error from the Server.\n\nErrorCode: " + params.code + " Message: " + params.message
                            + "\n" + (artifactID != null ? "Problematic artifactID: " + artifactID + " is not a locally known id." : "")
            );
            return;
        }

        //HACK: If a previous update was of an unkown ID, resend as create OR if create was of known ID, send as Update
        //added because the server at the time of writing this code was incomplete, having an empty database most of the time
        if (params.code == 1) {
            new ChangeArtifactMetaDataRequestBuilder(knownAnchor).queue(project, inboundChannel, false);
        } else {
            if (params.code == 2) {
                new ChangeArtifactMetaDataRequestBuilder(knownAnchor).queue(project, inboundChannel, true);
            }
        }

    }


    public void sendRequest(@Nullable Payload data, @Nullable Channel channel, @Nullable UUID uuid, boolean sendAsBroadCast) {
        @NotNull CommandTopParameter topParameter = new CommandTopParameter(uuid);
        topParameter.type = sendAsBroadCast ? "broadcast" : "request"; //NON-NLS NON-NLS
        topParameter.message = getCommandIdentifier();
        topParameter.data = data;
        //System.out.println("Sending: " + topParameter.toString());

        if (channel != null && channel.isOpen()) {

            myRequestHistory.put(topParameter.id, data);

            channel.writeAndFlush(new TextWebSocketFrame(topParameter.toString()));
        } else {
            @Nullable Project project = CommandsManager.getProjectForCommand(this);
            if (project == null) {
                LOG.warn("Tried sending a request, but the supplied channel wasn't open and the Project couldn't be looked up: " + getCommandIdentifier());
                return;
            }

            myRequestHistory.put(topParameter.id, data);
            slWebsocketManager.sendBroadcast(topParameter.toString(), project);
        }
        //System.out.println("Sending request " + getCommandIdentifier() + " " + data);

    }


    public void sendResponse(@NotNull Payload data, @Nullable Channel channel, @NotNull UUID uuid) {
        @NotNull CommandTopParameter topParameter = new CommandTopParameter(uuid);
        topParameter.type = "response";
        topParameter.message = getCommandIdentifier();
        topParameter.data = data;
        if (channel != null && channel.isOpen()) {
            myRequestHistory.remove(uuid);
            channel.writeAndFlush(new TextWebSocketFrame(topParameter.toString()));
        } else {
            @Nullable Project project = CommandsManager.getProjectForCommand(this);
            if (project == null) {
                LOG.error("Tried sending a response, but the supplied channel wasn't open and the Project couldn't be looked up " + topParameter.id + " " + getCommandIdentifier());
                return;
            }
            myRequestHistory.remove(uuid);
            slWebsocketManager.sendBroadcast(topParameter.toString(), project);
        }

        //System.out.println("Sending response " + getCommandIdentifier() + " " + uuid.toString() + " Data:\n\t" + data);

    }

    public void sendError(@NotNull ErrorParameter parameter, @NotNull Channel channel, @NotNull UUID uuid) {
        @NotNull CommandTopParameter topParameter = new CommandTopParameter(uuid);
        topParameter.type = "error";
        topParameter.message = getCommandIdentifier();
        topParameter.data = parameter;
        channel.writeAndFlush(new TextWebSocketFrame(topParameter.toString()));
        myRequestHistory.remove(uuid);
    }


    /**
     * Transforms the JSON-formatted parameter to a class object of type T and then calls the overriden Method response(T)
     *
     * @param params         The parameters needed to response this command in JSon Format
     * @param inboundChannel The channel this message was received on
     */
    @Override
    @SuppressWarnings("unchecked")
    public void executeCmd(@NotNull CommandTopParameterReceive params, @NotNull Channel inboundChannel, @NotNull String receivedCommand) {
        LOG.info("Stub executing: " + this.getCommandIdentifier() + " Type: " + params.type + " GUID:" + params.id + "\nParameter: " + receivedCommand);
        //System.out.println("Stub executing: " + getCommandIdentifier() + " Type: " + params.type + " GUID:" + params.id + "\nOriginal: " + receivedCommand + "\n");
        /**
         * No Class Cast Exception should occur, since this class is abstract, has parameters and therefore all instances are an
         * extension to this.
         * @see plg.tools.slToolBox.getType for details
         *
         * Code unused, because the whole object has to be deserialized.
         */
//        Class<T> in;
//        Class<V> out;
//        try {
//            in = (Class<T>) slToolBox.getType(getClass(), 0);
//            out = (Class<V>) slToolBox.getType(getClass(), 1);
//        } catch (ClassCastException e) {
//            LOG.error(e);
//            return;
//        }
        if ("request".equals(params.type) || "broadcast".equals(params.type)) { //NON-NLS NON-NLS
            //Type test = new TypeToken<getRequestWrapClass()>(){}.getType();
            Type test = getRequestWrapClass();
            Object o = myGson.fromJson(receivedCommand, test);
            if (o instanceof PayloadWithData) {
                request((T) ((PayloadWithData) o).getData(), params.id, inboundChannel);
            } else {
                slGuiTools.ShowNotificationAtBottom("Forgot to implement request changes in " + getCommandIdentifier(), manager.getProject());
//                        JOptionPane.showMessageDialog(null, "Forgot to implement request changes in " + getCommandIdentifier());
            }

            return;
        }
        if ("response".equals(params.type)) {
            //Type test = new com.google.gson.reflect.TypeToken<responseWrap>(){}.getType();
            Type test = getResponseWrapClass();

            Object o = myGson.fromJson(receivedCommand, test);

            if (myRequestHistory.remove(params.id) == null) {
                LOG.warn("Received a response to an unknown request! " + params.id + " " + params.message + " " + params.type);
                return;
            }
            if (o instanceof PayloadWithData) {
                response((V) ((PayloadWithData) o).getData(), params.id, inboundChannel);
            } else {
                slGuiTools.ShowNotificationAtBottom("Forgot to implement response changes in " + getCommandIdentifier(), manager.getProject());
//                JOptionPane.showMessageDialog(null, "Forgot to implement response changes in " + getCommandIdentifier());
            }
            return;
        }
        if ("error".equals(params.type)) {
            if (myRequestHistory.remove(params.id) == null) {
                LOG.warn("Received an error to an unknown request! " + params.id + " " + params.message + " " + params.type);
                return;
            }

            errorWrap incParam = myGson.fromJson(receivedCommand, errorWrap.class);
            error(incParam.data, params.id, inboundChannel);
        }
    }

    public abstract Type getRequestWrapClass();

    public abstract Type getResponseWrapClass();

//doesn't work the simple way because of java's type erasure
//    private static class requestWrap<T extends Payload> {
//        public UUID id = null;
//        public String message = null;
//        public String type = null;
//        public T data;
//
//    }
//
//    private class responseWrap {
//        public UUID id = null;
//        public String message = null;
//        public String type = null;
//        public V data;
//
//        public V getData() {
//            return this.data;
//        }
//    }

    private class errorWrap {
        @Nullable
        public UUID id = null;
        @Nullable
        public String message = null;
        @Nullable
        public String type = null;
        @Nullable
        public ErrorParameter data = null;
    }


}
