/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.websocket;

import com.intellij.openapi.project.Project;
import io.netty.channel.Channel;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.websocket.commands.builder.slApiBuilder;
import plg.websocket.commands.builder.slApiRequestBuilder;
import plg.websocket.commands.builder.slApiResponseBuilder;

import java.util.*;

/**
 * Created by Peter on 29.10.2014.
 * Queues important messages under certain conditions.
 * Currently: While the createAnchorMode is active, keep the user from creating links. Only send them after
 */
public class SendQueue {
    private final List<MessageEntry> queuedMessages = new Vector<>();
    private final slWebsocketManager manager;
    private boolean sendNow = true;

    public SendQueue(@NotNull slWebsocketManager manager) {
        this.manager = manager;
    }

    public void sendQueue() {
        synchronized (queuedMessages) {
            Iterator<MessageEntry> iterator = queuedMessages.iterator();
            while (iterator.hasNext()) {
                MessageEntry next = iterator.next();
                next.send(manager.getProject());
                iterator.remove();
            }
        }
    }

    public int getQueueLenght() {
        return queuedMessages.size();
    }

    public boolean isSendDirect() {
        return this.sendNow;
    }

    public void setSendDirect(boolean sendNow) {
        this.sendNow = sendNow;
        if (sendNow) {
            sendQueue();
        }
    }

    public void sendRequest(@NotNull slApiRequestBuilder builder, @Nullable Channel channel) {
        if (sendNow) {
            builder.send(manager.getProject(), channel);
        } else {
            synchronized (queuedMessages) {
                RequestEntry entry = new RequestEntry(builder, channel);
                if (!queuedMessages.contains(entry))
                    queuedMessages.add(entry);
            }
        }
    }

    public void sendResponse(@NotNull slApiResponseBuilder builder, @NotNull Channel channel, @NotNull UUID responseID) {
        if (sendNow) {
            builder.send(manager.getProject(), channel, responseID);
        } else {
            synchronized (queuedMessages) {
                queuedMessages.add(new ResponseEntry(builder, channel, responseID));
            }
        }
    }

    @Nullable
    public slApiBuilder removeQueuedMessage(@NotNull slApiBuilder builder) {

        synchronized (queuedMessages) {
            MessageEntry result = null;
            for (MessageEntry message : queuedMessages) {
                if (builder.equals(message.getBuilder())) {
                    result = message;
                    break;
                }
            }
            if (result != null) {
                queuedMessages.remove(result);
                return result.getBuilder();
            }
        }
        return null;
    }

    public List<slApiBuilder> getQueuedMessageBuilders() {
        List<slApiBuilder> result = new ArrayList<>(queuedMessages.size());
        synchronized (queuedMessages) {
            for (MessageEntry queuedMessage : queuedMessages) {
                result.add(queuedMessage.getBuilder());
            }
        }
        return result;
    }


    private interface MessageEntry {

        public void send(@NotNull Project project);

        public slApiBuilder getBuilder();


    }

    private class RequestEntry implements MessageEntry {
        private final slApiRequestBuilder builder;
        private final Channel channel;

        public RequestEntry(@NotNull slApiRequestBuilder builder, @Nullable Channel channel) {
            this.builder = builder;
            this.channel = channel;
        }

        @Override
        public void send(@NotNull Project project) {
            this.builder.send(project, channel);
        }

        @Override
        public slApiBuilder getBuilder() {
            return builder;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            RequestEntry that = (RequestEntry) o;

            if (!builder.equals(that.builder)) return false;

            return true;
        }

        @Override
        public int hashCode() {
            return builder.hashCode();
        }
    }

    private class ResponseEntry implements MessageEntry {
        @NotNull
        private final slApiResponseBuilder builder;
        @Nullable
        private final Channel channel;
        @NotNull
        private final UUID uuid;

        public ResponseEntry(@NotNull slApiResponseBuilder builder, @Nullable Channel channel, @NotNull UUID uuid) {
            this.builder = builder;
            this.channel = channel;
            this.uuid = uuid;
        }

        @Override
        public void send(@NotNull Project project) {
            builder.send(project, channel, uuid);
        }

        @Override
        public slApiBuilder getBuilder() {
            return builder;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ResponseEntry that = (ResponseEntry) o;

            if (!builder.equals(that.builder)) return false;
            if (channel != null ? !channel.equals(that.channel) : that.channel != null) return false;
            if (!uuid.equals(that.uuid)) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = builder.hashCode();
            result = 31 * result + (channel != null ? channel.hashCode() : 0);
            result = 31 * result + uuid.hashCode();
            return result;
        }
    }

}
