/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.websocket;

import com.intellij.openapi.project.Project;
import io.netty.bootstrap.Bootstrap;
import io.netty.channel.Channel;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioSocketChannel;
import io.netty.handler.codec.http.DefaultHttpHeaders;
import io.netty.handler.codec.http.HttpClientCodec;
import io.netty.handler.codec.http.HttpHeaders;
import io.netty.handler.codec.http.HttpObjectAggregator;
import io.netty.handler.codec.http.websocketx.CloseWebSocketFrame;
import io.netty.handler.codec.http.websocketx.TextWebSocketFrame;
import io.netty.handler.codec.http.websocketx.WebSocketClientHandshakerFactory;
import io.netty.handler.codec.http.websocketx.WebSocketVersion;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

/**
 * Created by Peter on 22.02.14.
 */
public class SketchlinkChannelImpl implements SketchlinkChannel {
    @NotNull
    private final URI myUri;
    @NotNull
    private final ChannelManager myManager;
    @NotNull
    private final Project project;
    private String myName;
    private Channel myChannel;


    public SketchlinkChannelImpl(@NotNull URI uri, @NotNull String name, @NotNull Project project) {
        this.myUri = uri;
        this.myName = name;
        this.myManager = slWebsocketManager.getInstance(project).getChannelManager();
        this.project = project;

        //TODO: actual connect
//        ApplicationManager.getApplication().executeOnPooledThread(new Runnable() {
//            @Override
//            public void run() {
//                try {
//                    open();
//                } catch (Exception e) {
//                    e.printStackTrace();
//                }
//            }
//        });

    }

    @Nullable
    @Override
    public Channel getChannel() {
        return myChannel;
    }

    @Override
    public boolean send(@NotNull String textFrameData) {
        if (!this.isOpen()) {
            return false;
        }
        myChannel.writeAndFlush(new TextWebSocketFrame(textFrameData));
        return true;
    }

    @Nullable
    @Override
    public ChannelManager getManager() {
        //return slWebsocketManager.getInstance(this.project).getChannelManager();
        return myManager;
    }

    @Override
    public void requestClosing(int statuscode, @Nullable String reason) {
        if (!isOpen()) {
            return;
        }
        this.myChannel.writeAndFlush(new CloseWebSocketFrame(statuscode, reason));
        try {
            this.myChannel.closeFuture().sync();
        } catch (InterruptedException e) {
            JOptionPane.showMessageDialog(null, "SketchlinkChannelImpl: Tried to close channel, but got interrupted.\n\n" + e.getMessage());
        }

    }

    @Override
    public void forceClosing(int statuscode, @Nullable String reason) {
        if (!isOpen()) {
            return;
        }
        this.myChannel.writeAndFlush(new CloseWebSocketFrame(statuscode, reason));
        this.myChannel.close();
    }

    @Override
    public boolean isOpen() {
        return this.myChannel != null && this.myChannel.isOpen();
    }

    @NotNull
    @Override
    public String getName() {
        return this.myName;
    }

    @Override
    public boolean isConnectedTo(@NotNull URI uri) {
        return isOpen() && this.myUri.equals(uri);
    }

    @Override
    public boolean isConnectedTo(@NotNull String uri) {
        try {
            @NotNull URI uri2 = new URI(uri);
            if (isConnectedTo(uri2)) {
                return true;
            }
        } catch (URISyntaxException ignored) {
        }
        return false;
    }

    @NotNull
    @Override
    public URI getEndpointUri() {
        return this.myUri;
    }

    @Nullable
    @Override
    public URL getEndpointAsHTTP() {
        try {

            @NotNull URI asUri = new URI("http", myUri.getUserInfo(), myUri.getHost(), myUri.getPort(), ""/*myUri.getPath()*/, myUri.getQuery(), myUri.getFragment());
            return asUri.toURL();
        } catch (@NotNull MalformedURLException | URISyntaxException e) {
            e.printStackTrace();
        }
        return null;
    }

    @NotNull
    @Override
    public String getMyName() {
        return this.myName;
    }


    /**
     * Opens this channel, making it able to send and Receive Data
     *
     * @throws Exception
     */
    @Override
    public void open() throws Exception {
        if (this.isOpen()) {
            //already open
            return;
        }

        @NotNull EventLoopGroup group = new NioEventLoopGroup();
        try {

            @NotNull Bootstrap b = new Bootstrap();
            String protocol = this.myUri.getScheme();
            if (!"ws".equals(protocol)) {
                throw new IllegalArgumentException("Unsupported protocol: " + protocol);
            }

            @NotNull HttpHeaders customHeaders = new DefaultHttpHeaders();
            customHeaders.add("MyHeader", "MyValue");

            // Connect with V13 (RFC 6455 aka HyBi-17). You can change it to V08 or V00.
            // If you change it to V00, ping is not supported and remember to change
            // HttpResponseDecoder to WebSocketHttpResponseDecoder in the pipeline.
            @NotNull final ClientHandler handler =
                    new ClientHandler(
                            WebSocketClientHandshakerFactory.newHandshaker(
                                    this.myUri, WebSocketVersion.V13, null, false, customHeaders), project
                    );

            b.group(group)
                    .channel(NioSocketChannel.class)
                    .handler(new ChannelInitializer<SocketChannel>() {
                        @Override
                        public void initChannel(@NotNull SocketChannel ch) throws Exception {
                            ChannelPipeline pipeline = ch.pipeline();
                            pipeline.addLast("http-codec", new HttpClientCodec());
                            pipeline.addLast("aggregator", new HttpObjectAggregator(8192));
                            pipeline.addLast("ws-handler", handler);
                        }
                    });


            //System.out.println("WebSocket Client connecting to " + this.myUri);
            myChannel = b.connect(this.myUri.getHost(), this.myUri.getPort()).sync().channel();
            myManager.addChannelToDefaultGroup(this);
            handler.handshakeFuture().sync();

            // WebSocketClientHandler will close the connection when the server
            // responds to the CloseWebSocketFrame.
            myChannel.closeFuture().sync();

        } finally {
            group.shutdownGracefully();
        }

    }


}
