/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.websocket.commands;

import com.google.gson.reflect.TypeToken;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.project.Project;
import com.intellij.util.ui.UIUtil;
import io.netty.channel.Channel;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.LinkAnchor;
import plg.sketchlinks.RemoteLinkAnchorS;
import plg.sketchlinks.SketchlinkManager;
import plg.tools.slGuiTools;
import plg.ui.preview.PreviewHandler;
import plg.ui.preview.PreviewManager;
import plg.websocket.CommandsManager;
import plg.websocket.PayloadWithData;
import plg.websocket.RunCmdStub;
import plg.websocket.commands.parameter.GetSketchesRequestParameter;
import plg.websocket.commands.parameter.GetSketchesResponseParameter;

import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.UUID;

/**
 * Created by Peter on 07.10.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class GetSketchesCMD extends RunCmdStub<GetSketchesRequestParameter, GetSketchesResponseParameter> {
    public static final String COMMAND = "getSketches";


    /**
     * Receives a list of sketches and shows them in a toolwindow, if this list is not empty.
     *
     * @param params         the parameters for execution, supplied from the server
     * @param receivedUUID   the UUID of the response Frame
     * @param inboundChannel the channel this Command got called from, might be null for direct calls
     */
    @Override
    public void response(@NotNull final GetSketchesResponseParameter params, @NotNull UUID receivedUUID, @NotNull Channel inboundChannel) {
//        System.out.println("unlinked sketches received: " + params.toString());
        @Nullable final Project project = CommandsManager.getProjectForCommand(this);
        if (project == null) {
            System.out.println("Command not registered");
            return;
        }
        @NotNull final List<LinkAnchor> sketches = new ArrayList<>();
        @NotNull final SketchlinkManager manager = SketchlinkManager.getInstance(project);
        ApplicationManager.getApplication().executeOnPooledThread(new Runnable() {
            @Override
            public void run() {
                for (@NotNull String sketchID : params.sketches) {
                    if (!sketchID.toLowerCase().startsWith("s-")) {
                        continue;
                    }
                    @NotNull LinkAnchor knownAnchor = manager.addRemote(sketchID, null, project);
                    if (knownAnchor instanceof RemoteLinkAnchorS) {
                        sketches.add(knownAnchor);
                        ((RemoteLinkAnchorS) knownAnchor).preparePreview();
                    }
                }
                if (sketches.isEmpty()) {
                    slGuiTools.ShowNotificationAtBottom("No currently unlinked sketches found.", project);
//                    JOptionPane.showMessageDialog(null, "No currently unlinked sketches found.");
                    return;
                }
                String title = "Filter: " + params.filter;
                if ("all".equals(params.filter)) {
                    title = "Existing sketches on the server";
                } else {
                    if ("unlinked".equals(params.filter)) {
                        title = "Currently unlinked sketches";
                    }
                }
                showSketchesInPreview(sketches, title, project);
            }
        });


    }

    private void showSketchesInPreview(@NotNull final List<LinkAnchor> sketches, @NotNull final String title, @NotNull final Project project) {
        UIUtil.invokeLaterIfNeeded(new Runnable() {
            @Override
            public void run() {
                PreviewManager.getInstance(project).setToolwindowMode();
                @NotNull PreviewHandler previewHandler = PreviewManager.getInstance(project).getPreviewHandler();
                previewHandler.setHighlighting(new HashSet<String>()); //disable highlighting
                previewHandler.setPinnedContent(null);
                previewHandler.setTitle(title);
                previewHandler.setContent(sketches);
                previewHandler.setVisible(true);
            }
        });
    }

//    @Override
//    public void sendRequest(@Nullable Payload data, @Nullable Channel channel, @Nullable UUID uuid, boolean sendAsBroadCast) {
//        super.sendRequest(data, channel, uuid, sendAsBroadCast);
//        System.out.println("Send? "+ data);
//    }

    @NotNull
    @Override
    public String getCommandIdentifier() {
        return COMMAND;
    }

    @Override
    public Type getRequestWrapClass() {
        return new TypeToken<requestWrap>() {
        }.getType();
    }

    @Override
    public Type getResponseWrapClass() {
        return new TypeToken<responseWrap>() {
        }.getType();
    }

    private static class requestWrap implements PayloadWithData<GetSketchesRequestParameter> {
        @Nullable
        public UUID id = null;
        @Nullable
        public String message = null;
        @Nullable
        public String type = null;
        public GetSketchesRequestParameter data;

        @Override
        public GetSketchesRequestParameter getData() {
            return data;
        }
    }

    private static class responseWrap implements PayloadWithData<GetSketchesResponseParameter> {
        @Nullable
        public UUID id = null;
        @Nullable
        public String message = null;
        @Nullable
        public String type = null;
        public GetSketchesResponseParameter data;

        public GetSketchesResponseParameter getData() {
            return this.data;
        }
    }

}
