/*
 * Copyright (c) 2014. by Peter Schmitz (ptsmail01@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package plg.websocket.commands.builder;

import com.google.gson.GsonBuilder;
import com.intellij.openapi.project.Project;
import io.netty.channel.Channel;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import plg.sketchlinks.LinkAnchor;
import plg.sketchlinks.LocalLinkAnchor;
import plg.websocket.RunCmd;
import plg.websocket.commands.CreateLinkCMD;
import plg.websocket.commands.DeleteLinkCMD;
import plg.websocket.commands.UpdateLinkCMD;
import plg.websocket.commands.parameter.LinkRequestParameter;
import plg.websocket.slWebsocketManager;

import java.util.Collection;
import java.util.HashSet;

/**
 * Created by Peter on 05.07.2014.
 * Part of Sketchlink Plugin for IntelliJ
 */
public class LinkRequestBuilder implements slApiRequestBuilder {
    @NotNull
    private final LinkRequestParameter parameter;
    private final LinkAnchor anchorA;
    private final LinkAnchor anchorB;
    private boolean delete = false;
    private boolean update = false;


    public LinkRequestBuilder(@NotNull LinkAnchor anchor1, @NotNull LinkAnchor anchor2, boolean isDeprecated, boolean update) {
        this(anchor1, anchor2, isDeprecated, update, false);
    }

    public LinkRequestBuilder(@NotNull LinkAnchor anchor1, @NotNull LinkAnchor anchor2, boolean isDeprecated, boolean update, boolean delete) {
        anchorA = anchor1;
        anchorB = anchor2;
        parameter = new LinkRequestParameter();
        parameter.artifactID1 = anchor1.getAnchorId();
        parameter.artifactID2 = anchor2.getAnchorId();
        parameter.isDeprecated = isDeprecated;
        this.delete = delete;
        this.update = update;
    }


    @NotNull
    @Override
    public Collection<String> getContainedAnchorIDs() {
        Collection<String> result = new HashSet<>(4);
        result.add(parameter.artifactID1);
        result.add(parameter.artifactID2);
        return result;
    }


    public String toString() {
        return new GsonBuilder().serializeNulls().create().toJson(parameter);
    }

    @NotNull
    private String getCommand() {
        if (delete) {
            return DeleteLinkCMD.COMMAND;
        }
        if (update) {
            return UpdateLinkCMD.COMMAND;
        } else {
            return CreateLinkCMD.COMMAND;
        }
    }

    public void queue(@NotNull Project project, @Nullable Channel channel) {
        slWebsocketManager.getInstance(project).getSendQueue().sendRequest(this, channel);
    }

    public void send(@NotNull Project project, @Nullable Channel channel) {
        if (anchorA instanceof LocalLinkAnchor && !((LocalLinkAnchor) anchorA).isValid()) {
            return;
        }
        if (anchorB instanceof LocalLinkAnchor && !((LocalLinkAnchor) anchorB).isValid()) {
            return;
        }
//        System.out.println("LinkRequest.java: Link send between "+anchorA.getDescription()+" and "+anchorB.getDescription());
        @Nullable RunCmd cmd = slWebsocketManager.getInstance(project).getCommandsManager().getCommand(getCommand());
        if (cmd instanceof CreateLinkCMD) {
            cmd.sendRequest(parameter, channel, null, false);
        }
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        LinkRequestBuilder that = (LinkRequestBuilder) o;

        if (delete != that.delete) return false;
        if (update != that.update) return false;
        if (!anchorA.equals(that.anchorA)) return false;
        if (!anchorB.equals(that.anchorB)) return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = (delete ? 1 : 0);
        result = 31 * result + (update ? 1 : 0);
        result = 31 * result + anchorA.hashCode();
        result = 31 * result + anchorB.hashCode();
        return result;
    }

    //    public void send(@NotNull CreateLinkCMD cmd, @Nullable Channel channel) {
//        Project project = CommandsManager.getProjectForCommand(cmd);
//        if (project == null)
//            return;
//        cmd.sendRequest(parameter, channel, null, false);
//    }
}
